#define THRESHOLD_SUPPORT

#include <stdlib.h>
#include <stdio.h>
#include <assert.h>

#include <set>

#include "OSGConfig.h"
#include "OSGLog.h"
#include "OSGBaseFunctions.h"
#include "OSGBaseThread.h"
#include "OSGSocketSelection.h"
#include "OSGBinaryMessage.h"
#include "OSGGroupIBPipeline.h"
#include "OSGConnectionType.h"
#include "OSGBinaryMessage.h"
#include "OSGTime.h"

#include "OSGminiz.h"
#include "OSGLZ4.h"

#pragma intrinsic(memcpy)

OSG_USING_NAMESPACE

GroupIBPipeline::GroupIBPipeline() : Inherited(), _initialized(false),
    _maxNodeCount(1)
{
}

GroupIBPipeline::~GroupIBPipeline(void)
{
    for (int index = 0; index < _connections.size(); ++index)
        delete _connections[index];
    for (int index = 0; index < _next.size(); ++index)
        delete _next[index];
    _connections.clear();
    _next.clear();
}

const ConnectionType *GroupIBPipeline::getType() const
{
    return &_type;
}

GroupConnection::Channel
GroupIBPipeline::connectPoint(const std::string &address, Time timeout)
{
    auto connection = new PointIBConnection();
    Channel channel;
    try
    {
        channel = connection->connectPoint(address, timeout);
        if (channel >= 0)
        {
            channel = newChannelIndex(ChannelIndex(_connections.size()));
            _connections.push_back(connection);
            _readIndex = 0;
            setLocalInterface(connection->getLocalInterface());
            setRemoteInterface(connection->getRemoteInterface());
        }
    }
    catch (std::exception e)
    {
        delete connection;
        throw ReadError(e.what());
    }
    return channel;
}

void GroupIBPipeline::disconnect(Channel channel)
{
    ChannelIndex index = channelToIndex(channel);
    try
    {
        delete _connections[index];
        _connections[index] = NULL;
    }
    catch (...)
    {
    }
    _connections.erase(_connections.begin() + index);
    delChannelIndex(index);
    _readIndex = 0;
}

void GroupIBPipeline::disconnect(const std::vector<Channel> &channels)
{
    for(UInt32 index=0;index<channels.size();++index)
    {
        ChannelIndex cIndex = channelToIndex(channels[index]);
        delete _connections[cIndex];
        _connections[cIndex] = NULL;
    }
    for(UInt32 index=0;index<_connections.size();++index)
    {
        if (_connections[index] == NULL)
        {
            _connections.erase(_connections.begin() + index);
            _remoteAddresses.erase(_remoteAddresses.begin() + 1);
            delChannelIndex(index);
            index--;
        }
    }
}

/*! accept an icomming point connection. If timeout is reached,
    -1 is returned. If timeout is -1 then wait without timeout
 */
GroupConnection::Channel GroupIBPipeline::acceptPoint(Time timeout)
{
    Channel channel = -1;
    auto connection = new PointIBConnection(_acceptSocket);
    try
    {
        if (connection->acceptPoint(timeout))
        {
            channel = newChannelIndex(ChannelIndex(_connections.size()));
            _connections.push_back(connection);
            _readIndex = 0;
            setLocalInterface(connection->getLocalInterface());
            setRemoteInterface(connection->getRemoteInterface());
        }
        else
        {
            if (_connections.empty())
            {
                setLocalInterface("");
                setRemoteInterface("");
            }
        }
    }
    catch (std::exception e)
    {
        delete connection;
        throw ReadError(e.what());
    }
    return channel;
}

std::string GroupIBPipeline::bind(const std::string &address)
{
    return Inherited::bind(address);
}

void GroupIBPipeline::setParams(const std::string &params)
{
    Inherited::setParams(params);
}

Connection::Channel
GroupIBPipeline::selectChannel(Time timeout) throw(ReadError)
{
    Int32 maxnread = 0, nread;
    ChannelIndex index;
    SocketSelection selection, result;

    // if there is data in the read buffer, return current channel
    if (_zeroCopyThreshold != 1 && _currentReadBuffer != readBufEnd())
    {
        FFATAL(("Channel change ignores data in current buffer"))
        return indexToChannel(_readIndex);
    }

    if (_selection[_readIndex] && _connections[_readIndex]->getAvailable())
        return indexToChannel(_readIndex);

    int newIndex = -1;
    Time startTime = getSystemTime();
    do
    {
        for (index = 0; index < _connections.size(); ++index)
        {
            if (!_selection[index])
                continue;
            if (_connections[index]->getAvailable())
            {
                newIndex = index;
                break;
            }
        }
        if (newIndex < 0)
            osgyield();
    } while (newIndex < 0 &&
             (timeout < 0 || getSystemTime() - startTime < timeout));
    if (newIndex < 0)
        return -1;
    _readIndex = newIndex;
    return indexToChannel(_readIndex);
}

int GroupIBPipeline::getAvailable(void)
{
    return _connections[_readIndex]->getAvailable();
}

int GroupIBPipeline::getAvailable(Channel channel)
{
    return _connections[channelToIndex(channel)]->getAvailable();
}

int GroupIBPipeline::waitAvailable(Channel channel)
{
    _connections[channelToIndex(channel)]->selectChannel();
    return _connections[channelToIndex(channel)]->getAvailable();
}

Connection::Channel
GroupIBPipeline::selectSingleChannel(Channel channel,
                                     Time timeout) throw(ReadError)
{
    try
    {
        _readIndex = channelToIndex(channel);
        if (_selection[_readIndex] && _connections[_readIndex]->getAvailable())
            return channel;
        return _connections[_readIndex]->selectChannel(timeout);
    }
    catch (SocketException &e)
    {
        throw ReadError(e.what());
    }
}

Int32 GroupIBPipeline::selectSingleChannelAndGetAvaiable(
    Channel channel, Time timeout) throw(ReadError)
{
    auto readIndex = channelToIndex(channel);
    auto connection = _connections[readIndex];

    if (_selection[readIndex] && connection->getAvailable())
        return connection->getAvailable();

    try
    {
        if (connection->selectChannel(timeout) == -1)
            return -1;
        return connection->getAvailable();
    }
    catch (SocketException &e)
    {
        throw ReadError(e.what());
    }
}

bool GroupIBPipeline::hasAvaiable(Channel channel,
                                  Int32 timeout) throw(ReadError)
{
    auto readIndex = channelToIndex(channel);
    auto connection = _connections[readIndex];

    if (!_selection[readIndex])
        return false;
    if (connection->getAvailable())
        return true;
    if (connection->selectChannel(timeout) == -1)
        return -1;
    return connection->getAvailable();
}

std::vector<Connection::Channel>
GroupIBPipeline::selectChannels(Time timeout) throw(ReadError)
{
    std::vector<Connection::Channel> channels;

    if (selectChannel(timeout) == -1)
        return channels;
    try
    {
        for (UInt32 index = 0; index < _connections.size(); ++index)
        {
            if (_connections[index]->getAvailable() > 0)
                channels.push_back(indexToChannel(index));
        }
    }
    catch (std::exception &e)
    {
        throw ReadError(e.what());
    }
    return channels;
}

void GroupIBPipeline::flush(void)
{
    GroupConnection::flush();
}

void GroupIBPipeline::setDelay(bool s)
{
    for (UInt32 index = 0; index < _sockets.size(); ++index)
        _connections[index]->setDelay(s);
}

bool GroupIBPipeline::wait(Time timeout) throw(ReadError)
{
    UInt32 len;
    UInt32 index;
    UInt32 tag = 314156;
    int missing = _connections.size();

    try
    {
        for (index = 0; index < _connections.size(); ++index)
        {
            auto connection = _connections[index];
            if (connection->selectChannel(timeout) >= 0)
            {
                connection->getValue(tag);
                missing--;
                if (tag != 314156)
                {
                    FFATAL(("Stream out of sync in SockConnection (tag is "
                            "%u)\n",
                            tag));
                    throw ReadError("Stream out of sync");
                }
            }
            else
            {
                return false;
            }
        }
    }
    catch (SocketException &e)
    {
        throw ReadError(e.what());
    }
    return true;
}

void GroupIBPipeline::signal(void) throw(WriteError)
{
    UInt32 tag = 314156;
    putValue(tag);
    flush();
}

GroupConnection *GroupIBPipeline::create(void)
{
    return new GroupIBPipeline();
}

SizeT GroupIBPipeline::read(MemoryHandle mem, SizeT size)
{
    return _connections[_readIndex]->read(mem, size);
}

SizeT GroupIBPipeline::read(Channel channel, MemoryHandle mem, SizeT size)
{
    ChannelIndex readIndex = channelToIndex(channel);
    return _connections[readIndex]->read(mem, size);
}

void GroupIBPipeline::readBuffer()
{
    int size;
    int len;

    // read buffer header
    len = _connections[_readIndex]
              ->read(&_socketReadBuffer[0], sizeof(SocketBufferHeader));
    if (len == 0)
        throw ReadError("Channel closed");

    // read remaining data
    size = osgntohl(
        reinterpret_cast<SocketBufferHeader *>(&_socketReadBuffer[0])->size);
    len = _connections[_readIndex]->read(
        (MemoryHandle) & _socketReadBuffer[sizeof(SocketBufferHeader)], size);
    if (len == 0)
        throw ReadError("Channel closed");
    readBufBegin()->setDataSize(size);
}

void GroupIBPipeline::write(const MemoryHandle mem, SizeT size, bool force_unicast)
{
    if (force_unicast)
    {
        for (int index = 0; index < _connections.size(); ++index)
            _connections[index]->write(mem, size);
        return;
    }
    if (!_initialized)
        initialize();
    for (int index = 0; index < _next.size(); ++index)
        _next[index]->write(mem, size);
}

void GroupIBPipeline::write(Channel channel, const MemoryHandle mem, SizeT size)
{
    ChannelIndex index = channelToIndex(channel);
    try
    {
        _connections[index]->write(mem, size);
    }
    catch (std::exception &e)
    {
        throw WriteError(e.what());
    }
}

void GroupIBPipeline::writeBuffer(void)
{
    UInt32 size = writeBufBegin()->getDataSize();

    if (!size)
        return;
    auto header =
        reinterpret_cast<SocketBufferHeader *>(&_socketWriteBuffer[0]);
    header->magic = 0x12345678;
    header->id = ++_id;
    header->size = osghtonl(size);
    header->compressed_size = osghtonl(0);

    UInt8 *data = &_socketWriteBuffer[0];
    _write_buffer_bytes += size;

    Real64 t = OSG::getSystemTime();
    write(data, size + sizeof(SocketBufferHeader));

    _write_buffer_cbytes += size;
    _write_buffer_time += (OSG::getSystemTime() - t);
}

void GroupIBPipeline::initialize()
{
    UInt32 index, len;
    std::string address;
    UInt32 port;
    std::vector<std::string> addresses;

    for (index = 0; index < _connections.size(); ++index)
    {
        _connections[index]->getValue(address);
        addresses.push_back(address);
    }
    _maxNodeCount = 1;
    if (_connections.size() > 32)
        _maxNodeCount = 2;
    initialize(NULL, _connections, addresses);
    _initialized = true;
}

void GroupIBPipeline::initialize(PointIBConnection *target,
                                 const PointConnections &connections,
                                 const std::vector<std::string> &addresses)
{
    std::vector<std::string> next;

    int start = 0;
    for (int index = 0; index < _maxNodeCount; ++index)
    {
        int groupCount = addresses.size() / _maxNodeCount;
        int groupCountRem = addresses.size() % _maxNodeCount;
        if (groupCountRem > 0)
            groupCount++;
        if (index == _maxNodeCount - 1)
            groupCount = addresses.size() - start;
        int end = start + groupCount;
        if (groupCount == 0)
            continue;
        next.push_back(addresses[start]);
        PointConnections newConnections(connections.begin() + start + 1,
                                        connections.begin() + end);
        std::vector<std::string> newAddresses(addresses.begin() + start + 1,
                                              addresses.begin() + end);
        initialize(connections[start], newConnections, newAddresses);
        start += groupCount;
    }
    if (target != NULL)
    {
        target->putValue((UInt32)next.size());
        for (int index = 0; index < next.size(); ++index)
            target->putValue(next[index]);
        target->flush();
    }
    else
    {
        for (int index = 0; index < next.size(); ++index)
        {
            PointIBConnection *nextConnection = new PointIBConnection();
            nextConnection->connectPoint(next[index]);
            _next.push_back(nextConnection);
        }
    }
    FLOG(("Init group done"));
}

ConnectionType GroupIBPipeline::_type(&GroupIBPipeline::create, "IBPipeline");

#ifdef __sgi
#pragma set woff 1174
#endif

#ifdef OSG_LINUX_ICC
#pragma warning(disable : 177)
#endif

namespace
{
static Char8 cvsid_cpp[] = "@(#)$Id: $";
static Char8 cvsid_hpp[] = OSG__HEADER_CVSID;
}
