#ifdef __linux

#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <dlfcn.h>
#include <poll.h>
#include <malloc.h>
#include <fcntl.h>
#include <pthread.h>

#include <numeric>

#include "OSGVerbsDefinition.h"
#include "OSGConfig.h"
#include "OSGLog.h"
#include "OSGBaseFunctions.h"
#include "OSGSocketSelection.h"
#include "OSGIBVerbs.h"
#include "OSGConnectionHelper.h"
#include "OSGConnectionType.h"

#include "OSGminiz.h"
#include "OSGLZ4.h"

#pragma intrinsic(memcpy)

OSG_USING_NAMESPACE

IBVerbs::IBVerbs()
    : _queuePair(NULL),
      _completionQueue(NULL), _protectionDomain(NULL),
      _completionChannel(NULL), _context(NULL), _deviceList(NULL),
      _eventsToAck(0),
      _maxBufferSize(65536 * 2),
      _maxBufferCount(6),
      _maxEventsToAck(200),
      _postedRecvCountAck(0),
      _connectionClosed(false),
      _isReading(false),
      _ibPort(0),
      _loadedRecvRegions(),
      _freeSendRegions()
{
    readSymbols();
    char thread_name[256];
    sprintf(thread_name, "IBVerbsThread%p", this);
    _checkThread = BaseThread::get(thread_name);
}

IBVerbs::~IBVerbs(void)
{
    disconnect();
    _checkThread->subRef();
}

int IBVerbs::getAvailable(void)
{
    int available = 0;
    if (_loadedRecvRegions.empty())
        pollCompletion();
    for (auto i = _loadedRecvRegions.begin(); i != _loadedRecvRegions.end();
         ++i)
        available += i->length;
    return available;
}

/* connect two infiniband devices. The given socket is used to transmit
 * infiniband address information */
void IBVerbs::connect(StreamSocket socket)
{
    _socket = socket;
    open();
    socket.send((MemoryHandle) & _localIBAddress, sizeof(IBAddress));
    socket.recv((MemoryHandle) & _remoteIBAddress, sizeof(IBAddress));
    connectQueuePair();
}

/* Accept an incomming connection request */
void IBVerbs::accept(StreamSocket socket)
{
    _socket = socket;
    open();
    socket.recv((MemoryHandle) & _remoteIBAddress, sizeof(IBAddress));
    socket.send((MemoryHandle) & _localIBAddress, sizeof(IBAddress));
    connectQueuePair();
}

/* Disconnect infiniband devices and free all resources */
void IBVerbs::disconnect()
{
    _connectionClosed = true;
    char buffer = ' ';
    _socket.send(&buffer,1);
    BaseThread::join(_checkThread);
    if (_completionQueue && _eventsToAck > 0)
        ibv_ack_cq_events_p(_completionQueue, _eventsToAck);
    _eventsToAck = 0;
    if (_queuePair)
        ibv_destroy_qp_p(_queuePair);
    _queuePair = NULL;
    if (_completionQueue)
        ibv_destroy_cq_p(_completionQueue);
    _completionQueue = NULL;

    std::for_each(_recvRegions.begin(), _recvRegions.end(),
                  [this](ibv_mr *region)
    { destroyRegion(region); });
    _recvRegions.clear();
    std::for_each(_sendRegions.begin(), _sendRegions.end(),
                  [this](ibv_mr *region)
    { destroyRegion(region); });
    _sendRegions.clear();

    if (_protectionDomain)
        ibv_dealloc_pd_p(_protectionDomain);
    _protectionDomain = NULL;
    if (_completionChannel != NULL)
        ibv_destroy_comp_channel_p(_completionChannel);
    _completionChannel = NULL;
    if (_context != NULL)
        ibv_close_device_p(_context);
    _context = NULL;
    if (_deviceList)
        ibv_free_device_list_p(_deviceList);
    _deviceList = NULL;
}

void IBVerbs::recv(void *buffer, size_t len)
{
    // Send ack if this is the first read after a write operation.
    // For received ack-packages no response is sent. So we are sending
    // the number of free recv-buffers to the writer.
    if (!_isReading)
    {
        postAck();
        _isReading = true;
    }
    u_int8_t *dstBuffer = (u_int8_t *)buffer;
    while (len > 0)
    {
        waitForRecvedData();
        LoadedData &data = _loadedRecvRegions.front();
        size_t amount = std::min(len, data.length);
        memcpy(dstBuffer, data.buffer, amount);
        dstBuffer += amount;
        data.buffer += amount;
        data.length -= amount;
        len -= amount;
        // keep current buffer, if bytes are remaining
        if (data.length)
            return;
        ibv_mr *region = data.region;
        _loadedRecvRegions.pop_front();
        postRecv(region);
        postAck();
    }
}

void IBVerbs::send(const void *buffer, size_t len)
{
    _isReading = false;
    u_int8_t *srcBuffer = (u_int8_t *)buffer;
    while (len > 0)
    {
        waitForSendRegion();
        ibv_mr *region = _freeSendRegions.front();
        _freeSendRegions.pop();
        size_t amount = std::min(len, _maxBufferSize);
        memcpy(region->addr, srcBuffer, amount);
        srcBuffer += amount;
        len -= amount;
        waitForReceiverCanRecv();
        // needed to repost ack buffers
        if ((char)(_postedRecvCount - _postedRecvCountAck) < (_maxBufferCount - 2))
            pollCompletion();
        postSend(region, amount);
    }
}

/* Initialize all infiniband data structures */
void IBVerbs::open(const std::string &name)
{
    _connectionClosed = false;
    _deviceList = ibv_get_device_list_p(NULL);
    if (!_deviceList)
        throw IBError("unable to get devices", errno);
    ibv_device *dev = NULL;
    for (int i = 0; _deviceList[i] != NULL; ++i)
        if (name == "" || name == ibv_get_device_name_p(_deviceList[i]))
            dev = _deviceList[i];
    if (dev == NULL)
        throw IBError("device not found");
    _context = ibv_open_device_p(dev);
    if (!_context)
        throw IBError("unable to open device", errno);
    _completionChannel = ibv_create_comp_channel_p(_context);
    if (!_completionChannel)
        throw IBError("unable to create event channel", errno);
    _protectionDomain = ibv_alloc_pd_p(_context);
    if (!_protectionDomain)
        throw IBError("unable to allocate protection domain", errno);
    _freeSendRegions = std::queue<ibv_mr *>();
    _recvRegions.clear();
    _sendRegions.clear();
    for (int i = 0; i < _maxBufferCount; ++i)
    {
        _recvRegions.push_back(createRegion());
        ibv_mr *sendRegion = createRegion();
        _sendRegions.push_back(sendRegion);
        _freeSendRegions.push(sendRegion);
    }

    _completionQueue = ibv_create_cq_p(_context, _maxBufferCount * 2 + 1, NULL,
                                       _completionChannel, 0);
    if (!_completionQueue)
        throw IBError("unable to create completion queue", errno);

    ibv_device_attr device_attr;
    if (ibv_query_device_p(_context,&device_attr))
        throw IBError("unable to get device attributes", errno);

    struct ibv_port_attr portAttr;
    int ibPort;
    for(ibPort=1;ibPort<=device_attr.phys_port_cnt;++ibPort)
    {
        if (ibv_query_port_p(_context, ibPort, &portAttr))
            throw IBError("Unable to query port");
        if (portAttr.state == IBV_PORT_ACTIVE)
            break;
        if (ibPort == device_attr.phys_port_cnt)
            throw IBError("No active infiniband port found");
    }
    _ibPort = ibPort;

    int maxQPWrite = std::min(256,device_attr.max_qp_wr);

    struct ibv_qp_init_attr initAttr;
    memset(&initAttr, 0, sizeof(initAttr));
    initAttr.send_cq = _completionQueue;
    initAttr.recv_cq = _completionQueue;
    initAttr.cap.max_send_wr = maxQPWrite;
    initAttr.cap.max_recv_wr = maxQPWrite;
    initAttr.cap.max_send_sge = 1;
    initAttr.cap.max_recv_sge = 1;
    initAttr.qp_type = IBV_QPT_RC;
    _queuePair = ibv_create_qp_p(_protectionDomain, &initAttr);
    if (!_queuePair)
        throw IBError("unable to create queue pair", errno);

    struct ibv_qp_attr modifyAttr;
    memset(&modifyAttr, 0, sizeof(modifyAttr));
    modifyAttr.qp_state = IBV_QPS_INIT;
    modifyAttr.pkey_index = 0;
    modifyAttr.port_num = _ibPort;
    modifyAttr.qp_access_flags = 0;
    modifyAttr.max_rd_atomic = 4;
    modifyAttr.max_dest_rd_atomic = 4;
    modifyAttr.qp_access_flags =
        (IBV_ACCESS_LOCAL_WRITE | IBV_ACCESS_REMOTE_WRITE |
         IBV_ACCESS_REMOTE_READ);

    if (ibv_modify_qp_p(_queuePair, &modifyAttr,
                        IBV_QP_STATE | IBV_QP_PKEY_INDEX | IBV_QP_PORT |
                            IBV_QP_ACCESS_FLAGS))
        throw IBError("unable to init queue pair", errno);

    _postedRecvCount = 0;
    _postedRecvCountAck = 0;
    std::for_each(_recvRegions.begin(), _recvRegions.end(),
                  [this](ibv_mr *region)
    { postRecv(region); });

    _remoteUsedRecvCount = 0;
    _remotePostedRecvCount = _recvRegions.size();

    // content of inline function ibv_req_notify_cq
    if (_completionQueue->context->ops.req_notify_cq(_completionQueue, 0))
        throw IBError("Couldn't request CQ notification");

    _localIBAddress.lid = portAttr.lid;
    _localIBAddress.qpn = _queuePair->qp_num;
    _localIBAddress.packetSequenceNumber = lrand48() & 0xffffff;
    _localIBAddress.maxMtu = portAttr.max_mtu;
    if (!_localIBAddress.lid)
        throw IBError("Couldn't get local LID");
}

void IBVerbs::connectQueuePair()
{
    struct ibv_qp_attr attr;
    memset(&attr, 0, sizeof(attr));
    attr.qp_state = IBV_QPS_RTR;
    int mtu = std::min(_remoteIBAddress.maxMtu,_localIBAddress.maxMtu);
    attr.path_mtu = (ibv_mtu)mtu;
    attr.dest_qp_num = _remoteIBAddress.qpn;
    attr.rq_psn = _remoteIBAddress.packetSequenceNumber;
    attr.max_dest_rd_atomic = 1;
    attr.min_rnr_timer = 12;
    attr.ah_attr.is_global = 0;
    attr.ah_attr.dlid = _remoteIBAddress.lid;
    attr.ah_attr.sl = 0;
    attr.ah_attr.src_path_bits = 0;
    attr.ah_attr.port_num = _ibPort;

    if (ibv_modify_qp_p(_queuePair, &attr,
                        IBV_QP_STATE | IBV_QP_AV | IBV_QP_PATH_MTU |
                            IBV_QP_DEST_QPN | IBV_QP_RQ_PSN |
                            IBV_QP_MAX_DEST_RD_ATOMIC | IBV_QP_MIN_RNR_TIMER))
        throw IBError("Failed to modify QP to RTR", errno);

    attr.qp_state = IBV_QPS_RTS;
    attr.timeout = 14;
    attr.retry_cnt = 7;
    attr.rnr_retry = 7;
    attr.sq_psn = _localIBAddress.packetSequenceNumber;
    attr.max_rd_atomic = 1;
    if (ibv_modify_qp_p(_queuePair, &attr,
                        IBV_QP_STATE | IBV_QP_TIMEOUT | IBV_QP_RETRY_CNT |
                            IBV_QP_RNR_RETRY | IBV_QP_SQ_PSN |
                            IBV_QP_MAX_QP_RD_ATOMIC))
        throw IBError("Failed to modify QP to RTS", errno);
    _checkThread->runFunction(checkConnection,this);
}

void IBVerbs::postRecv(ibv_mr *region)
{
    if (_completionChannel == NULL || _connectionClosed)
        throw IBError("Conneciton closed");
    struct ibv_sge list;
    memset(&list, 0, sizeof(ibv_sge));
    list.addr = (uintptr_t)region->addr;
    list.length = _maxBufferSize;
    list.lkey = region->lkey;
    struct ibv_recv_wr wr;
    memset(&wr, 0, sizeof(ibv_recv_wr));
    wr.wr_id = u_int64_t(region);
    wr.sg_list = &list;
    wr.num_sge = 1;
    struct ibv_recv_wr *bad_wr;
    if (_queuePair->context->ops.post_recv(_queuePair, &wr, &bad_wr))
        throw IBError("failed to post recv", errno);
    _postedRecvCount++;
}

void IBVerbs::postSend(ibv_mr *region, size_t size)
{
    if (_completionChannel == NULL || _connectionClosed)
        throw IBError("Conneciton closed");
    struct ibv_sge list;
    memset(&list, 0, sizeof(ibv_sge));
    list.addr = (uintptr_t)region->addr;
    list.length = size;
    list.lkey = region->lkey;
    struct ibv_send_wr wr;
    memset(&wr, 0, sizeof(ibv_send_wr));
    wr.wr_id = u_int64_t(region);
    wr.sg_list = &list;
    wr.num_sge = 1;
    wr.opcode = IBV_WR_SEND_WITH_IMM;
    wr.imm_data = htonl(_postedRecvCount);
    wr.send_flags = IBV_SEND_SIGNALED;

    struct ibv_send_wr *bad_wr;
    if (_queuePair->context->ops.post_send(_queuePair, &wr, &bad_wr))
        throw IBError("Unable to post send buffers", errno);
    _postedRecvCountAck = _postedRecvCount;
    _remoteUsedRecvCount++;
}

/* If we do not have acknowleged all received buffers, the send an ack
 * package */
void IBVerbs::postAck()
{
    if (_completionChannel == NULL || _connectionClosed)
        throw IBError("Conneciton closed");
    if ((char)(_postedRecvCount - _postedRecvCountAck) < (_maxBufferCount - 2))
        return;
    waitForReceiverCanRecv();
    struct ibv_send_wr wr;
    memset(&wr, 0, sizeof(ibv_send_wr));
    wr.opcode = IBV_WR_SEND_WITH_IMM;
    wr.imm_data = htonl(_postedRecvCount);
    wr.send_flags = IBV_SEND_SIGNALED;
    struct ibv_send_wr *bad_wr;
    int err = _queuePair->context->ops.post_send(_queuePair, &wr, &bad_wr);
    if (err)
        throw IBError("Unable to post ack send", err);
    _postedRecvCountAck = _postedRecvCount;
    _remoteUsedRecvCount++;
}

bool IBVerbs::waitCompletion(Time timeout)
{
    if (_completionChannel == NULL || _connectionClosed)
        throw IBError("Conneciton closed");
    if (timeout >= 0)
    {
        struct pollfd pollfd;
        pollfd.fd = _completionChannel->fd;
        pollfd.events = POLLIN;
        pollfd.revents = 0;
        int rc = ::poll(&pollfd, 1, (int)(timeout * 1000));
        if (rc == 0 && errno == EAGAIN)
            return false;
        if (rc != 1)
            throw IBError("Conneciton closed");
    }
    ibv_cq *cq;
    void *context;
    int err=ibv_get_cq_event_p(_completionChannel, &cq, &context);
    if (err)
        throw IBError("Unable get completion events", err);
    _eventsToAck++;
    if (_eventsToAck >= _maxEventsToAck)
    {
        ibv_ack_cq_events_p(cq, _eventsToAck);
        _eventsToAck = 0;
    }
    err = cq->context->ops.req_notify_cq(cq, 0);
    if (err)
        throw IBError("Couldn't request CQ notification");
    return pollCompletion();
}

bool IBVerbs::pollCompletion()
{
    if (_completionChannel == NULL || _connectionClosed)
        throw IBError("Conneciton closed");
    ibv_wc wc[4];
    int nc;
    bool doneSomething = false;
    while (
        (nc = _completionQueue->context->ops.poll_cq(_completionQueue, 4, wc)))
    {
        doneSomething = true;
        for (int n = 0; n < nc; ++n)
        {
            ibv_mr *region = (ibv_mr *)wc[n].wr_id;
            size_t byteLen = wc[n].byte_len;

            if (wc[n].status != IBV_WC_SUCCESS)
                throw IBError("work completion no success",wc[n].status);
            if (wc[n].opcode == IBV_WC_SEND)
            {
                if (region)
                    _freeSendRegions.push(region);
            }
            else if (wc[n].opcode == IBV_WC_RECV)
            {
                _remotePostedRecvCount = ntohl(wc[n].imm_data);
                if (byteLen == 0)
                {
                    postRecv(region);
                }
                else
                {
                    LoadedData data;
                    data.buffer = (u_int8_t *)(region->addr);
                    data.length = byteLen;
                    data.region = region;
                    _loadedRecvRegions.push_back(data);
                }
            }
            else
            {
                throw IBError("unexpected completion opcode", wc[n].opcode);
            }
        }
    }
    return doneSomething;
}

ibv_mr *IBVerbs::createRegion()
{
    void *buffer = memalign(getpagesize(), _maxBufferSize);
    if (!buffer)
        throw IBError("unable to allocate memory", errno);
    ibv_mr *region =
        ibv_reg_mr_p(_protectionDomain, buffer, _maxBufferSize,
                     IBV_ACCESS_LOCAL_WRITE | IBV_ACCESS_REMOTE_WRITE |
                         IBV_ACCESS_REMOTE_READ);
    if (!region)
    {
        free(buffer);
        throw IBError("unable to register memory region", errno);
    }
    return region;
}

void IBVerbs::destroyRegion(ibv_mr *region)
{
    if (region == NULL)
        return;
    void *buffer = region->addr;
    ibv_dereg_mr_p(region);
    free(buffer);
}

void IBVerbs::waitForSendRegion()
{
    while(_freeSendRegions.empty())
        waitCompletion();
}

void IBVerbs::waitForReceiverCanRecv()
{
    while ((char)(_remotePostedRecvCount - _remoteUsedRecvCount) == 0)
        waitCompletion();
}

void IBVerbs::waitForRecvedData()
{
    while (!_loadedRecvRegions.size())
        waitCompletion();
}

void IBVerbs::checkConnection(void *arg)
{
    IBVerbs *ib = (IBVerbs*)arg;
    if (ib == NULL)
        return;
    char buffer = ' ';
    try
    {
        while(!ib->_socket.waitReadable(1) && !ib->_connectionClosed);
        ib->_socket.send(&buffer,1);
    }
    catch(std::exception &ex)
    {
    }
    // force wakeup on wait for queue
    ib->_connectionClosed = true;
    struct ibv_send_wr wr;
    memset(&wr, 0, sizeof(ibv_send_wr));
    wr.opcode = IBV_WR_SEND_WITH_IMM;
    wr.imm_data = htonl(ib->_postedRecvCount);
    wr.send_flags = IBV_SEND_SIGNALED; // | IBV_SEND_INLINE;
    struct ibv_send_wr *bad_wr;
    ib->_queuePair->context->ops.post_send(ib->_queuePair, &wr, &bad_wr);
}

bool IBVerbs::waitReadable(Time timeout)
{
    if (!_loadedRecvRegions.empty())
        return true;
    Time startTime = getSystemTime();
    while (_loadedRecvRegions.empty())
    {
        Time waitTime = -1;
        if (timeout >= 0)
        {
            waitTime = timeout - (getSystemTime() - startTime);
            if (waitTime < 0)
                return false;
        }
        waitCompletion(waitTime);
    }
    return true;
}

void IBVerbs::readSymbols()
{
    if (ibv_get_device_list_p)
        return;
    void *lib = dlopen("libibverbs.so", RTLD_LAZY);
    if (lib == NULL)
        throw IBError("unable to load lib ibverbs");
    ibv_get_device_list_p =
        (ibv_get_device_list_t *)dlsym(lib, "ibv_get_device_list");
    ibv_query_device_p = (ibv_query_device_t *)dlsym(lib, "ibv_query_device");
    ibv_open_device_p = (ibv_open_device_t *)dlsym(lib, "ibv_open_device");
    ibv_close_device_p = (ibv_close_device_t *)dlsym(lib, "ibv_close_device");
    ibv_get_device_name_p =
        (ibv_get_device_name_t *)dlsym(lib, "ibv_get_device_name");
    ibv_create_comp_channel_p =
        (ibv_create_comp_channel_t *)dlsym(lib, "ibv_create_comp_channel");
    ibv_reg_mr_p = (ibv_reg_mr_t *)dlsym(lib, "ibv_reg_mr");
    ibv_dereg_mr_p = (ibv_dereg_mr_t *)dlsym(lib, "ibv_dereg_mr");
    ibv_alloc_pd_p = (ibv_alloc_pd_t *)dlsym(lib, "ibv_alloc_pd");
    ibv_dealloc_pd_p = (ibv_dealloc_pd_t *)dlsym(lib, "ibv_dealloc_pd");
    ibv_create_cq_p = (ibv_create_cq_t *)dlsym(lib, "ibv_create_cq");
    ibv_destroy_cq_p = (ibv_destroy_cq_t *)dlsym(lib, "ibv_destroy_cq");
    ibv_query_port_p = (ibv_query_port_t *)dlsym(lib, "ibv_query_port");
    ibv_get_cq_event_p = (ibv_get_cq_event_t *)dlsym(lib, "ibv_get_cq_event");
    ibv_create_qp_p = (ibv_create_qp_t *)dlsym(lib, "ibv_create_qp");
    ibv_modify_qp_p = (ibv_modify_qp_t *)dlsym(lib, "ibv_modify_qp");
    ibv_destroy_qp_p = (ibv_destroy_qp_t *)dlsym(lib, "ibv_destroy_qp");
    ibv_free_device_list_p =
        (ibv_free_device_list_t *)dlsym(lib, "ibv_free_device_list");
    ibv_destroy_comp_channel_p =
        (ibv_destroy_comp_channel_t *)dlsym(lib, "ibv_destroy_comp_channel");
    ibv_ack_cq_events_p =
        (ibv_ack_cq_events_t *)dlsym(lib, "ibv_ack_cq_events");
}

ibv_query_device_t *IBVerbs::ibv_query_device_p = NULL;
ibv_open_device_t *IBVerbs::ibv_open_device_p = NULL;
ibv_close_device_t *IBVerbs::ibv_close_device_p = NULL;
ibv_get_device_name_t *IBVerbs::ibv_get_device_name_p = NULL;
ibv_get_device_list_t *IBVerbs::ibv_get_device_list_p = NULL;
ibv_create_comp_channel_t *IBVerbs::ibv_create_comp_channel_p = NULL;
ibv_reg_mr_t *IBVerbs::ibv_reg_mr_p = NULL;
ibv_dereg_mr_t *IBVerbs::ibv_dereg_mr_p = NULL;
ibv_alloc_pd_t *IBVerbs::ibv_alloc_pd_p = NULL;
ibv_dealloc_pd_t *IBVerbs::ibv_dealloc_pd_p = NULL;
ibv_create_cq_t *IBVerbs::ibv_create_cq_p = NULL;
ibv_destroy_cq_t *IBVerbs::ibv_destroy_cq_p = NULL;
ibv_query_port_t *IBVerbs::ibv_query_port_p = NULL;
ibv_get_cq_event_t *IBVerbs::ibv_get_cq_event_p = NULL;
ibv_create_qp_t *IBVerbs::ibv_create_qp_p = NULL;
ibv_modify_qp_t *IBVerbs::ibv_modify_qp_p = NULL;
ibv_destroy_qp_t *IBVerbs::ibv_destroy_qp_p = NULL;
ibv_free_device_list_t *IBVerbs::ibv_free_device_list_p = NULL;
ibv_destroy_comp_channel_t *IBVerbs::ibv_destroy_comp_channel_p = NULL;
ibv_ack_cq_events_t *IBVerbs::ibv_ack_cq_events_p = NULL;

#ifdef __sgi
#pragma set woff 1174
#endif

#ifdef OSG_LINUX_ICC
#pragma warning(disable : 177)
#endif

namespace
{
static Char8 cvsid_cpp[] = "@(#)$Id: $";
static Char8 cvsid_hpp[] = OSG_IBVERBS_HEADER_CVSID;
}

#endif
