#ifndef _IBVERBS_H_
#define _IBVERBS_H_
#pragma once

#ifdef __linux

#include <sstream>
#include <string>
#include <queue>
#include <stdexcept>

#include <OSGBase.h>
#include <OSGBaseTypes.h>
#include <OSGPointConnection.h>
#include <OSGPointSockConnection.h>
#include <OSGTime.h>
#include <OSGStreamSocket.h>
#include <OSGPointConnection.h>
#include <OSGBaseThread.h>
#include "OSGIBImplementation.h"

OSG_BEGIN_NAMESPACE

typedef struct ibv_device **ibv_get_device_list_t(int *num_devices);
typedef int ibv_query_device_t(struct ibv_context *context,
                               struct ibv_device_attr *device_attr);
typedef struct ibv_context *ibv_open_device_t(struct ibv_device *device);
typedef int ibv_close_device_t(struct ibv_context *context);
typedef const char *ibv_get_device_name_t(struct ibv_device *device);
typedef struct ibv_comp_channel *
ibv_create_comp_channel_t(struct ibv_context *context);
typedef struct ibv_mr *ibv_reg_mr_t(struct ibv_pd *pd, void *addr,
                                    size_t length, int access);
typedef int ibv_dereg_mr_t(struct ibv_mr *mr);
typedef struct ibv_pd *ibv_alloc_pd_t(struct ibv_context *context);
typedef int ibv_dealloc_pd_t(struct ibv_pd *pd);
typedef struct ibv_cq *ibv_create_cq_t(struct ibv_context *context, int cqe,
                                       void *cq_context,
                                       struct ibv_comp_channel *channel,
                                       int comp_vector);
typedef int ibv_destroy_cq_t(struct ibv_cq *cq);
typedef int ibv_query_port_t(struct ibv_context *context, uint8_t port_num,
                             struct ibv_port_attr *port_attr);
typedef int ibv_get_cq_event_t(struct ibv_comp_channel *channel,
                               struct ibv_cq **cq, void **cq_context);
typedef struct ibv_qp *ibv_create_qp_t(struct ibv_pd *pd,
                                       struct ibv_qp_init_attr *qp_init_attr);
typedef int ibv_modify_qp_t(struct ibv_qp *qp, struct ibv_qp_attr *attr,
                            int attr_mask);
typedef int ibv_destroy_qp_t(struct ibv_qp *qp);
typedef void ibv_free_device_list_t(struct ibv_device **list);
typedef int ibv_destroy_comp_channel_t(struct ibv_comp_channel *channel);
typedef void ibv_ack_cq_events_t(struct ibv_cq *cq, unsigned int nevents);

struct IBError : std::runtime_error
{
  public:
    IBError(const std::string &what) : std::runtime_error(what)
    {
    }
    IBError(const std::string &what, int err)
        : std::runtime_error(static_cast<std::stringstream &>(
              std::stringstream() << what << " errno:" << err).str())
    {
    }
};

class OSG_BASE_DLLMAPPING IBVerbs : public IBImplementation
{
  public:
    IBVerbs();
    virtual ~IBVerbs(void);
    virtual void recv(void *buffer, size_t len);
    virtual void send(const void *buffer, size_t len);
    virtual bool waitReadable(Time timeout=-1);
    virtual int getAvailable(void);
    virtual void connect(StreamSocket socket);
    virtual void accept(StreamSocket socket);
    virtual void disconnect();

  private:
    struct IBAddress
    {
        int lid;
        int qpn;
        int packetSequenceNumber;
        int maxMtu;
    };
    struct LoadedData
    {
        u_int8_t *buffer;
        size_t length;
        ibv_mr *region;
    };
    void open(const std::string &deviceName = "");
    void connectQueuePair();
    void postRecv(ibv_mr *region);
    void postSend(ibv_mr *region, size_t size);
    void postAck();
    void waitForSendRegion();
    void waitForReceiverCanRecv();
    void waitForRecvedData();
    bool pollCompletion();
    bool waitCompletion(Time timeout = -1);
    static void readSymbols();
    static void checkConnection(void *arg);

    BaseThread *_checkThread;
    ibv_mr *createRegion();
    void destroyRegion(ibv_mr *region);

    ibv_device **_deviceList;
    ibv_context *_context;
    ibv_comp_channel *_completionChannel;
    ibv_pd *_protectionDomain;
    ibv_cq *_completionQueue;
    ibv_qp *_queuePair;

    StreamSocket _socket;
    std::vector<ibv_mr *> _sendRegions;
    std::vector<ibv_mr *> _recvRegions;
    std::queue<ibv_mr *> _freeSendRegions;
    std::list<LoadedData> _loadedRecvRegions;
    int _ibPort;
    IBAddress _localIBAddress;
    IBAddress _remoteIBAddress;
    ibv_mr *_currentRecvRegion;

    u_int8_t _remoteUsedRecvCount;
    u_int8_t _remotePostedRecvCount;
    u_int8_t _postedRecvCount;
    u_int8_t _postedRecvCountAck;

    bool _isReading;
    int _eventsToAck;
    bool _connectionClosed;

    size_t _maxBufferSize;
    int _maxBufferCount;
    int _maxEventsToAck;

    static ibv_query_device_t *ibv_query_device_p;
    static ibv_open_device_t *ibv_open_device_p;
    static ibv_close_device_t *ibv_close_device_p;
    static ibv_get_device_name_t *ibv_get_device_name_p;
    static ibv_get_device_list_t *ibv_get_device_list_p;
    static ibv_create_comp_channel_t *ibv_create_comp_channel_p;
    static ibv_reg_mr_t *ibv_reg_mr_p;
    static ibv_dereg_mr_t *ibv_dereg_mr_p;
    static ibv_alloc_pd_t *ibv_alloc_pd_p;
    static ibv_dealloc_pd_t *ibv_dealloc_pd_p;
    static ibv_create_cq_t *ibv_create_cq_p;
    static ibv_destroy_cq_t *ibv_destroy_cq_p;
    static ibv_query_port_t *ibv_query_port_p;
    static ibv_get_cq_event_t *ibv_get_cq_event_p;
    static ibv_create_qp_t *ibv_create_qp_p;
    static ibv_modify_qp_t *ibv_modify_qp_p;
    static ibv_destroy_qp_t *ibv_destroy_qp_p;
    static ibv_free_device_list_t *ibv_free_device_list_p;
    static ibv_destroy_comp_channel_t *ibv_destroy_comp_channel_p;
    static ibv_ack_cq_events_t *ibv_ack_cq_events_p;

    // prohibit default functions (move to 'public' if you need one)
    IBVerbs(const IBVerbs &source);
    IBVerbs &operator=(const IBVerbs &source);
};

typedef IBVerbs *IBVerbsP;

#endif

OSG_END_NAMESPACE

#define OSG_IBVERBS_HEADER_CVSID "@(#)$Id: $"

#endif /* _IBVERBS_H_ */
