#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <numeric>

#include "OSGConfig.h"
#include "OSGLog.h"
#include "OSGBaseFunctions.h"
#include "OSGSocketSelection.h"
#include "OSGPointIBConnection.h"
#include "OSGConnectionHelper.h"
#include "OSGConnectionType.h"

#include "OSGminiz.h"
#include "OSGLZ4.h"
#include "OSGIBImplementation.h"
#ifdef __linux
#include "OSGIBVerbs.h"
#endif

#pragma intrinsic(memcpy)

OSG_USING_NAMESPACE

PointIBConnection::PointIBConnection(StreamSocket acceptSocket)
    : Inherited(acceptSocket),
#ifdef __linux
      _ib(new IBVerbs())
#else
      _ib(NULL)
#endif
{
}

PointIBConnection::~PointIBConnection(void)
{
    if (_ib)
        delete _ib;
    _ib=NULL;
}

const ConnectionType *PointIBConnection::getType(void) const
{
    return &_type;
}

Connection::Channel PointIBConnection::connectPoint(const std::string &address,
                                                    Time timeout)
{
    try
    {
        if (!_ib)
            throw ReadError("Infiniband not supported");
        auto channel = Inherited::connectPoint(address, timeout);
        if (channel < 0)
            return channel;
        _ib->connect(_socket);
        return channel;
    }
    catch (std::exception &ex)
    {
        throw ReadError(ex.what());
    }
}

Connection::Channel PointIBConnection::connectGroup(const std::string &address,
                                                    Time timeout)
{
    try
    {
        if (!_ib)
            throw ReadError("Infiniband not supported");
        auto channel = Inherited::connectGroup(address, timeout);
        if (channel < 0)
            return channel;
        _ib->connect(_socket);
        return channel;
    }
    catch (std::exception &ex)
    {
        throw ReadError(ex.what());
    }
}

Connection::Channel PointIBConnection::acceptPoint(Time timeout)
{
    try
    {
        if (!_ib)
            throw ReadError("Infiniband not supported");
        auto channel = Inherited::acceptPoint(timeout);
        if (channel < 0)
            return channel;
        _ib->accept(_socket);
        return channel;
    }
    catch (std::exception &ex)
    {
        throw ReadError(ex.what());
    }
}

Connection::Channel PointIBConnection::acceptGroup(Time timeout)
{
    try
    {
        if (!_ib)
            throw ReadError("Infiniband not supported");
        auto channel = Inherited::acceptGroup(timeout);
        if (channel < 0)
            return channel;
        _ib->accept(_socket);
        return channel;
    }
    catch (std::exception &ex)
    {
        throw ReadError(ex.what());
    }
}

void PointIBConnection::disconnect(void)
{
    Inherited::disconnect();
    _ib->disconnect();
}

std::string PointIBConnection::bind(const std::string &address)
{
    return Inherited::bind(address);
}

Connection::Channel
PointIBConnection::selectChannel(Time timeout) throw(ReadError)
{
    try
    {
        if (_ib->waitReadable(timeout))
            return 0;
        return -1;
    }
    catch (std::exception &ex)
    {
        throw ReadError(ex.what());
    }
}

int PointIBConnection::getAvailable(void)
{
    try
    {
        return _ib->getAvailable();
    }
    catch (std::exception &ex)
    {
        throw ReadError(ex.what());
    }
}

bool PointIBConnection::wait(Time timeout) throw(ReadError)
{
    try
    {
        UInt32 tag;
        if (selectChannel(timeout) == -1)
            return false;
        getValue(tag);
        if (tag != 314156)
        {
            FFATAL(("Stream out of sync in IBConnection (tag is %u)\n", tag));
            throw ReadError("Stream out of sync");
        }
    }
    catch (std::exception &e)
    {
        throw ReadError(e.what());
    }
    return true;
}

void PointIBConnection::signal(void) throw(WriteError)
{
    UInt32 tag = 314156;
    try
    {
        putValue(tag);
        flush();
    }
    catch (std::exception &e)
    {
        throw WriteError(e.what());
    }
}

PointConnection *PointIBConnection::create(void)
{
    return new PointIBConnection();
}

SizeT PointIBConnection::read(MemoryHandle mem, SizeT size)
{
    try
    {
        _ib->recv(mem, size);
        return size;
    }
    catch (std::exception &ex)
    {
        throw ReadError(ex.what());
    }
}

void PointIBConnection::readBuffer()
{
    UInt32 size;
    read(&_socketReadBuffer[0], sizeof(SocketBufferHeader));
    SocketBufferHeader *header =
        reinterpret_cast<SocketBufferHeader *>(&_socketReadBuffer[0]);
    size = osgntohl(header->size);
    read(&_socketReadBuffer[sizeof(SocketBufferHeader)], size);
    readBufBegin()->setDataSize(size);
}

void PointIBConnection::write(const MemoryHandle mem, SizeT size)
{
    try
    {
        _ib->send(mem, size);
    }
    catch (std::exception &ex)
    {
        throw WriteError(ex.what());
    }
}

void PointIBConnection::writeBuffer(void)
{
    UInt32 size = writeBufBegin()->getDataSize();
    auto header =
        reinterpret_cast<SocketBufferHeader *>(&_socketWriteBuffer[0]);
    header->size = osghtonl(size);
    header->compressed_size = osghtonl(0);
    write(&_socketWriteBuffer[0], size + sizeof(SocketBufferHeader));
}

void PointIBConnection::setCork(bool s)
{
}

void PointIBConnection::setDelay(bool s)
{
    Inherited::setDelay(s);
}

int PointIBConnection::waitAvailable(void)
{
    if(_ib->waitReadable())
        return _ib->getAvailable();

    return 0;
}

ConnectionType PointIBConnection::_type(&PointIBConnection::create,
                                        "Infiniband");

#ifdef __sgi
#pragma set woff 1174
#endif

#ifdef OSG_LINUX_ICC
#pragma warning(disable : 177)
#endif

namespace
{
static Char8 cvsid_cpp[] = "@(#)$Id: $";
static Char8 cvsid_hpp[] = OSG_CONNECTION_HEADER_CVSID;
}
