/*---------------------------------------------------------------------------*\
 *                                OpenSG                                     *
 *                                                                           *
 *                                                                           *
 *                 Copyright (C) 2011 by the OpenSG Forum                    *
 *                                                                           *
 *                            www.opensg.org                                 *
 *                                                                           *
 *   contact: dirk@opensg.org, gerrit.voss@vossg.org, jbehr@zgdv.de          *
 *                                                                           *
\*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*\
 *                                License                                    *
 *                                                                           *
 * This library is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU Library General Public License as published    *
 * by the Free Software Foundation, version 2.                               *
 *                                                                           *
 * This library is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU         *
 * Library General Public License for more details.                          *
 *                                                                           *
 * You should have received a copy of the GNU Library General Public         *
 * License along with this library; if not, write to the Free Software       *
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 *
 *                                                                           *
\*---------------------------------------------------------------------------*/

#include <sstream>

#include "OSGConnectionHelper.h"
#include "OSGLog.h"

OSG_BEGIN_NAMESPACE

// amz enabled socket delay (nagle algorithm) makes not much sense as we already buffer
// our data. But with two computers connected with a 10 GBit ethernet and disabled delay
// sometimes the data distribution gets really slow like only 5% network usage (36 MB/s).
// In the normal case network usage is about 40% (500 MB/s). Now with anabled delay we also
// get slowdowns sometimes but only between 200 MB/s and 500 MB/s.
// ok switched back to disabled delay doesn't work for the display cluster!
// I switch it off in the remote aspect!

namespace
{
    const UInt32 _sockReadBufferSize  = 1024 * 1024;
    const UInt32 _sockWriteBufferSize = 1024 * 1024;

} // namespace

/*! Connect socket to address. Retries until success or timeout
    (negative timeout retries until success).
    Returns true on success and the address of the other endpoint
    in destination.
 */
bool connectSocket(StreamSocket      &socket,
                   const std::string &address,
                   SocketAddress     &destination,
                   Time               timeout     )
{
    Time        startT = getSystemTime();
    std::string host   = "unknown";
    UInt16      port   = 0;

    std::string::size_type pos = address.find(':');

    if(pos != std::string::npos)
    {
        host = address.substr(0, pos);
        port = portFromString(address.substr(pos + 1, std::string::npos));
    }
    else
    {
        host = address;
    }

    socket.open    ();
    socket.setDelay(false);
    socket.setReadBufferSize (_sockReadBufferSize );
    socket.setWriteBufferSize(_sockWriteBufferSize);

    destination = SocketAddress(host, port);

    bool connected = false;

    while(!connected && (timeout < 0 || (getSystemTime() - startT) < timeout))
    {
        try
        {
            socket.connect(destination);
            connected = true;
        }
        catch(const SocketException &)
        {
        }
    }

    return connected;
}

/*! Accept a socket connection on accept, binding it to socket.
    Waits for the connection or until timeout (negative timeout
    waits until connection is accepted).
    Returns true on success and the address of the other endpoint
    in destination.
 */
bool acceptSocket(StreamSocket  &accept,
                  StreamSocket  &socket,
                  SocketAddress &destination,
                  Time           timeout     )
{
    bool retVal = false;

    if(accept.waitReadable(timeout))
    {
        socket = accept.acceptFrom(destination);
        socket.setDelay(false);
        socket.setReadBufferSize (_sockReadBufferSize );
        socket.setWriteBufferSize(_sockWriteBufferSize);

        retVal = true;
    }

    return retVal;
}

UInt16 portFromString(const std::string &str)
{
    UInt16 port = 0;

    std::istringstream iss(str);
    iss >> port;

    if(iss.fail())
    {
        SWARNING << "Could not get port from string '" << str << "'."
                 << std::endl;
        port = 0;
    }

    return port;
}

OSG_END_NAMESPACE
