/*---------------------------------------------------------------------------*\
 *                                OpenSG                                     *
 *                                                                           *
 *                                                                           *
 *             Copyright (C) 2000-2002 by the OpenSG Forum                   *
 *                                                                           *
 *                            www.opensg.org                                 *
 *                                                                           *
 *   contact: dirk@opensg.org, gerrit.voss@vossg.org, jbehr@zgdv.de          *
 *                                                                           *
\*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*\
 *                                License                                    *
 *                                                                           *
 * This library is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU Library General Public License as published    *
 * by the Free Software Foundation, version 2.                               *
 *                                                                           *
 * This library is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU         *
 * Library General Public License for more details.                          *
 *                                                                           *
 * You should have received a copy of the GNU Library General Public         *
 * License along with this library; if not, write to the Free Software       *
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 *
 *                                                                           *
\*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*\
 *                                Changes                                    *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *                                                                           *
\*---------------------------------------------------------------------------*/

//#define MULTI_THREADED
#define THRESHOLD_SUPPORT

//---------------------------------------------------------------------------
//  Includes
//---------------------------------------------------------------------------
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>

#include <set>

#include "OSGConfig.h"
#include "OSGLog.h"
#include "OSGBaseFunctions.h"
#include "OSGBaseThread.h"
#include "OSGSocketSelection.h"
#include "OSGBinaryMessage.h"
#include "OSGGroupSockPipeline.h"
#include "OSGConnectionType.h"
#include "OSGBinaryMessage.h"

#include "OSGminiz.h"
#include "OSGLZ4.h"

#pragma intrinsic(memcpy)

OSG_USING_NAMESPACE

/** \class osg::GroupSockPipeline
 *  \ingroup GrpBaseNetwork
 **/

/*-------------------------------------------------------------------------*/
/*                            constructor destructor                       */

/*! Constructor
 */

GroupSockPipeline::GroupSockPipeline():
    Inherited(),
    _initialized(false)
{
    _next.open();
}

/*! Destructor
 */
GroupSockPipeline::~GroupSockPipeline(void)
{
    // to get out if the busy waiting loop looking for socket data
    // we need to send a end marker.
    UInt32 end = 0x87654321;
    Inherited::write((const MemoryHandle) &end, sizeof(end));
    _next.close();
}

/*! get connection type
 */
const ConnectionType *GroupSockPipeline::getType() const
{
    return &_type;
}

/*-------------------------------------------------------------------------*/
/*                            connection                                   */

/*! connect to the given point. If timeout is reached, -1 is
    returned
 */
GroupConnection::Channel GroupSockPipeline::connectPoint(
    const std::string &address,
    Time               timeout)
{
    Channel channel = Inherited::connectPoint(address, timeout);
    return channel;
}

/*! disconnect the given channel
 */
void GroupSockPipeline::disconnect(Channel channel)
{
    Inherited::disconnect(channel);
}

/*! disconnect the given channels
 */
void GroupSockPipeline::disconnect(const std::vector<Channel> &channels)
{
    Inherited::disconnect(channels);
}

/*! accept an icomming point connection. If timeout is reached,
    -1 is returned. If timeout is -1 then wait without timeout
 */
GroupConnection::Channel GroupSockPipeline::acceptPoint(Time timeout)
{
    Connection::Channel channel = Inherited::acceptPoint(timeout);
    return channel;
}

/*-------------------------- helpers --------------------------------------*/

/** \brief create conneciton
 */

GroupConnection *GroupSockPipeline::create(void)
{
    return new GroupSockPipeline();
}

/*-------------------------------------------------------------------------*/
/*                              read write                                 */

/** Write data to all destinations
 *
 * \param mem   Pointer to data buffer
 * \param size  Size of bytes to write
 *
 **/

void GroupSockPipeline::write(const MemoryHandle mem, SizeT size, bool force_unicast)
{
    if(force_unicast)
    {
        Inherited::write(mem, size);
        return;
    }

    if(!_initialized)
        initialize();

    try
    {
        if(getChannelCount())
            _next.send(mem, size);
    }
    catch(SocketException &e)
    {
        throw WriteError(e.what());
    }
}

typedef struct
{
    std::vector<UInt8> buffer;
    GroupSockPipeline *the;
} ThreadData;

/** Write buffer
 *
 * Write blocksize and data.
 *
 **/
void GroupSockPipeline::writeBuffer(void)
{
    Int32 index;

    if(!_initialized)
        initialize();

    Real64 t = OSG::getSystemTime();

    UInt32 size = writeBufBegin()->getDataSize();

#ifdef THRESHOLD_SUPPORT
    // for small changes we use unicast to reduce latency.
    if(size < _interactive_threshold * 736)
    {
        Inherited::writeBuffer();
        return;
    }
#endif

    if(size)
    {
        // write size to header
        reinterpret_cast<SocketBufferHeader *>(&_socketWriteBuffer[0])->magic = 0x12345678;
        reinterpret_cast<SocketBufferHeader *>(&_socketWriteBuffer[0])->id = ++_id;
        reinterpret_cast<SocketBufferHeader *>(&_socketWriteBuffer[0])->size = osghtonl(size);
        reinterpret_cast<SocketBufferHeader *>(&_socketWriteBuffer[0])->compressed_size = osghtonl(0);

        UInt8 *data = &_socketWriteBuffer[0];

        _write_buffer_bytes += size;

        Real64 t = OSG::getSystemTime();

        if(_compression == Connection::CCOMPRESSION_BEST && size > 1)
        {
            static std::vector<UInt8> compressed;
            compressed.resize(mz_compressBound(size) + sizeof(SocketBufferHeader));
            mz_ulong compressed_size = compressed.size() - sizeof(SocketBufferHeader);
            int r = mz_compress2((unsigned char *) &compressed[sizeof(SocketBufferHeader)], &compressed_size, (const unsigned char *) &_socketWriteBuffer[sizeof(SocketBufferHeader)], size, 1);
            if(r == 0 && compressed_size != 0 && compressed_size < size)
            {
                reinterpret_cast<SocketBufferHeader *>(&compressed[0])->magic = 0x12345678;
                reinterpret_cast<SocketBufferHeader *>(&compressed[0])->id = reinterpret_cast<SocketBufferHeader *>(&_socketWriteBuffer[0])->id;
                reinterpret_cast<SocketBufferHeader *>(&compressed[0])->size = osghtonl(size);
                size = compressed_size;
                reinterpret_cast<SocketBufferHeader *>(&compressed[0])->compressed_size = osghtonl(size);
                data = &compressed[0];
            }
            else
            {
                //printf("GroupSockConnection::writeBuffer: send uncompressed %d!\n", size);
            }
        }
        else if(_compression == Connection::CCOMPRESSION_FAST && size > 1)
        {
            static std::vector<UInt8> compressed;
            // ok we allocate size - 1, lzf_compress returns 0 when the compression failed,
            // in this case the compressed data is greater than the input data.
            //compressed.resize(size - 1 + sizeof(SocketBufferHeader));
            compressed.resize(LZ4_compressBound(size) + sizeof(SocketBufferHeader));
            int compressed_size = LZ4_compress((const char *) &_socketWriteBuffer[sizeof(SocketBufferHeader)], (char *) &compressed[sizeof(SocketBufferHeader)], size);

            if(compressed_size != 0 && compressed_size < size)
            {
                //printf("PointSockConnection::writeBuffer: compressed %d ==> %d\n", size, compressed_size);
                reinterpret_cast<SocketBufferHeader *>(&compressed[0])->magic = 0x12345678;
                reinterpret_cast<SocketBufferHeader *>(&compressed[0])->id = reinterpret_cast<SocketBufferHeader *>(&_socketWriteBuffer[0])->id;
                reinterpret_cast<SocketBufferHeader *>(&compressed[0])->size = osghtonl(size);
                size = compressed_size;
                reinterpret_cast<SocketBufferHeader *>(&compressed[0])->compressed_size = osghtonl(size);
                data = &compressed[0];
            }
            else
            {
                //printf("GroupSockConnection::writeBuffer: send uncompressed %d!\n", size);
            }
        }
        else
        {
            //printf("GroupSockConnection::writeBuffer: send uncompressed %d!\n", size);
        }

#ifdef MULTI_THREADED
        _tp.wait();

        static ThreadData *td = new ThreadData;
        td->buffer.resize(size + sizeof(SocketBufferHeader));
        memcpy(&td->buffer[0], data, size + sizeof(SocketBufferHeader));
        td->the = this;
        _tp.run(writeThread, static_cast<void *>(td));
#else
        _next.send(data, size + sizeof(SocketBufferHeader));
#endif

        _write_buffer_cbytes += size;
        _write_buffer_time += (OSG::getSystemTime() - t);
    }
}

/*! write thread
 */
void GroupSockPipeline::writeThread(void *arg)
{
    ThreadData *td = (ThreadData *) arg;
    td->the->_next.send(&td->buffer[0], td->buffer.size());
    // we use one thread only so one td instance is enough!
    //delete td;
}

/*-------------------------------------------------------------------------*/
/*                              private helpers                            */

/*! initialize pipeline
 */
void GroupSockPipeline::initialize(void)
{
    UInt32         index, len;
    UInt32         nextPort;
    std::string    nextHost;
    BinaryMessage  message;

    for(index = 0 ; index < _sockets.size() ; ++index)
    {
        len = _sockets[index].recv(message);
        if(len == 0)
            throw ReadError("Channel closed\n");
        nextHost = message.getString();
        nextPort = message.getUInt32();

        message.clear();
        if(index == 0)
        {
            // flag to detect last machine.
            message.putUInt32(true);
            // send last machine the ip from the first machine.
            _sockets[_sockets.size() - 1].send(message);
            for(;;)
            {
                try
                {
                    _next.connect(SocketAddress(nextHost.c_str(),
                                                nextPort));
                    break;
                }
                catch(...)
                {
                }
            }
        }
        else
        {
            message.clear();
            message.putUInt32(false);
            message.putString(nextHost);
            message.putUInt32(nextPort);
            // send the previous machine the ip of the current machine.
            _sockets[index - 1].send(message);
        }
    }
    _initialized = true;
}

/*-------------------------------------------------------------------------*/
/*                              static type                                */

ConnectionType GroupSockPipeline::_type(
    &GroupSockPipeline::create,
    "SockPipeline");

/*-------------------------------------------------------------------------*/
/*                              cvs id's                                   */

#ifdef __sgi
#pragma set woff 1174
#endif

#ifdef OSG_LINUX_ICC
#pragma warning( disable : 177 )
#endif

namespace
{
    static Char8 cvsid_cpp       [] = "@(#)$Id: $";
    static Char8 cvsid_hpp       [] = OSG_GROUPSOCKPIPELINE_HEADER_CVSID;
}

