/*---------------------------------------------------------------------------*\
*                                OpenSG                                     *
*                                                                           *
*                                                                           *
*             Copyright (C) 2000-2002 by the OpenSG Forum                   *
*                                                                           *
*                            www.opensg.org                                 *
*                                                                           *
*   contact: dirk@opensg.org, gerrit.voss@vossg.org, jbehr@zgdv.de          *
*                                                                           *
\*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*\
*                                License                                    *
*                                                                           *
*                                                                           *
*                                                                           *
* This library is free software; you can redistribute it and/or modify it   *
* under the terms of the GNU Library General Public License as published    *
* by the Free Software Foundation, version 2.                               *
*                                                                           *
* This library is distributed in the hope that it will be useful, but       *
* WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU         *
* Library General Public License for more details.                          *
*                                                                           *
* You should have received a copy of the GNU Library General Public         *
* License along with this library; if not, write to the Free Software       *
* Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 *
*                                                                           *
*                                                                           *
\*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*\
*                                Changes                                    *
*                                                                           *
*                                                                           *
*                                                                           *
*                                                                           *
*                                                                           *
*                                                                           *
\*---------------------------------------------------------------------------*/

//#define OSG_USE_VAO

//---------------------------------------------------------------------------
//  Includes
//---------------------------------------------------------------------------

#include <stdlib.h>
#include <stdio.h>
#include <stddef.h>
#include <iostream>

#include "OSGConfig.h"

#include <OSGGL.h>
#include <OSGGLEXT.h>

#include "OSGLog.h"

#include "OSGFieldContainerPtr.h"
#include "OSGGeometry.h"
#include "OSGDrawAction.h"

#include "OSGMaterial.h"

#include "OSGGeoPumpFactory.h"
#include "OSGGeoFunctions.h"



OSG_USING_NAMESPACE


    /***************************************************************************\
    *                            Description                                  *
    \***************************************************************************/

    /* \class osg::GeoPumpFactory
    \ingroup GrpSystemNodeCoresDrawablesGeometry

    The PumpFactory is responsible for selecting the most appropriate pump
    function to send the geometry's data to OpenGL in the most efficient manner.

    The pump is selected inside the rendering function and should not be accessed
    from the outside.

    \dev The pump factory chooses the pump based on the used properties. For
    non- and single-indexed geometry OpenGL VertexArrays are used, for
    multi-indexed geometry an immediate mode pump has to be used. Specialized Pumps
    for all combinations of properties are created to optimizie this case as much
    as possible.
    */

    /***************************************************************************\
    *                           Class variables                               *
    \***************************************************************************/

    /*! The Singleton instance of the GeoPumpFactory.
    */
    GeoPumpFactory *GeoPumpFactory::_the = NULL;

/*! An InitFuncWrapper to initialize the GeoPumpFactory.
*/
InitFuncWrapper
    GeoPumpFactory::_glextInitFuncWrapper(GeoPumpFactory::glextInitFunction);

/*! OpenGL extension indices.
*/
UInt32 GeoPumpFactory::_extSecondaryColor = Window::invalidExtensionID;
UInt32 GeoPumpFactory::_extMultitexture = Window::invalidExtensionID;
UInt32 GeoPumpFactory::_extCompiledVertexArray = Window::invalidExtensionID;
UInt32 GeoPumpFactory::_extDrawRangeElements = Window::invalidExtensionID;
UInt32 GeoPumpFactory::_arbVBO = Window::invalidExtensionID;
UInt32 GeoPumpFactory::_arbVAO = Window::invalidExtensionID;

#ifdef OSG_NV_BINDLESS_SUPPORT
UInt32 GeoPumpFactory::_nvVertexBufferUnifiedMemory = Window::invalidExtensionID;
UInt32 GeoPumpFactory::_nvShaderBufferLoad = Window::invalidExtensionID;
#endif

/*! OpenGL extension function indices.
*/
UInt32 GeoPumpFactory::_funcglSecondaryColorPointer = Window::invalidFunctionID;
UInt32 GeoPumpFactory::_funcglClientActiveTextureARB = Window::invalidFunctionID;
UInt32 GeoPumpFactory::_funcglLockArraysEXT = Window::invalidFunctionID;
UInt32 GeoPumpFactory::_funcglUnlockArraysEXT = Window::invalidFunctionID;
UInt32 GeoPumpFactory::_funcglDrawRangeElementsEXT = Window::invalidFunctionID;

UInt32 GeoPumpFactory::_funcglGenBuffers = Window::invalidFunctionID;
UInt32 GeoPumpFactory::_funcglDeleteBuffers = Window::invalidFunctionID;
UInt32 GeoPumpFactory::_funcglBindBuffer = Window::invalidFunctionID;
UInt32 GeoPumpFactory::_funcglBufferData = Window::invalidFunctionID;
UInt32 GeoPumpFactory::_funcglBufferSubData = Window::invalidFunctionID;

UInt32 GeoPumpFactory::_funcglGenBuffersARB = Window::invalidFunctionID;
UInt32 GeoPumpFactory::_funcglDeleteBuffersARB = Window::invalidFunctionID;
UInt32 GeoPumpFactory::_funcglBindBufferARB = Window::invalidFunctionID;
UInt32 GeoPumpFactory::_funcglBufferDataARB = Window::invalidFunctionID;
UInt32 GeoPumpFactory::_funcglBufferSubDataARB = Window::invalidFunctionID;

UInt32 GeoPumpFactory::_funcglGenVertexArrays = Window::invalidFunctionID;
UInt32 GeoPumpFactory::_funcglBindVertexArray = Window::invalidFunctionID;
UInt32 GeoPumpFactory::_funcglDeleteVertexArrays = Window::invalidFunctionID;

#ifdef OSG_NV_BINDLESS_SUPPORT
// bindless VBO stuff
UInt32 GeoPumpFactory::_funcglMakeBufferResidentNV = Window::invalidFunctionID;
UInt32 GeoPumpFactory::_funcglBufferAddressRangeNV = Window::invalidFunctionID;
UInt32 GeoPumpFactory::_funcglVertexFormatNV = Window::invalidFunctionID;
UInt32 GeoPumpFactory::_funcglNormalFormatNV = Window::invalidFunctionID;
UInt32 GeoPumpFactory::_funcglColorFormatNV = Window::invalidFunctionID;
UInt32 GeoPumpFactory::_funcglSecondaryColorFormatNV = Window::invalidFunctionID;
UInt32 GeoPumpFactory::_funcglTexCoordFormatNV = Window::invalidFunctionID;
UInt32 GeoPumpFactory::_funcglIndexFormatNV = Window::invalidFunctionID;
UInt32 GeoPumpFactory::_funcglGetBufferParameterui64vNV = Window::invalidFunctionID;
#endif

/***************************************************************************\
*                           Instance methods                              *
\***************************************************************************/

/*------------- constructors & destructors --------------------------------*/

GeoPumpFactory::GeoPumpFactory(void)
{
}

GeoPumpFactory::~GeoPumpFactory(void)
{
}

/*-------------------------- your_category---------------------------------*/

/*! Access the index of the given Geometry. Needed to access the Pump.

\dev The index is organised as follows:
-  there are 2^7-1+3 different functions
-  the first (0) is for invalid (empty) geometry
-  the next 127 define all multi-indexed functions
-  the last 2 define non-indexed and single-indexed geometries
\enddev
*/
GeoPumpFactory::Index GeoPumpFactory::getIndex(Geometry * geo)
{
    if (geo->getPositions() == NullFC ||
        !geo->getPositions()->getData() ) return 0; //INVALID

    if (geo->getTypes() == NullFC || !geo->getTypes()->getData() ||
        geo->getTypes()->getSize() == 0) return 0; //INVALID

    if (geo->getIndices() == NullFC) return 128; //NON_INDEXED
    if (geo->getMFIndexMapping()->size() < 2) return 129; //SINGLE_INDEXED

    if ((geo->getIndices() != NullFC &&
        geo->getIndices()->getFormatSize() != 4) ||
        (geo->getLengths() != NullFC &&
        geo->getLengths()->getFormatSize() != 4)
        )
        return 130; // Needd to use the master pump

    UInt32 uiIndexMask = 0;

    for(UInt32 i = 0; i < geo->getMFIndexMapping()->size(); ++i)
    {
        uiIndexMask |= geo->getIndexMapping(i);
    }

    // ok the multi index table supports only 4 texcoord units!
    if((uiIndexMask & Geometry::MapTexCoords4 &&
        geo->getTexCoords4() != NullFC         &&
        geo->getTexCoords4()->getData()          ) ||
        (uiIndexMask & Geometry::MapTexCoords5 &&
        geo->getTexCoords5() != NullFC         &&
        geo->getTexCoords5()->getData()          ) ||
        (uiIndexMask & Geometry::MapTexCoords6 &&
        geo->getTexCoords6() != NullFC         &&
        geo->getTexCoords6()->getData()          ) ||
        (uiIndexMask & Geometry::MapTexCoords7 &&
        geo->getTexCoords7() != NullFC         &&
        geo->getTexCoords7()->getData()          ))
    {
        return 130; // Needd to use the master pump
    }

    int a[7];

    if(uiIndexMask & Geometry::MapColor &&
        geo->getColors() != NullFC       &&
        geo->getColors()->getData()        )
    {
        a[0]=1;
    }
    else
    {
        a[0] = 0;
    }

    if(uiIndexMask & Geometry::MapSecondaryColor &&
        geo->getSecondaryColors() != NullFC       &&
        geo->getSecondaryColors()->getData()       )
    {
        a[1]=1;
    }
    else
    {
        a[1]=0;
    }

    if(uiIndexMask & Geometry::MapNormal &&
        geo->getNormals() != NullFC       &&
        geo->getNormals()->getData()        )
    {
        a[2]=1;
    }
    else
    {
        a[2]=0;
    }

    if(uiIndexMask & Geometry::MapTexCoords &&
        geo->getTexCoords() != NullFC        &&
        geo->getTexCoords()->getData()        )
    {
        a[3]=1;
    }
    else
    {
        a[3]=0;
    }

    if(uiIndexMask & Geometry::MapTexCoords1 &&
        geo->getTexCoords1() != NullFC        &&
        geo->getTexCoords1()->getData()        )
    {
        a[4]=1;
    }
    else
    {
        a[4]=0;
    }

    if(uiIndexMask & Geometry::MapTexCoords2 &&
        geo->getTexCoords2() != NullFC        &&
        geo->getTexCoords2()->getData()        )
    {
        a[5]=1;
    }
    else
    {
        a[5]=0;
    }

    if(uiIndexMask & Geometry::MapTexCoords3 &&
        geo->getTexCoords3() != NullFC        &&
        geo->getTexCoords3()->getData()        )
    {
        a[6]=1;
    }
    else
    {
        a[6]=0;
    }

    int index=0;

    for (int i=0; i<7; i++)
        if (a[i]) index=index|(1<<i);

    /* if the geometry was detected as multi-indexed,but
    there was no index data, then index will be 0, which
    corresponds to invalid geometry.
    */

    return index;
}


/*! Get the full Geometry pump for the given \a index.
*/
GeoPumpFactory::GeoPump GeoPumpFactory::getGeoPump(
    Window * ,
    GeoPumpFactory::Index index )
{
    //FWARNING(("GeoPump%d used\n",index));
    return GeoPumps[index]; //crashed windows while testing, watch it.
    //return &masterGeoPump;
}

GeoPumpFactory::PartialGeoPump GeoPumpFactory::getPartialGeoPump(
    Window * ,
    GeoPumpFactory::Index  )
{
    FWARNING(("GeoPumpFactory::getPartialGeoPump: not implemented yet!\n"));
    return NULL;
}

GeoPumpFactory::InterfacePump GeoPumpFactory::getInterfacePump(
    Window * ,
    GeoPumpFactory::Index  )
{
    FWARNING(("GeoPumpFactory::getInterfacePump: not implemented yet!\n"));
    return NULL;
}

GeoPumpFactory::PartialInterfacePump GeoPumpFactory::getPartialInterfacePump(
    Window * ,
    GeoPumpFactory::Index  )
{
    FWARNING(("GeoPumpFactory::getPartialInterfacePump: not "
        "implemented yet!\n"));
    return NULL;
}

/*-------------------------------------------------------------------------*\
-  protected                                                              -
\*-------------------------------------------------------------------------*/


GeoPumpFactory::Index GeoPumpFactory::numIndices( void )
{
    return 131;
}

/*-------------------------------------------------------------------------*\
-  private                                                                -
\*-------------------------------------------------------------------------*/

// Pumping function definitions

typedef void (OSG_APIENTRY *pumpFunc)(GLubyte * data);
typedef void (OSG_APIENTRY *multiPumpFunc)(GLenum which, GLubyte * data);


// Some helper arrays for function selection
// indexed by data type and dimension

static const int formatBase = GL_BYTE;
static const int numFormats = GL_DOUBLE - GL_BYTE + 1;

static const char *formatNames[] =
{   "GL_BYTE", "GL_UNSIGNED_BYTE", "GL_SHORT", "GL_UNSIGNED_SHORT",
"GL_INT", "GL_UNSIGNED_INT", "GL_FLOAT", "GL_2_BYTES",
"GL_3_BYTES", "GL_4_BYTES", "GL_DOUBLE"
};

// little helper class for function init

#if !defined(OSG_DO_DOC) || defined(OSG_DOC_DEV)

class glextFuncInit
{
public:

    glextFuncInit(const char *name, UInt32 format, UInt32 dim) :
      _name(name), _format(format), _dim(dim) {};

      void init(UInt32 (&extids)[numFormats][4], UInt32 extension)
      {
          extids[_format - formatBase][_dim - 1] =
              Window::registerFunction(_name, extension);
      }

private:

    const char   *_name;
    UInt32  _format;
    UInt32  _dim;
};

static glextFuncInit secondaryColorInitFuncs[8] = {
    glextFuncInit(OSG_DLSYM_UNDERSCORE"glSecondaryColor3bvEXT",
    GL_BYTE,
    3),
    glextFuncInit(OSG_DLSYM_UNDERSCORE"glSecondaryColor3ubvEXT",
    GL_UNSIGNED_BYTE,
    3),
    glextFuncInit(OSG_DLSYM_UNDERSCORE"glSecondaryColor3svEXT",
    GL_SHORT,
    3),
    glextFuncInit(OSG_DLSYM_UNDERSCORE"glSecondaryColor3usvEXT",
    GL_UNSIGNED_SHORT,
    3),
    glextFuncInit(OSG_DLSYM_UNDERSCORE"glSecondaryColor3ivEXT",
    GL_INT,
    3),
    glextFuncInit(OSG_DLSYM_UNDERSCORE"glSecondaryColor3uivEXT",
    GL_UNSIGNED_INT,
    3),
    glextFuncInit(OSG_DLSYM_UNDERSCORE"glSecondaryColor3fvEXT",
    GL_FLOAT,
    3),
    glextFuncInit(OSG_DLSYM_UNDERSCORE"glSecondaryColor3dvEXT",
    GL_DOUBLE,
    3)
};

static glextFuncInit multiTexCoordsInitFuncs[16] = {
    glextFuncInit(OSG_DLSYM_UNDERSCORE"glMultiTexCoord1svARB",
    GL_SHORT,
    1),
    glextFuncInit(OSG_DLSYM_UNDERSCORE"glMultiTexCoord2svARB",
    GL_SHORT,
    2),
    glextFuncInit(OSG_DLSYM_UNDERSCORE"glMultiTexCoord3svARB",
    GL_SHORT,
    3),
    glextFuncInit(OSG_DLSYM_UNDERSCORE"glMultiTexCoord4svARB",
    GL_SHORT,
    4),
    glextFuncInit(OSG_DLSYM_UNDERSCORE"glMultiTexCoord1ivARB",
    GL_INT,
    1),
    glextFuncInit(OSG_DLSYM_UNDERSCORE"glMultiTexCoord2ivARB",
    GL_INT,
    2),
    glextFuncInit(OSG_DLSYM_UNDERSCORE"glMultiTexCoord3ivARB",
    GL_INT,
    3),
    glextFuncInit(OSG_DLSYM_UNDERSCORE"glMultiTexCoord4ivARB",
    GL_INT,
    4),
    glextFuncInit(OSG_DLSYM_UNDERSCORE"glMultiTexCoord1fvARB",
    GL_FLOAT,
    1),
    glextFuncInit(OSG_DLSYM_UNDERSCORE"glMultiTexCoord2fvARB",
    GL_FLOAT,
    2),
    glextFuncInit(OSG_DLSYM_UNDERSCORE"glMultiTexCoord3fvARB",
    GL_FLOAT,
    3),
    glextFuncInit(OSG_DLSYM_UNDERSCORE"glMultiTexCoord4fvARB",
    GL_FLOAT,
    4),
    glextFuncInit(OSG_DLSYM_UNDERSCORE"glMultiTexCoord1dvARB",
    GL_DOUBLE,
    1),
    glextFuncInit(OSG_DLSYM_UNDERSCORE"glMultiTexCoord2dvARB",
    GL_DOUBLE,
    2),
    glextFuncInit(OSG_DLSYM_UNDERSCORE"glMultiTexCoord3dvARB",
    GL_DOUBLE,
    3),
    glextFuncInit(OSG_DLSYM_UNDERSCORE"glMultiTexCoord4dvARB",
    GL_DOUBLE,
    4)
};

// The real pumping functions

static pumpFunc PositionFuncs[numFormats][4] = {
    { NULL, NULL, NULL, NULL },                           // GL_BYTE
    { NULL, NULL, NULL, NULL },                           // GL_UNSIGNED_BYTE
    { NULL,
    reinterpret_cast<pumpFunc>(glVertex2sv),
    reinterpret_cast<pumpFunc>(glVertex3sv),
    reinterpret_cast<pumpFunc>(glVertex4sv) },     // GL_SHORT
    { NULL, NULL, NULL, NULL },                           // GL_UNSIGNED_SHORT
    { NULL,
    reinterpret_cast<pumpFunc>(glVertex2iv),
    reinterpret_cast<pumpFunc>(glVertex3iv),
    reinterpret_cast<pumpFunc>(glVertex4iv) },     // GL_INT
    { NULL, NULL, NULL, NULL },                           // GL_UNSIGNED_INT
    { NULL,
    reinterpret_cast<pumpFunc>(glVertex2fv),
    reinterpret_cast<pumpFunc>(glVertex3fv),
    reinterpret_cast<pumpFunc>(glVertex4fv) },     // GL_FLOAT
    { NULL, NULL, NULL, NULL },                           // GL_2_BYTES
    { NULL, NULL, NULL, NULL },                           // GL_3_BYTES
    { NULL, NULL, NULL, NULL },                           // GL_4_BYTES
    { NULL,
    reinterpret_cast<pumpFunc>(glVertex2dv),
    reinterpret_cast<pumpFunc>(glVertex3dv),
    reinterpret_cast<pumpFunc>(glVertex4dv) },     // GL_DOUBLE
};

static pumpFunc NormalFuncs[numFormats][4] = {
    { NULL,
    NULL,
    reinterpret_cast<pumpFunc>(glNormal3sv),
    NULL },          // GL_BYTE
    { NULL, NULL, NULL, NULL },                           // GL_UNSIGNED_BYTE
    { NULL,
    NULL,
    reinterpret_cast<pumpFunc>(glNormal3sv),
    NULL },          // GL_SHORT
    { NULL, NULL, NULL, NULL },                           // GL_UNSIGNED_SHORT
    { NULL,
    NULL,
    reinterpret_cast<pumpFunc>(glNormal3iv),
    NULL },          // GL_INT
    { NULL, NULL, NULL, NULL },                           // GL_UNSIGNED_INT
    { NULL,
    NULL,
    reinterpret_cast<pumpFunc>(glNormal3fv),
    NULL },          // GL_FLOAT
    { NULL, NULL, NULL, NULL },                           // GL_2_BYTES
    { NULL, NULL, NULL, NULL },                           // GL_3_BYTES
    { NULL, NULL, NULL, NULL },                           // GL_4_BYTES
    { NULL,
    NULL,
    reinterpret_cast<pumpFunc>(glNormal3dv), NULL },          // GL_DOUBLE
};

static pumpFunc ColorFuncs[numFormats][4] = {
    { NULL,
    NULL,
    reinterpret_cast<pumpFunc>(glColor3bv),
    reinterpret_cast<pumpFunc>(glColor4bv) },       // GL_BYTE
    { NULL,
    NULL,
    reinterpret_cast<pumpFunc>(glColor3ubv),
    reinterpret_cast<pumpFunc>(glColor4ubv) },     // GL_UNSIGNED_BYTE
    { NULL,
    NULL,
    reinterpret_cast<pumpFunc>(glColor3sv),
    reinterpret_cast<pumpFunc>(glColor4sv) },       // GL_SHORT
    { NULL,
    NULL,
    reinterpret_cast<pumpFunc>(glColor3usv),
    reinterpret_cast<pumpFunc>(glColor4usv) },     // GL_UNSIGNED_SHORT
    { NULL,
    NULL,
    reinterpret_cast<pumpFunc>(glColor3iv),
    reinterpret_cast<pumpFunc>(glColor4iv) },       // GL_INT
    { NULL,
    NULL,
    reinterpret_cast<pumpFunc>(glColor3uiv),
    reinterpret_cast<pumpFunc>(glColor4uiv) },     // GL_UNSIGNED_INT
    { NULL,
    NULL,
    reinterpret_cast<pumpFunc>(glColor3fv),
    reinterpret_cast<pumpFunc>(glColor4fv) },       // GL_FLOAT
    { NULL, NULL, NULL, NULL },                           // GL_2_BYTES
    { NULL, NULL, NULL, NULL },                           // GL_3_BYTES
    { NULL, NULL, NULL, NULL },                           // GL_4_BYTES
    { NULL,
    NULL,
    reinterpret_cast<pumpFunc>(glColor3dv),
    reinterpret_cast<pumpFunc>(glColor4dv) },       // GL_DOUBLE
};

static UInt32 SecColorIDs[numFormats][4];

static pumpFunc TexCoordsFuncs[numFormats][4] = {
    { NULL, NULL, NULL, NULL },                           // GL_BYTE
    { NULL, NULL, NULL, NULL },                           // GL_UNSIGNED_BYTE
    { reinterpret_cast<pumpFunc>(glTexCoord1sv),
    reinterpret_cast<pumpFunc>(glTexCoord2sv),
    reinterpret_cast<pumpFunc>(glTexCoord3sv),
    reinterpret_cast<pumpFunc>(glTexCoord4sv) }, // GL_SHORT
    { NULL, NULL, NULL, NULL },                           // GL_UNSIGNED_SHORT
    { reinterpret_cast<pumpFunc>(glTexCoord1iv),
    reinterpret_cast<pumpFunc>(glTexCoord2iv),
    reinterpret_cast<pumpFunc>(glTexCoord3iv),
    reinterpret_cast<pumpFunc>(glTexCoord4iv) }, // GL_INT
    { NULL, NULL, NULL, NULL },                           // GL_UNSIGNED_INT
    { reinterpret_cast<pumpFunc>(glTexCoord1fv),
    reinterpret_cast<pumpFunc>(glTexCoord2fv),
    reinterpret_cast<pumpFunc>(glTexCoord3fv),
    reinterpret_cast<pumpFunc>(glTexCoord4fv) }, // GL_FLOAT
    { NULL, NULL, NULL, NULL },                           // GL_2_BYTES
    { NULL, NULL, NULL, NULL },                           // GL_3_BYTES
    { NULL, NULL, NULL, NULL },                           // GL_4_BYTES
    { reinterpret_cast<pumpFunc>(glTexCoord1dv),
    reinterpret_cast<pumpFunc>(glTexCoord2dv),
    reinterpret_cast<pumpFunc>(glTexCoord3dv),
    reinterpret_cast<pumpFunc>(glTexCoord4dv) }, // GL_DOUBLE
};

static UInt32 TexCoords1IDs[numFormats][4];

#define TexCoords2IDs TexCoords1IDs
#define TexCoords3IDs TexCoords1IDs
#define TexCoords4IDs TexCoords1IDs
#define TexCoords5IDs TexCoords1IDs
#define TexCoords6IDs TexCoords1IDs
#define TexCoords7IDs TexCoords1IDs

#endif      // remove from all but dev docs

/*! MasterPump for empty geometry, does nothing
*/

static
    void GeoPump0(Window   *OSG_CHECK_ARG(win),
    Geometry *OSG_CHECK_ARG(geo))
{
    return;
}

#define pumpSetup( name, typename, getmethod )                              \
    typename name##Ptr;                                                     \
    GLubyte * name##Data = NULL;                                            \
    UInt32 name##Stride = 0;                                                \
    \
    name##Ptr = geo->getmethod();                                           \
    if ( name##Ptr != NullFC )                                              \
{                                                                       \
    name##Data = name##Ptr->getData();                                  \
    if ( ! ( name##Stride = name##Ptr->getStride() ) )                  \
    name##Stride = name##Ptr->getFormatSize() *                     \
    name##Ptr->getDimension();                                      \
}

/*! MasterPump for non-indexed geometry
*/
static
    void GeoPump128(Window   *win,
    Geometry *geo               )
{
    Int16 modified=0;
    pumpSetup(Positions  , GeoPositionsPtr , getPositions  );
    pumpSetup(Colors     , GeoColorsPtr    , getColors     );
    pumpSetup(SecColors  , GeoColorsPtr    , getSecondaryColors );
    pumpSetup(TexCoords  , GeoTexCoordsPtr , getTexCoords  );
    pumpSetup(TexCoords1 , GeoTexCoordsPtr , getTexCoords1 );
    pumpSetup(TexCoords2 , GeoTexCoordsPtr , getTexCoords2 );
    pumpSetup(TexCoords3 , GeoTexCoordsPtr , getTexCoords3 );
    pumpSetup(TexCoords4 , GeoTexCoordsPtr , getTexCoords4 );
    pumpSetup(TexCoords5 , GeoTexCoordsPtr , getTexCoords5 );
    pumpSetup(TexCoords6 , GeoTexCoordsPtr , getTexCoords6 );
    pumpSetup(TexCoords7 , GeoTexCoordsPtr , getTexCoords7 );
    pumpSetup(Normals    , GeoNormalsPtr   , getNormals    );

    pumpSetup(Lengths    , GeoPLengthsPtr  , getLengths    );
    pumpSetup(Types      , GeoPTypesPtr    , getTypes      );

    if (PositionsData)
    {
        glVertexPointer(PositionsPtr->getDimension(), PositionsPtr->getFormat(),
            PositionsStride, PositionsData);
        glEnableClientState(GL_VERTEX_ARRAY);
        modified|=(1<<0);
    }

    if (ColorsData)
    {
        glColorPointer(ColorsPtr->getDimension(), ColorsPtr->getFormat(),
            ColorsStride, ColorsData);
        glEnableClientState(GL_COLOR_ARRAY);
        modified|=(1<<1);
    }

    if (SecColorsData)
    {
        if (win->hasExtension(GeoPumpFactory::_extSecondaryColor))
        {
            void (OSG_APIENTRY*_glSecondaryColorPointerEXT)(
                GLint size,
                GLenum type,
                GLsizei stride,
                const GLvoid *pointer)=
                reinterpret_cast<void (OSG_APIENTRY*)(
                GLint size,
                GLenum type,
                GLsizei stride,
                const GLvoid *pointer)>(
                win->getFunction(
                GeoPumpFactory::_funcglSecondaryColorPointer));

            _glSecondaryColorPointerEXT(SecColorsPtr->getDimension(),
                SecColorsPtr->getFormat(),
                SecColorsStride, SecColorsData);
            glEnableClientState(GL_SECONDARY_COLOR_ARRAY_EXT);
            modified|=(1<<2);
        }
        else
            FWARNING(("GeoPump128: Window has no Secondary Color extension\n"));
    }


    if (NormalsData)
    {
        glNormalPointer(NormalsPtr->getFormat(), NormalsStride, NormalsData);
        glEnableClientState(GL_NORMAL_ARRAY);
        modified|=(1<<3);
    }

    if (TexCoordsData)
    {
        if (win->hasExtension(GeoPumpFactory::_extMultitexture))
        {
            void (OSG_APIENTRY*_glClientActiveTextureARB) (GLenum type)=
                reinterpret_cast<void (OSG_APIENTRY*) (GLenum type)>(
                win->getFunction(
                GeoPumpFactory::_funcglClientActiveTextureARB));

            _glClientActiveTextureARB(GL_TEXTURE0_ARB);
            glTexCoordPointer (TexCoordsPtr->getDimension(),
                TexCoordsPtr->getFormat(),
                TexCoordsStride, TexCoordsData);
            glEnableClientState(GL_TEXTURE_COORD_ARRAY);
        }
        else
        {
            glTexCoordPointer (TexCoordsPtr->getDimension(),
                TexCoordsPtr->getFormat(),
                TexCoordsStride, TexCoordsData);
            glEnableClientState(GL_TEXTURE_COORD_ARRAY);
        }
        modified|=(1<<4);
    }

    if (TexCoords1Data)
    {
        if (win->hasExtension(GeoPumpFactory::_extMultitexture))
        {
            void (OSG_APIENTRY*_glClientActiveTextureARB) (GLenum type)=
                reinterpret_cast<void (OSG_APIENTRY*) (GLenum type)>(
                win->getFunction(
                GeoPumpFactory::_funcglClientActiveTextureARB));

            _glClientActiveTextureARB(GL_TEXTURE1_ARB);
            glTexCoordPointer (TexCoords1Ptr->getDimension(),
                TexCoords1Ptr->getFormat(),
                TexCoords1Stride, TexCoords1Data);
            glEnableClientState(GL_TEXTURE_COORD_ARRAY);
            modified|=(1<<5);
        }
        else
            FWARNING(("GeoPump128: Window has no MultitextureARB extension\n"));
    }

    if (TexCoords2Data)
    {
        if (win->hasExtension(GeoPumpFactory::_extMultitexture))
        {
            void (OSG_APIENTRY*_glClientActiveTextureARB) (GLenum type)=
                reinterpret_cast<void (OSG_APIENTRY*) (GLenum type)>(
                win->getFunction(
                GeoPumpFactory::_funcglClientActiveTextureARB));

            _glClientActiveTextureARB(GL_TEXTURE2_ARB);
            glTexCoordPointer (TexCoords2Ptr->getDimension(),
                TexCoords2Ptr->getFormat(),
                TexCoords2Stride, TexCoords2Data);
            glEnableClientState(GL_TEXTURE_COORD_ARRAY);
            modified|=(1<<6);
        }
        else
            FWARNING(("GeoPump128: Window has no MultitextureARB extension\n"));
    }

    if (TexCoords3Data)
    {
        if (win->hasExtension(GeoPumpFactory::_extMultitexture))
        {
            void (OSG_APIENTRY*_glClientActiveTextureARB) (GLenum type)=
                reinterpret_cast<void (OSG_APIENTRY*) (GLenum type)>(
                win->getFunction(
                GeoPumpFactory::_funcglClientActiveTextureARB));

            _glClientActiveTextureARB(GL_TEXTURE3_ARB);
            glTexCoordPointer (TexCoords3Ptr->getDimension(),
                TexCoords3Ptr->getFormat(),
                TexCoords3Stride, TexCoords3Data);
            glEnableClientState(GL_TEXTURE_COORD_ARRAY);
            modified|=(1<<7);
        }
        else
            FWARNING(("GeoPump128: Window has no MultitextureARB extension\n"));
    }

    if (TexCoords4Data)
    {
        if (win->hasExtension(GeoPumpFactory::_extMultitexture))
        {
            void (OSG_APIENTRY*_glClientActiveTextureARB) (GLenum type)=
                reinterpret_cast<void (OSG_APIENTRY*) (GLenum type)>(
                win->getFunction(
                GeoPumpFactory::_funcglClientActiveTextureARB));

            _glClientActiveTextureARB(GL_TEXTURE4_ARB);
            glTexCoordPointer (TexCoords4Ptr->getDimension(),
                TexCoords4Ptr->getFormat(),
                TexCoords4Stride, TexCoords4Data);
            glEnableClientState(GL_TEXTURE_COORD_ARRAY);
            modified|=(1<<8);
        }
        else
            FWARNING(("GeoPump128: Window has no MultitextureARB extension\n"));
    }

    if (TexCoords5Data)
    {
        if (win->hasExtension(GeoPumpFactory::_extMultitexture))
        {
            void (OSG_APIENTRY*_glClientActiveTextureARB) (GLenum type)=
                reinterpret_cast<void (OSG_APIENTRY*) (GLenum type)>(
                win->getFunction(
                GeoPumpFactory::_funcglClientActiveTextureARB));

            _glClientActiveTextureARB(GL_TEXTURE5_ARB);
            glTexCoordPointer (TexCoords5Ptr->getDimension(),
                TexCoords5Ptr->getFormat(),
                TexCoords5Stride, TexCoords5Data);
            glEnableClientState(GL_TEXTURE_COORD_ARRAY);
            modified|=(1<<9);
        }
        else
            FWARNING(("GeoPump128: Window has no MultitextureARB extension\n"));
    }

    if (TexCoords6Data)
    {
        if (win->hasExtension(GeoPumpFactory::_extMultitexture))
        {
            void (OSG_APIENTRY*_glClientActiveTextureARB) (GLenum type)=
                reinterpret_cast<void (OSG_APIENTRY*) (GLenum type)>(
                win->getFunction(
                GeoPumpFactory::_funcglClientActiveTextureARB));

            _glClientActiveTextureARB(GL_TEXTURE6_ARB);
            glTexCoordPointer (TexCoords6Ptr->getDimension(),
                TexCoords6Ptr->getFormat(),
                TexCoords6Stride, TexCoords6Data);
            glEnableClientState(GL_TEXTURE_COORD_ARRAY);
            modified|=(1<<10);
        }
        else
            FWARNING(("GeoPump128: Window has no MultitextureARB extension\n"));
    }

    if (TexCoords7Data)
    {
        if (win->hasExtension(GeoPumpFactory::_extMultitexture))
        {
            void (OSG_APIENTRY*_glClientActiveTextureARB) (GLenum type)=
                reinterpret_cast<void (OSG_APIENTRY*) (GLenum type)>(
                win->getFunction(
                GeoPumpFactory::_funcglClientActiveTextureARB));

            _glClientActiveTextureARB(GL_TEXTURE7_ARB);
            glTexCoordPointer (TexCoords7Ptr->getDimension(),
                TexCoords7Ptr->getFormat(),
                TexCoords7Stride, TexCoords7Data);
            glEnableClientState(GL_TEXTURE_COORD_ARRAY);
            modified|=(1<<11);
        }
        else
            FWARNING(("GeoPump128: Window has no MultitextureARB extension\n"));
    }

    UInt32 lendummy;
    UInt32 LengthsSize;
    UInt32 LengthsFormatSize;

    if(LengthsPtr != NullFC && LengthsData != NULL)
    {
        LengthsSize = LengthsPtr->getSize();
        LengthsFormatSize = LengthsPtr->getFormatSize();
    }
    else
    {
        LengthsSize = 1;
        LengthsFormatSize = 4;
        LengthsData = reinterpret_cast<UChar8*>(&lendummy);
        lendummy = PositionsPtr->getSize();
    }

    UInt32 LengthsInd,TypesInd = 0;
    UInt32 first=0;

    if(LengthsFormatSize == 4)
    {
        for(LengthsInd = 0; LengthsInd < LengthsSize; LengthsInd++)
        {
            UInt32 count = *reinterpret_cast<UInt32*>(LengthsData +
                LengthsInd *
                LengthsStride);
            glDrawArrays(*(TypesData + TypesInd++ * TypesStride),first,count);
            first+=count;
        }
    }
    else if(LengthsFormatSize == 2)
    {
        for(LengthsInd = 0; LengthsInd < LengthsSize; LengthsInd++)
        {
            UInt16 count = *reinterpret_cast<UInt16*>(LengthsData +
                LengthsInd *
                LengthsStride);
            glDrawArrays(*(TypesData + TypesInd++ * TypesStride),first,count);
            first+=count;
        }
    }
    else
    {
        FWARNING(("GeoPump128: unknown Lengths format size\n"));
    }

    if(modified&(1<<0)) glDisableClientState(GL_VERTEX_ARRAY);
    if(modified&(1<<1)) glDisableClientState(GL_COLOR_ARRAY);
    if(modified&(1<<2)) glDisableClientState(GL_SECONDARY_COLOR_ARRAY_EXT);
    if(modified&(1<<3)) glDisableClientState(GL_NORMAL_ARRAY);

    if (win->hasExtension(GeoPumpFactory::_extMultitexture))
    {
        void (OSG_APIENTRY*_glClientActiveTextureARB) (GLenum type)=
            reinterpret_cast<void (OSG_APIENTRY*) (GLenum type)>(
            win->getFunction(
            GeoPumpFactory::_funcglClientActiveTextureARB));

        if(modified&(1<<4))
        {
            _glClientActiveTextureARB(GL_TEXTURE0_ARB);
            glDisableClientState(GL_TEXTURE_COORD_ARRAY);
        }

        if(modified&(1<<5))
        {
            _glClientActiveTextureARB(GL_TEXTURE1_ARB);
            glDisableClientState(GL_TEXTURE_COORD_ARRAY);
        }

        if(modified&(1<<6))
        {
            _glClientActiveTextureARB(GL_TEXTURE2_ARB);
            glDisableClientState(GL_TEXTURE_COORD_ARRAY);
        }

        if(modified&(1<<7))
        {
            _glClientActiveTextureARB(GL_TEXTURE3_ARB);
            glDisableClientState(GL_TEXTURE_COORD_ARRAY);
        }

        if(modified&(1<<8))
        {
            _glClientActiveTextureARB(GL_TEXTURE4_ARB);
            glDisableClientState(GL_TEXTURE_COORD_ARRAY);
        }

        if(modified&(1<<9))
        {
            _glClientActiveTextureARB(GL_TEXTURE5_ARB);
            glDisableClientState(GL_TEXTURE_COORD_ARRAY);
        }

        if(modified&(1<<10))
        {
            _glClientActiveTextureARB(GL_TEXTURE6_ARB);
            glDisableClientState(GL_TEXTURE_COORD_ARRAY);
        }

        if(modified&(1<<11))
        {
            _glClientActiveTextureARB(GL_TEXTURE7_ARB);
            glDisableClientState(GL_TEXTURE_COORD_ARRAY);
        }
    }
    else
    {
        if(modified&(1<<4)) glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    }
}

/*! MasterPump for single-indexed geometry
*/

static
    void GeoPump129(Window   *win,
    Geometry *geo)
{
    Int16 modified=0;
    pumpSetup(Positions  , GeoPositionsPtr , getPositions );
    pumpSetup(Colors     , GeoColorsPtr    , getColors    );
    pumpSetup(SecColors  , GeoColorsPtr    , getSecondaryColors );
    pumpSetup(TexCoords  , GeoTexCoordsPtr , getTexCoords  );
    pumpSetup(TexCoords1 , GeoTexCoordsPtr , getTexCoords1 );
    pumpSetup(TexCoords2 , GeoTexCoordsPtr , getTexCoords2 );
    pumpSetup(TexCoords3 , GeoTexCoordsPtr , getTexCoords3 );
    pumpSetup(TexCoords4 , GeoTexCoordsPtr , getTexCoords4 );
    pumpSetup(TexCoords5 , GeoTexCoordsPtr , getTexCoords5 );
    pumpSetup(TexCoords6 , GeoTexCoordsPtr , getTexCoords6 );
    pumpSetup(TexCoords7 , GeoTexCoordsPtr , getTexCoords7 );
    pumpSetup(Normals    , GeoNormalsPtr   , getNormals    );

    pumpSetup(Lengths    , GeoPLengthsPtr  , getLengths    );
    pumpSetup(Types      , GeoPTypesPtr    , getTypes      );
    pumpSetup(Indices    , GeoIndicesPtr   , getIndices    );

    if (PositionsData)
    {
        glVertexPointer(PositionsPtr->getDimension(), PositionsPtr->getFormat(),
            PositionsStride, PositionsData);
        glEnableClientState(GL_VERTEX_ARRAY);
        modified|=(1<<0);
    }

    if (ColorsData)
    {
        glColorPointer(ColorsPtr->getDimension(), ColorsPtr->getFormat(),
            ColorsStride, ColorsData);
        glEnableClientState(GL_COLOR_ARRAY);
        modified|=(1<<1);
    }

    if (SecColorsData)
    {
        if (win->hasExtension(GeoPumpFactory::_extSecondaryColor))
        {
            void (OSG_APIENTRY*_glSecondaryColorPointerEXT)(
                GLint size,
                GLenum type,
                GLsizei stride,
                const GLvoid *pointer)=
                reinterpret_cast<void (OSG_APIENTRY*)(
                GLint size,
                GLenum type,
                GLsizei stride,
                const GLvoid *pointer)>(
                win->getFunction(
                GeoPumpFactory::_funcglSecondaryColorPointer));

            _glSecondaryColorPointerEXT(SecColorsPtr->getDimension(),
                SecColorsPtr->getFormat(),
                SecColorsStride, SecColorsData);
            glEnableClientState(GL_SECONDARY_COLOR_ARRAY_EXT);
            modified|=(1<<2);
        }
        else
            FWARNING(("GeoPump129: Window has no Secondary Color extension\n"));
    }


    if (NormalsData)
    {
        glNormalPointer(NormalsPtr->getFormat(), NormalsStride, NormalsData);
        glEnableClientState(GL_NORMAL_ARRAY);
        modified|=(1<<3);
    }

    if (TexCoordsData)
    {
        if (win->hasExtension(GeoPumpFactory::_extMultitexture))
        {
            void (OSG_APIENTRY*_glClientActiveTextureARB) (GLenum type)=
                reinterpret_cast<void (OSG_APIENTRY*) (GLenum type)>(
                win->getFunction(
                GeoPumpFactory::_funcglClientActiveTextureARB));

            _glClientActiveTextureARB(GL_TEXTURE0_ARB);
            glTexCoordPointer (TexCoordsPtr->getDimension(),
                TexCoordsPtr->getFormat(),
                TexCoordsStride, TexCoordsData);
            glEnableClientState(GL_TEXTURE_COORD_ARRAY);
        }
        else
        {
            glTexCoordPointer (TexCoordsPtr->getDimension(),
                TexCoordsPtr->getFormat(),
                TexCoordsStride, TexCoordsData);
            glEnableClientState(GL_TEXTURE_COORD_ARRAY);
        }
        modified|=(1<<4);
    }

    if (TexCoords1Data)
    {
        if (win->hasExtension(GeoPumpFactory::_extMultitexture))
        {
            void (OSG_APIENTRY*_glClientActiveTextureARB) (GLenum type)=
                reinterpret_cast<void (OSG_APIENTRY*) (GLenum type)>(
                win->getFunction(
                GeoPumpFactory::_funcglClientActiveTextureARB));

            _glClientActiveTextureARB(GL_TEXTURE1_ARB);
            glTexCoordPointer (TexCoords1Ptr->getDimension(),
                TexCoords1Ptr->getFormat(),
                TexCoords1Stride, TexCoords1Data);
            glEnableClientState(GL_TEXTURE_COORD_ARRAY);
            modified|=(1<<5);
        }
        else
            FWARNING(("GeoPump129: Window has no MultitextureARB extension\n"));
    }

    if (TexCoords2Data)
    {
        if (win->hasExtension(GeoPumpFactory::_extMultitexture))
        {
            void (OSG_APIENTRY*_glClientActiveTextureARB) (GLenum type)=
                reinterpret_cast<void (OSG_APIENTRY*) (GLenum type)>(
                win->getFunction(
                GeoPumpFactory::_funcglClientActiveTextureARB));

            _glClientActiveTextureARB(GL_TEXTURE2_ARB);
            glTexCoordPointer (TexCoords2Ptr->getDimension(),
                TexCoords2Ptr->getFormat(),
                TexCoords2Stride, TexCoords2Data);
            glEnableClientState(GL_TEXTURE_COORD_ARRAY);
            modified|=(1<<6);
        }
        else
            FWARNING(("GeoPump129: Window has no MultitextureARB extension\n"));
    }

    if (TexCoords3Data)
    {
        if (win->hasExtension(GeoPumpFactory::_extMultitexture))
        {
            void (OSG_APIENTRY*_glClientActiveTextureARB) (GLenum type)=
                reinterpret_cast<void (OSG_APIENTRY*) (GLenum type)>(
                win->getFunction(
                GeoPumpFactory::_funcglClientActiveTextureARB));

            _glClientActiveTextureARB(GL_TEXTURE3_ARB);
            glTexCoordPointer (TexCoords3Ptr->getDimension(),
                TexCoords3Ptr->getFormat(),
                TexCoords3Stride, TexCoords3Data);
            glEnableClientState(GL_TEXTURE_COORD_ARRAY);
            modified|=(1<<7);
        }
        else
            FWARNING(("GeoPump129: Window has no MultitextureARB extension\n"));
    }

    if (TexCoords4Data)
    {
        if (win->hasExtension(GeoPumpFactory::_extMultitexture))
        {
            void (OSG_APIENTRY*_glClientActiveTextureARB) (GLenum type)=
                reinterpret_cast<void (OSG_APIENTRY*) (GLenum type)>(
                win->getFunction(
                GeoPumpFactory::_funcglClientActiveTextureARB));

            _glClientActiveTextureARB(GL_TEXTURE4_ARB);
            glTexCoordPointer (TexCoords4Ptr->getDimension(),
                TexCoords4Ptr->getFormat(),
                TexCoords4Stride, TexCoords4Data);
            glEnableClientState(GL_TEXTURE_COORD_ARRAY);
            modified|=(1<<8);
        }
        else
            FWARNING(("GeoPump129: Window has no MultitextureARB extension\n"));
    }

    if (TexCoords5Data)
    {
        if (win->hasExtension(GeoPumpFactory::_extMultitexture))
        {
            void (OSG_APIENTRY*_glClientActiveTextureARB) (GLenum type)=
                reinterpret_cast<void (OSG_APIENTRY*) (GLenum type)>(
                win->getFunction(
                GeoPumpFactory::_funcglClientActiveTextureARB));

            _glClientActiveTextureARB(GL_TEXTURE5_ARB);
            glTexCoordPointer (TexCoords5Ptr->getDimension(),
                TexCoords5Ptr->getFormat(),
                TexCoords5Stride, TexCoords5Data);
            glEnableClientState(GL_TEXTURE_COORD_ARRAY);
            modified|=(1<<9);
        }
        else
            FWARNING(("GeoPump129: Window has no MultitextureARB extension\n"));
    }

    if (TexCoords6Data)
    {
        if (win->hasExtension(GeoPumpFactory::_extMultitexture))
        {
            void (OSG_APIENTRY*_glClientActiveTextureARB) (GLenum type)=
                reinterpret_cast<void (OSG_APIENTRY*) (GLenum type)>(
                win->getFunction(
                GeoPumpFactory::_funcglClientActiveTextureARB));

            _glClientActiveTextureARB(GL_TEXTURE6_ARB);
            glTexCoordPointer (TexCoords6Ptr->getDimension(),
                TexCoords6Ptr->getFormat(),
                TexCoords6Stride, TexCoords6Data);
            glEnableClientState(GL_TEXTURE_COORD_ARRAY);
            modified|=(1<<10);
        }
        else
            FWARNING(("GeoPump129: Window has no MultitextureARB extension\n"));
    }

    if (TexCoords7Data)
    {
        if (win->hasExtension(GeoPumpFactory::_extMultitexture))
        {
            void (OSG_APIENTRY*_glClientActiveTextureARB) (GLenum type)=
                reinterpret_cast<void (OSG_APIENTRY*) (GLenum type)>(
                win->getFunction(
                GeoPumpFactory::_funcglClientActiveTextureARB));

            _glClientActiveTextureARB(GL_TEXTURE7_ARB);
            glTexCoordPointer (TexCoords7Ptr->getDimension(),
                TexCoords7Ptr->getFormat(),
                TexCoords7Stride, TexCoords7Data);
            glEnableClientState(GL_TEXTURE_COORD_ARRAY);
            modified|=(1<<11);
        }
        else
            FWARNING(("GeoPump129: Window has no MultitextureARB extension\n"));
    }

    UInt32 lendummy;
    UInt32 LengthsSize;
    bool   len16 = false;

    if(LengthsPtr != NullFC && LengthsData != NULL)
    {
        LengthsSize = LengthsPtr->getSize();
        len16 = (LengthsPtr->getFormatSize() == 2);
    }
    else
    {
        LengthsSize = 1;
        LengthsData = reinterpret_cast<UChar8*>(&lendummy);
        lendummy = PositionsPtr->getSize();
    }

    UInt32 LengthsInd = 0, TypesInd = 0, IndicesInd = 0,
        IndicesSize = IndicesPtr->getSize();

    if (win->hasExtension(GeoPumpFactory::_extCompiledVertexArray))
    {
        void (OSG_APIENTRY*_glLockArraysEXT) (GLint first, GLsizei count)=
            reinterpret_cast<void (OSG_APIENTRY*) (GLint first,
            GLsizei count)>(
            win->getFunction(GeoPumpFactory::_funcglLockArraysEXT));

        _glLockArraysEXT(geo->getMinindex(), geo->getMaxindex() + 1);
    }

    void (OSG_APIENTRY *osgGLDrawRangeElementsEXT)(GLenum mode,
        GLuint start,
        GLuint end,
        GLsizei count,
        GLenum type,
        const GLvoid*indices) =
        reinterpret_cast<void (OSG_APIENTRY *)(GLenum mode,
        GLuint start,
        GLuint end,
        GLsizei count,
        GLenum type,
        const GLvoid*indices)>(
        win->getFunction(GeoPumpFactory::_funcglDrawRangeElementsEXT));


    if(len16 == false)
    {
        for(LengthsInd = 0; LengthsInd < LengthsSize; LengthsInd++)
        {
            UInt32 count  = *reinterpret_cast<UInt32*>(LengthsData +
                LengthsInd *
                LengthsStride);
            UInt8 * vind = IndicesData + IndicesStride * IndicesInd;
            IndicesInd += count;

#if 0
            glDrawElements(*(TypesData + TypesInd++ * TypesStride),count,
                IndicesPtr->getFormat(),vind);
#else
            if(win->hasExtension(GeoPumpFactory::_extDrawRangeElements))
            {
                if(count != 0)
                {
                    osgGLDrawRangeElementsEXT(*(TypesData + TypesInd++ * TypesStride),
                        geo->getLowindices(LengthsInd),
                        geo->getHighindices(LengthsInd),
                        count,
                        IndicesPtr->getFormat(),
                        vind);
                }
            }
            else
            {
                // hope this still works
                if(count != 0)
                {
                    glDrawElements(*(TypesData + TypesInd++ * TypesStride),
                        count,
                        IndicesPtr->getFormat(),
                        vind);
                }
            }
#endif
        }
    }
    else if(len16 == true)
    {
        for(LengthsInd = 0; LengthsInd < LengthsSize; LengthsInd++)
        {
            UInt16 count  = *reinterpret_cast<UInt16*>(LengthsData +
                LengthsInd *
                LengthsStride);
            UInt8 * vind = IndicesData + IndicesStride * IndicesInd;
            IndicesInd += count;

#if 0
            glDrawElements(*(TypesData + TypesInd++ * TypesStride),count,
                IndicesPtr->getFormat(),vind);
#else
            if(win->hasExtension(GeoPumpFactory::_extDrawRangeElements))
            {
                if(count != 0)
                {
                    osgGLDrawRangeElementsEXT(*(TypesData + TypesInd++ * TypesStride),
                        geo->getLowindices(LengthsInd),
                        geo->getHighindices(LengthsInd),
                        count,
                        IndicesPtr->getFormat(),
                        vind);
                }
            }
            else
            {
                if(count != 0)
                {
                    // hope this still works
                    glDrawElements(*(TypesData + TypesInd++ * TypesStride),
                        count,
                        IndicesPtr->getFormat(),
                        vind);
                }
            }
#endif
        }
    }
    else
    {
        FWARNING(("GeoPump129: unknown Lengths format size\n"));
    }

    if (win->hasExtension(GeoPumpFactory::_extCompiledVertexArray))
    {
        void (OSG_APIENTRY*_glUnlockArraysEXT) (void)=
            reinterpret_cast<void (OSG_APIENTRY*) (void)>(
            win->getFunction(GeoPumpFactory::_funcglUnlockArraysEXT));

        _glUnlockArraysEXT();
    }

    if(modified&(1<<0)) glDisableClientState(GL_VERTEX_ARRAY);
    if(modified&(1<<1)) glDisableClientState(GL_COLOR_ARRAY);
    if(modified&(1<<2)) glDisableClientState(GL_SECONDARY_COLOR_ARRAY_EXT);
    if(modified&(1<<3)) glDisableClientState(GL_NORMAL_ARRAY);

    if (win->hasExtension(GeoPumpFactory::_extMultitexture))
    {
        void (OSG_APIENTRY*_glClientActiveTextureARB) (GLenum type)=
            reinterpret_cast<void (OSG_APIENTRY*) (GLenum type)>(
            win->getFunction(
            GeoPumpFactory::_funcglClientActiveTextureARB));

        if(modified&(1<<4))
        {
            _glClientActiveTextureARB(GL_TEXTURE0_ARB);
            glDisableClientState(GL_TEXTURE_COORD_ARRAY);
        }

        if(modified&(1<<5))
        {
            _glClientActiveTextureARB(GL_TEXTURE1_ARB);
            glDisableClientState(GL_TEXTURE_COORD_ARRAY);
        }

        if(modified&(1<<6))
        {
            _glClientActiveTextureARB(GL_TEXTURE2_ARB);
            glDisableClientState(GL_TEXTURE_COORD_ARRAY);
        }

        if(modified&(1<<7))
        {
            _glClientActiveTextureARB(GL_TEXTURE3_ARB);
            glDisableClientState(GL_TEXTURE_COORD_ARRAY);
        }

        if(modified&(1<<8))
        {
            _glClientActiveTextureARB(GL_TEXTURE4_ARB);
            glDisableClientState(GL_TEXTURE_COORD_ARRAY);
        }

        if(modified&(1<<9))
        {
            _glClientActiveTextureARB(GL_TEXTURE5_ARB);
            glDisableClientState(GL_TEXTURE_COORD_ARRAY);
        }

        if(modified&(1<<10))
        {
            _glClientActiveTextureARB(GL_TEXTURE6_ARB);
            glDisableClientState(GL_TEXTURE_COORD_ARRAY);
        }

        if(modified&(1<<11))
        {
            _glClientActiveTextureARB(GL_TEXTURE7_ARB);
            glDisableClientState(GL_TEXTURE_COORD_ARRAY);
        }
    }
    else
    {
        if(modified&(1<<4)) glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    }
}


// The master pump. Can render everything, but takes ages to do so.
// A real fallback, when everything else fails.
// Its second function is the example of how things are supposed to work.
// If an optimized pump creates a different result, it's wrong.
// The formal specification is defined in the design document though.

// some little helper macros to make the code more readable

// define and initialize the variables needed to access the data

#define pumpInternalSetup( name, typename, getmethod, mandatory )           \
    GLubyte * name##Data;                                                   \
    UInt32 name##Stride;                                                    \
    UInt32 name##Ind = 0;                                                   \
    typename name##Ptr;                                                     \
    \
    name##Ptr = geo->getmethod();                                           \
    if(mandatory && name##Ptr == NullFC)                                    \
{                                                                       \
    SWARNING << "masterPump: Geometry " << geo << " has no "            \
    << #name << "s!" << std::endl;                             \
    return;                                                             \
}                                                                       \
    else if(name##Ptr != NullFC)                                            \
{                                                                       \
    name##Data = name##Ptr->getData();                                  \
    if(!(name##Stride = name##Ptr->getStride()))                        \
    name##Stride =  name##Ptr->getFormatSize() *                    \
    name##Ptr->getDimension();                      \
}                                                                       \
    else                                                                    \
{                                                                       \
    name##Data = NULL;                                                  \
    name##Stride = 0;                                                   \
}

#define pumpGLSetup( name, typename, getmethod )                            \
    GLubyte * name##Data;                                                   \
    UInt32 name##Stride;                                                    \
    UInt32 name##Ind = 0;                                                   \
    typename name##Ptr;                                                     \
    pumpFunc name##Func;                                                    \
    \
    name##Ptr = geo->getmethod();                                           \
    if(name##Ptr != NullFC)                                                 \
{                                                                       \
    name##Data = name##Ptr->getData();                                  \
    if(!(name##Stride = name##Ptr->getStride()))                        \
    name##Stride = name##Ptr->getFormatSize() *                     \
    name##Ptr->getDimension();                                      \
    if(!(name##Func = name##Funcs[name##Ptr->getFormat() -              \
    formatBase]                     \
    [name##Ptr->getDimension() - 1]        \
    ) )                                                               \
{                                                                   \
    SWARNING << "masterPump: Geometry " << geo << " has illegal "   \
    << #name << "s: " << name##Ptr->getDimension()         \
    << "D " << formatNames[ name##Ptr->getFormat() -       \
    formatBase ]                    \
    << "!" << std::endl;                                   \
    return;                                                         \
}                                                                   \
}                                                                       \
    else                                                                    \
{                                                                       \
    name##Data = NULL;                                                  \
    name##Func = NULL;                                                  \
    name##Stride = 0;                                                   \
}

#define pumpGLExtSetup( name, typename, getmethod )                         \
    GLubyte * name##Data;                                                   \
    UInt32 name##Stride;                                                    \
    UInt32 name##Ind = 0;                                                   \
    typename name##Ptr;                                                     \
    pumpFunc name##Func;                                                    \
    \
    name##Ptr = geo->getmethod();                                           \
    if(name##Ptr != NullFC)                                                 \
{                                                                       \
    name##Data = name##Ptr->getData();                                  \
    if(!(name##Stride = name##Ptr->getStride()))                        \
    name##Stride = name##Ptr->getFormatSize() *                     \
    name##Ptr->getDimension();                                      \
    if(name##IDs[name##Ptr->getFormat() - formatBase]                   \
    [name##Ptr->getDimension() - 1] ==                      \
    Window::invalidFunctionID)                              \
{                                                                   \
    SWARNING << "masterPump: Geometry " << geo << " has illegal "   \
    << #name << "s: " << name##Ptr->getDimension()         \
    << "D " << formatNames[ name##Ptr->getFormat() -       \
    formatBase ]                    \
    << "!" << std::endl;                                   \
    return;                                                         \
}                                                                   \
    if(!(name##Func = reinterpret_cast<pumpFunc>(                       \
    win->getFunction(name##IDs                    \
    [name##Ptr->getFormat() - formatBase]  \
    [name##Ptr->getDimension() - 1]))      \
    ) )                                                               \
{                                                                   \
    SWARNING << "masterPump: Geometry " << geo << " uses "          \
    << #name << "s: " << name##Ptr->getDimension()         \
    << "D " << formatNames[ name##Ptr->getFormat() -       \
    formatBase ]                    \
    << " which are not supported by Window " << win        \
    <<  "!" << std::endl;                                  \
    return;                                                         \
}                                                                   \
}                                                                       \
    else                                                                    \
{                                                                       \
    name##Data = NULL;                                                  \
    name##Func = NULL;                                                  \
    name##Stride = 0;                                                   \
}

#define pumpMultiGLExtSetup(name, typename, getmethod)                      \
    GLubyte * name##Data;                                                   \
    UInt32 name##Stride;                                                    \
    UInt32 name##Ind = 0;                                                   \
    typename name##Ptr;                                                     \
    multiPumpFunc name##Func;                                               \
    \
    name##Ptr = geo->getmethod();                                           \
    if(name##Ptr != NullFC)                                                 \
{                                                                       \
    name##Data = name##Ptr->getData();                                  \
    if(!(name##Stride = name##Ptr->getStride() ) )                      \
    name##Stride = name##Ptr->getFormatSize() *                     \
    name##Ptr->getDimension();                                      \
    if(name##IDs[name##Ptr->getFormat() - formatBase]                   \
    [name##Ptr->getDimension() - 1] ==                      \
    Window::invalidFunctionID)                              \
{                                                                   \
    SWARNING << "masterPump: Geometry " << geo << "has illegal "    \
    << #name << "s: " << name##Ptr->getDimension()         \
    << "D " << formatNames[ name##Ptr->getFormat() -       \
    formatBase ]                    \
    << "!" << std::endl;                                   \
    return;                                                         \
}                                                                   \
    if(!(name##Func = reinterpret_cast<multiPumpFunc>(                  \
    win->getFunction(name##IDs                    \
    [name##Ptr->getFormat() - formatBase]  \
    [name##Ptr->getDimension() - 1]))      \
    ) )                                                               \
{                                                                   \
    SWARNING << "masterPump: Geometry " << geo << "uses "           \
    << #name << "s: " << name##Ptr->getDimension()         \
    << "D " << formatNames[ name##Ptr->getFormat() -       \
    formatBase ]                    \
    << "which are not supported by Window " << win         \
    <<  "!" << std::endl;                                  \
    return;                                                         \
}                                                                   \
}                                                                       \
    else                                                                    \
{                                                                       \
    name##Data = NULL;                                                  \
    name##Func = NULL;                                                  \
    name##Stride = 0;                                                   \
}

#define iC  pumpGLSetup         ( Color     , GeoColorsPtr   , getColors         )\
    Int16 ColorIndex  =  geo->calcMappingIndex( Geometry::MapColor );     \
    if(ColorData && ColorPtr->getSize() == 1)                             \
    ColorFunc(ColorData);

#define iS  pumpGLExtSetup      ( SecColor  , GeoColorsPtr   , getSecondaryColors)\
    Int16 SecColorIndex = geo->calcMappingIndex(Geometry::MapSecondaryColor);

#define iN  pumpGLSetup         ( Normal    , GeoNormalsPtr  , getNormals        )\
    Int16 NormalIndex = geo->calcMappingIndex(Geometry::MapNormal);

#define iT  pumpGLSetup         ( TexCoords , GeoTexCoordsPtr, getTexCoords      )\
    Int16 TexCoordsIndex  = geo->calcMappingIndex(Geometry::MapTexCoords);

#define iT1 pumpMultiGLExtSetup ( TexCoords1, GeoTexCoordsPtr, getTexCoords1     )\
    Int16 TexCoords1Index = geo->calcMappingIndex(Geometry::MapTexCoords1);

#define iT2 pumpMultiGLExtSetup ( TexCoords2, GeoTexCoordsPtr, getTexCoords2     )\
    Int16 TexCoords2Index = geo->calcMappingIndex(Geometry::MapTexCoords2);

#define iT3 pumpMultiGLExtSetup ( TexCoords3, GeoTexCoordsPtr, getTexCoords3     )\
    Int16 TexCoords3Index = geo->calcMappingIndex(Geometry::MapTexCoords3);

#define iT4 pumpMultiGLExtSetup ( TexCoords4, GeoTexCoordsPtr, getTexCoords4     )\
    Int16 TexCoords4Index = geo->calcMappingIndex(Geometry::MapTexCoords4);

#define iT5 pumpMultiGLExtSetup ( TexCoords5, GeoTexCoordsPtr, getTexCoords5     )\
    Int16 TexCoords5Index = geo->calcMappingIndex(Geometry::MapTexCoords5);

#define iT6 pumpMultiGLExtSetup ( TexCoords6, GeoTexCoordsPtr, getTexCoords6     )\
    Int16 TexCoords6Index = geo->calcMappingIndex(Geometry::MapTexCoords6);

#define iT7 pumpMultiGLExtSetup ( TexCoords7, GeoTexCoordsPtr, getTexCoords7     )\
    Int16 TexCoords7Index = geo->calcMappingIndex(Geometry::MapTexCoords7);

#define rC ColorFunc     ( ColorData     + ColorStride     * vind[ColorIndex]     );
#define rS SecColorFunc  ( SecColorData  + SecColorStride  * vind[SecColorIndex]  );
#define rN NormalFunc    ( NormalData    + NormalStride    * vind[NormalIndex]    );
#define rT TexCoordsFunc ( TexCoordsData + TexCoordsStride * vind[TexCoordsIndex] );
#define rT1 TexCoords1Func( GL_TEXTURE1_ARB , TexCoords1Data + TexCoords1Stride * vind[TexCoords1Index] );
#define rT2 TexCoords2Func( GL_TEXTURE2_ARB , TexCoords2Data + TexCoords2Stride * vind[TexCoords2Index] );
#define rT3 TexCoords3Func( GL_TEXTURE3_ARB , TexCoords3Data + TexCoords3Stride * vind[TexCoords3Index] );
#define rT4 TexCoords4Func( GL_TEXTURE4_ARB , TexCoords4Data + TexCoords4Stride * vind[TexCoords4Index] );
#define rT5 TexCoords5Func( GL_TEXTURE5_ARB , TexCoords5Data + TexCoords5Stride * vind[TexCoords5Index] );
#define rT6 TexCoords6Func( GL_TEXTURE6_ARB , TexCoords6Data + TexCoords6Stride * vind[TexCoords6Index] );
#define rT7 TexCoords7Func( GL_TEXTURE7_ARB , TexCoords7Data + TexCoords7Stride * vind[TexCoords7Index] );

/*
nmappings for sure is greater or equal than two because this is multi
index geometry and there is for sure IndexData!!!
*/

#define defMultiGeoPump( func, init , render )                              \
    static                                                                      \
    void GeoPump##func( Window   *win,                                          \
    Geometry *geo)                                   \
{                                                                           \
    /*if (win) ; */                                                         \
    \
    pumpInternalSetup( Type  , GeoPTypesPtr   , getTypes  , true );         \
    pumpInternalSetup( Length, GeoPLengthsPtr , getLengths, false );        \
    pumpInternalSetup( Index , GeoIndicesPtr  , getIndices, false );        \
    \
    pumpGLSetup( Position    , GeoPositionsPtr, getPositions );             \
    Int16 PositionIndex = geo->calcMappingIndex(Geometry::MapPosition);     \
    \
    init;                                                                   \
    UInt16 nmappings = geo->getMFIndexMapping()->size();                    \
    \
    UInt32 lendummy;                                                        \
    UInt32 LengthSize;                                                      \
    \
    if ( ! LengthData )                                                     \
{                                                                       \
    LengthSize = 1;                                                     \
    LengthData = reinterpret_cast<UChar8 *>(&lendummy);                 \
    lendummy = IndexPtr->getSize() / nmappings;                         \
}                                                                       \
    else                                                                    \
{                                                                       \
    LengthSize = LengthPtr->getSize();                                  \
}                                                                       \
    \
    for ( LengthInd = 0; LengthInd < LengthSize; LengthInd++ )              \
{                                                                       \
    glBegin( *(TypeData + TypeInd++ * TypeStride) );                    \
    \
    for ( UInt32 l = *reinterpret_cast<UInt32*>(LengthData +            \
    LengthInd *             \
    LengthStride);          \
    l > 0; l-- )                                           \
{                                                                   \
    UInt32 * vind;                                                  \
    \
    vind =                                                          \
    reinterpret_cast<UInt32*>(IndexData +                       \
    IndexStride *                     \
    IndexInd);                        \
    IndexInd += nmappings;                                          \
    \
    render;                                                         \
    PositionFunc(PositionData+PositionStride*vind[PositionIndex]);  \
}                                                                   \
    glEnd();                                                            \
}                                                                       \
    \
}


#ifdef __sgi
#pragma set woff 1174,1209
#endif

/*
Defining all multi-indexed Geometry pumps (127:))
*/

defMultiGeoPump(   1 , iC;                            , rC;                             )
    defMultiGeoPump(   2 ,     iS;                        ,     rS;                         )
    defMultiGeoPump(   3 , iC; iS;                        , rC; rS;                         )
    defMultiGeoPump(   4 ,         iN;                    ,         rN;                     )
    defMultiGeoPump(   5 , iC;     iN;                    , rC;     rN;                     )
    defMultiGeoPump(   6 ,     iS; iN;                    ,     rS; rN;                     )
    defMultiGeoPump(   7 , iC; iS; iN;                    , rC; rS; rN;                     )
    defMultiGeoPump(   8 ,             iT;                ,             rT;                 )
    defMultiGeoPump(   9 , iC;         iT;                , rC;         rT;                 )
    defMultiGeoPump(  10 ,     iS;     iT;                ,     rS;     rT;                 )
    defMultiGeoPump(  11 , iC; iS;     iT;                , rC; rS;     rT;                 )
    defMultiGeoPump(  12 ,         iN; iT;                ,         rN; rT;                 )
    defMultiGeoPump(  13 , iC;     iN; iT;                , rC;     rN; rT;                 )
    defMultiGeoPump(  14 ,     iS; iN; iT;                ,     rS; rN; rT;                 )
    defMultiGeoPump(  15 , iC; iS; iN; iT;                , rC; rS; rN; rT;                 )
    defMultiGeoPump(  16 ,                 iT1;           ,                 rT1;            )
    defMultiGeoPump(  17 , iC;             iT1;           , rC;             rT1;            )
    defMultiGeoPump(  18 ,     iS;         iT1;           ,     rS;         rT1;            )
    defMultiGeoPump(  19 , iC; iS;         iT1;           , rC; rS;         rT1;            )
    defMultiGeoPump(  20 ,         iN;     iT1;           ,         rN;     rT1;            )
    defMultiGeoPump(  21 , iC;     iN;     iT1;           , rC;     rN;     rT1;            )
    defMultiGeoPump(  22 ,     iS; iN;     iT1;           ,     rS; rN;     rT1;            )
    defMultiGeoPump(  23 , iC; iS; iN;     iT1;           , rC; rS; rN;     rT1;            )
    defMultiGeoPump(  24 ,             iT; iT1;           ,             rT; rT1;            )
    defMultiGeoPump(  25 , iC;         iT; iT1;           , rC;         rT; rT1;            )
    defMultiGeoPump(  26 ,     iS;     iT; iT1;           ,     rS;     rT; rT1;            )
    defMultiGeoPump(  27 , iC; iS;     iT; iT1;           , rC; rS;     rT; rT1;            )
    defMultiGeoPump(  28 ,         iN; iT; iT1;           ,         rN; rT; rT1;            )
    defMultiGeoPump(  29 , iC;     iN; iT; iT1;           , rC;     rN; rT; rT1;            )
    defMultiGeoPump(  30 ,     iS; iN; iT; iT1;           ,     rS; rN; rT; rT1;            )
    defMultiGeoPump(  31 , iC; iS; iN; iT; iT1;           , rC; rS; rN; rT; rT1;            )
    defMultiGeoPump(  32 ,                      iT2;      ,                      rT2;       )
    defMultiGeoPump(  33 , iC;                  iT2;      , rC;                  rT2;       )
    defMultiGeoPump(  34 ,     iS;              iT2;      ,     rS;              rT2;       )
    defMultiGeoPump(  35 , iC; iS;              iT2;      , rC; rS;              rT2;       )
    defMultiGeoPump(  36 ,         iN;          iT2;      ,         rN;          rT2;       )
    defMultiGeoPump(  37 , iC;     iN;          iT2;      , rC;     rN;          rT2;       )
    defMultiGeoPump(  38 ,     iS; iN;          iT2;      ,     rS; rN;          rT2;       )
    defMultiGeoPump(  39 , iC; iS; iN;          iT2;      , rC; rS; rN;          rT2;       )
    defMultiGeoPump(  40 ,             iT;      iT2;      ,             rT;      rT2;       )
    defMultiGeoPump(  41 , iC;         iT;      iT2;      , rC;         rT;      rT2;       )
    defMultiGeoPump(  42 ,     iS;     iT;      iT2;      ,     rS;     rT;      rT2;       )
    defMultiGeoPump(  43 , iC; iS;     iT;      iT2;      , rC; rS;     rT;      rT2;       )
    defMultiGeoPump(  44 ,         iN; iT;      iT2;      ,         rN; rT;      rT2;       )
    defMultiGeoPump(  45 , iC;     iN; iT;      iT2;      , rC;     rN; rT;      rT2;       )
    defMultiGeoPump(  46 ,     iS; iN; iT;      iT2;      ,     rS; rN; rT;      rT2;       )
    defMultiGeoPump(  47 , iC; iS; iN; iT;      iT2;      , rC; rS; rN; rT;      rT2;       )
    defMultiGeoPump(  48 ,                 iT1; iT2;      ,                 rT1; rT2;       )
    defMultiGeoPump(  49 , iC;             iT1; iT2;      , rC;             rT1; rT2;       )
    defMultiGeoPump(  50 ,     iS;         iT1; iT2;      ,     rS;         rT1; rT2;       )
    defMultiGeoPump(  51 , iC; iS;         iT1; iT2;      , rC; rS;         rT1; rT2;       )
    defMultiGeoPump(  52 ,         iN;     iT1; iT2;      ,         rN;     rT1; rT2;       )
    defMultiGeoPump(  53 , iC;     iN;     iT1; iT2;      , rC;     rN;     rT1; rT2;       )
    defMultiGeoPump(  54 ,     iS; iN;     iT1; iT2;      ,     rS; rN;     rT1; rT2;       )
    defMultiGeoPump(  55 , iC; iS; iN;     iT1; iT2;      , rC; rS; rN;     rT1; rT2;       )
    defMultiGeoPump(  56 ,             iT; iT1; iT2;      ,             rT; rT1; rT2;       )
    defMultiGeoPump(  57 , iC;         iT; iT1; iT2;      , rC;         rT; rT1; rT2;       )
    defMultiGeoPump(  58 ,     iS;     iT; iT1; iT2;      ,     rS;     rT; rT1; rT2;       )
    defMultiGeoPump(  59 , iC; iS;     iT; iT1; iT2;      , rC; rS;     rT; rT1; rT2;       )
    defMultiGeoPump(  60 ,         iN; iT; iT1; iT2;      ,         rN; rT; rT1; rT2;       )
    defMultiGeoPump(  61 , iC;     iN; iT; iT1; iT2;      , rC;     rN; rT; rT1; rT2;       )
    defMultiGeoPump(  62 ,     iS; iN; iT; iT1; iT2;      ,     rS; rN; rT; rT1; rT2;       )
    defMultiGeoPump(  63 , iC; iS; iN; iT; iT1; iT2;      , rC; rS; rN; rT; rT1; rT2;       )
    defMultiGeoPump(  64 ,                           iT3; ,                           rT3;  )
    defMultiGeoPump(  65 , iC;                       iT3; , rC;                       rT3;  )
    defMultiGeoPump(  66 ,     iS;                   iT3; ,     rS;                   rT3;  )
    defMultiGeoPump(  67 , iC; iS;                   iT3; , rC; rS;                   rT3;  )
    defMultiGeoPump(  68 ,         iN;               iT3; ,         rN;               rT3;  )
    defMultiGeoPump(  69 , iC;     iN;               iT3; , rC;     rN;               rT3;  )
    defMultiGeoPump(  70 ,     iS; iN;               iT3; ,     rS; rN;               rT3;  )
    defMultiGeoPump(  71 , iC; iS; iN;               iT3; , rC; rS; rN;               rT3;  )
    defMultiGeoPump(  72 ,             iT;           iT3; ,             rT;           rT3;  )
    defMultiGeoPump(  73 , iC;         iT;           iT3; , rC;         rT;           rT3;  )
    defMultiGeoPump(  74 ,     iS;     iT;           iT3; ,     rS;     rT;           rT3;  )
    defMultiGeoPump(  75 , iC; iS;     iT;           iT3; , rC; rS;     rT;           rT3;  )
    defMultiGeoPump(  76 ,         iN; iT;           iT3; ,         rN; rT;           rT3;  )
    defMultiGeoPump(  77 , iC;     iN; iT;           iT3; , rC;     rN; rT;           rT3;  )
    defMultiGeoPump(  78 ,     iS; iN; iT;           iT3; ,     rS; rN; rT;           rT3;  )
    defMultiGeoPump(  79 , iC; iS; iN; iT;           iT3; , rC; rS; rN; rT;           rT3;  )
    defMultiGeoPump(  80 ,                 iT1;      iT3; ,                 rT1;      rT3;  )
    defMultiGeoPump(  81 , iC;             iT1;      iT3; , rC;             rT1;      rT3;  )
    defMultiGeoPump(  82 ,     iS;         iT1;      iT3; ,     rS;         rT1;      rT3;  )
    defMultiGeoPump(  83 , iC; iS;         iT1;      iT3; , rC; rS;         rT1;      rT3;  )
    defMultiGeoPump(  84 ,         iN;     iT1;      iT3; ,         rN;     rT1;      rT3;  )
    defMultiGeoPump(  85 , iC;     iN;     iT1;      iT3; , rC;     rN;     rT1;      rT3;  )
    defMultiGeoPump(  86 ,     iS; iN;     iT1;      iT3; ,     rS; rN;     rT1;      rT3;  )
    defMultiGeoPump(  87 , iC; iS; iN;     iT1;      iT3; , rC; rS; rN;     rT1;      rT3;  )
    defMultiGeoPump(  88 ,             iT; iT1;      iT3; ,             rT; rT1;      rT3;  )
    defMultiGeoPump(  89 , iC;         iT; iT1;      iT3; , rC;         rT; rT1;      rT3;  )
    defMultiGeoPump(  90 ,     iS;     iT; iT1;      iT3; ,     rS;     rT; rT1;      rT3;  )
    defMultiGeoPump(  91 , iC; iS;     iT; iT1;      iT3; , rC; rS;     rT; rT1;      rT3;  )
    defMultiGeoPump(  92 ,         iN; iT; iT1;      iT3; ,         rN; rT; rT1;      rT3;  )
    defMultiGeoPump(  93 , iC;     iN; iT; iT1;      iT3; , rC;     rN; rT; rT1;      rT3;  )
    defMultiGeoPump(  94 ,     iS; iN; iT; iT1;      iT3; ,     rS; rN; rT; rT1;      rT3;  )
    defMultiGeoPump(  95 , iC; iS; iN; iT; iT1;      iT3; , rC; rS; rN; rT; rT1;      rT3;  )
    defMultiGeoPump(  96 ,                      iT2; iT3; ,                      rT2; rT3;  )
    defMultiGeoPump(  97 , iC;                  iT2; iT3; , rC;                  rT2; rT3;  )
    defMultiGeoPump(  98 ,     iS;              iT2; iT3; ,     rS;              rT2; rT3;  )
    defMultiGeoPump(  99 , iC; iS;              iT2; iT3; , rC; rS;              rT2; rT3;  )
    defMultiGeoPump( 100 ,         iN;          iT2; iT3; ,         rN;          rT2; rT3;  )
    defMultiGeoPump( 101 , iC;     iN;          iT2; iT3; , rC;     rN;          rT2; rT3;  )
    defMultiGeoPump( 102 ,     iS; iN;          iT2; iT3; ,     rS; rN;          rT2; rT3;  )
    defMultiGeoPump( 103 , iC; iS; iN;          iT2; iT3; , rC; rS; rN;          rT2; rT3;  )
    defMultiGeoPump( 104 ,             iT;      iT2; iT3; ,             rT;      rT2; rT3;  )
    defMultiGeoPump( 105 , iC;         iT;      iT2; iT3; , rC;         rT;      rT2; rT3;  )
    defMultiGeoPump( 106 ,     iS;     iT;      iT2; iT3; ,     rS;     rT;      rT2; rT3;  )
    defMultiGeoPump( 107 , iC; iS;     iT;      iT2; iT3; , rC; rS;     rT;      rT2; rT3;  )
    defMultiGeoPump( 108 ,         iN; iT;      iT2; iT3; ,         rN; rT;      rT2; rT3;  )
    defMultiGeoPump( 109 , iC;     iN; iT;      iT2; iT3; , rC;     rN; rT;      rT2; rT3;  )
    defMultiGeoPump( 110 ,     iS; iN; iT;      iT2; iT3; ,     rS; rN; rT;      rT2; rT3;  )
    defMultiGeoPump( 111 , iC; iS; iN; iT;      iT2; iT3; , rC; rS; rN; rT;      rT2; rT3;  )
    defMultiGeoPump( 112 ,                 iT1; iT2; iT3; ,                 rT1; rT2; rT3;  )
    defMultiGeoPump( 113 , iC;             iT1; iT2; iT3; , rC;             rT1; rT2; rT3;  )
    defMultiGeoPump( 114 ,     iS;         iT1; iT2; iT3; ,     rS;         rT1; rT2; rT3;  )
    defMultiGeoPump( 115 , iC; iS;         iT1; iT2; iT3; , rC; rS;         rT1; rT2; rT3;  )
    defMultiGeoPump( 116 ,         iN;     iT1; iT2; iT3; ,         rN;     rT1; rT2; rT3;  )
    defMultiGeoPump( 117 , iC;     iN;     iT1; iT2; iT3; , rC;     rN;     rT1; rT2; rT3;  )
    defMultiGeoPump( 118 ,     iS; iN;     iT1; iT2; iT3; ,     rS; rN;     rT1; rT2; rT3;  )
    defMultiGeoPump( 119 , iC; iS; iN;     iT1; iT2; iT3; , rC; rS; rN;     rT1; rT2; rT3;  )
    defMultiGeoPump( 120 ,             iT; iT1; iT2; iT3; ,             rT; rT1; rT2; rT3;  )
    defMultiGeoPump( 121 , iC;         iT; iT1; iT2; iT3; , rC;         rT; rT1; rT2; rT3;  )
    defMultiGeoPump( 122 ,     iS;     iT; iT1; iT2; iT3; ,     rS;     rT; rT1; rT2; rT3;  )
    defMultiGeoPump( 123 , iC; iS;     iT; iT1; iT2; iT3; , rC; rS;     rT; rT1; rT2; rT3;  )
    defMultiGeoPump( 124 ,         iN; iT; iT1; iT2; iT3; ,         rN; rT; rT1; rT2; rT3;  )
    defMultiGeoPump( 125 , iC;     iN; iT; iT1; iT2; iT3; , rC;     rN; rT; rT1; rT2; rT3;  )
    defMultiGeoPump( 126 ,     iS; iN; iT; iT1; iT2; iT3; ,     rS; rN; rT; rT1; rT2; rT3;  )
    defMultiGeoPump( 127 , iC; iS; iN; iT; iT1; iT2; iT3; , rC; rS; rN; rT; rT1; rT2; rT3;  )

    // the master pump function

    GeoPumpFactory::GeoPump GeoPumpFactory::GeoPumps[131]={&GeoPump0,
    &GeoPump1,&GeoPump2,&GeoPump3,&GeoPump4,&GeoPump5,&GeoPump6,&GeoPump7,&GeoPump8,
    &GeoPump9,&GeoPump10,&GeoPump11,&GeoPump12,&GeoPump13,&GeoPump14,&GeoPump15,&GeoPump16,
    &GeoPump17,&GeoPump18,&GeoPump19,&GeoPump20,&GeoPump21,&GeoPump22,&GeoPump23,&GeoPump24,
    &GeoPump25,&GeoPump26,&GeoPump27,&GeoPump28,&GeoPump29,&GeoPump30,&GeoPump31,&GeoPump32,
    &GeoPump33,&GeoPump34,&GeoPump35,&GeoPump36,&GeoPump37,&GeoPump38,&GeoPump39,&GeoPump40,
    &GeoPump41,&GeoPump42,&GeoPump43,&GeoPump44,&GeoPump45,&GeoPump46,&GeoPump47,&GeoPump48,
    &GeoPump49,&GeoPump50,&GeoPump51,&GeoPump52,&GeoPump53,&GeoPump54,&GeoPump55,&GeoPump56,
    &GeoPump57,&GeoPump58,&GeoPump59,&GeoPump60,&GeoPump61,&GeoPump62,&GeoPump63,&GeoPump64,
    &GeoPump65,&GeoPump66,&GeoPump67,&GeoPump68,&GeoPump69,&GeoPump70,&GeoPump71,&GeoPump72,
    &GeoPump73,&GeoPump74,&GeoPump75,&GeoPump76,&GeoPump77,&GeoPump78,&GeoPump79,&GeoPump80,
    &GeoPump81,&GeoPump82,&GeoPump83,&GeoPump84,&GeoPump85,&GeoPump86,&GeoPump87,&GeoPump88,
    &GeoPump89,&GeoPump90,&GeoPump91,&GeoPump92,&GeoPump93,&GeoPump94,&GeoPump95,&GeoPump96,
    &GeoPump97,&GeoPump98,&GeoPump99,&GeoPump100,&GeoPump101,&GeoPump102,&GeoPump103,&GeoPump104,
    &GeoPump105,&GeoPump106,&GeoPump107,&GeoPump108,&GeoPump109,&GeoPump110,&GeoPump111,&GeoPump112,
    &GeoPump113,&GeoPump114,&GeoPump115,&GeoPump116,&GeoPump117,&GeoPump118,&GeoPump119,&GeoPump120,
    &GeoPump121,&GeoPump122,&GeoPump123,&GeoPump124,&GeoPump125,&GeoPump126,&GeoPump127,&GeoPump128,
    &GeoPump129,&GeoPumpFactory::masterGeoPump};

void GeoPumpFactory::masterGeoPump(Window   *win,
    Geometry *geo)
{
    // Setup: get all the data

    pumpInternalSetup( Type, GeoPTypesPtr, getTypes, true );
    pumpInternalSetup( Length, GeoPLengthsPtr, getLengths, false );
    pumpInternalSetup( Index, GeoIndicesPtr, getIndices, false );

    pumpGLSetup        ( Position,   GeoPositionsPtr, getPositions      );
    pumpGLSetup        ( Color,      GeoColorsPtr,    getColors         );
    pumpGLExtSetup     ( SecColor,   GeoColorsPtr,    getSecondaryColors);
    pumpGLSetup        ( Normal,     GeoNormalsPtr,   getNormals        );
    pumpGLSetup        ( TexCoords,  GeoTexCoordsPtr, getTexCoords      );
    pumpMultiGLExtSetup( TexCoords1, GeoTexCoordsPtr, getTexCoords1     );
    pumpMultiGLExtSetup( TexCoords2, GeoTexCoordsPtr, getTexCoords2     );
    pumpMultiGLExtSetup( TexCoords3, GeoTexCoordsPtr, getTexCoords3     );
    pumpMultiGLExtSetup( TexCoords4, GeoTexCoordsPtr, getTexCoords4     );
    pumpMultiGLExtSetup( TexCoords5, GeoTexCoordsPtr, getTexCoords5     );
    pumpMultiGLExtSetup( TexCoords6, GeoTexCoordsPtr, getTexCoords6     );
    pumpMultiGLExtSetup( TexCoords7, GeoTexCoordsPtr, getTexCoords7     );

    // check if the node is empty
    if(! TypeData || TypePtr->getSize() == 0)
        return;

    // if it's not empty we need positions
    if(! PositionData)
    {
        SWARNING << "masterPump: Geometry " << geo << " has no positions!?!"
            << std::endl;
        return;
    }

    // find the mapping indices
    UInt16 nmappings = geo->getMFIndexMapping()->size();
    Int16 PositionIndex   = -1,
        NormalIndex     = -1,
        ColorIndex      = -1,
        SecColorIndex   = -1,
        TexCoordsIndex  = -1,
        TexCoords1Index = -1,
        TexCoords2Index = -1,
        TexCoords3Index = -1,
        TexCoords4Index = -1,
        TexCoords5Index = -1,
        TexCoords6Index = -1,
        TexCoords7Index = -1;

    if(nmappings)
    {
        PositionIndex   = geo->calcMappingIndex(Geometry::MapPosition      );
        NormalIndex     = geo->calcMappingIndex(Geometry::MapNormal        );
        ColorIndex      = geo->calcMappingIndex(Geometry::MapColor         );
        SecColorIndex   = geo->calcMappingIndex(Geometry::MapSecondaryColor);
        TexCoordsIndex  = geo->calcMappingIndex(Geometry::MapTexCoords     );
        TexCoords1Index = geo->calcMappingIndex(Geometry::MapTexCoords1    );
        TexCoords2Index = geo->calcMappingIndex(Geometry::MapTexCoords2    );
        TexCoords3Index = geo->calcMappingIndex(Geometry::MapTexCoords3    );
        TexCoords4Index = geo->calcMappingIndex(Geometry::MapTexCoords4    );
        TexCoords5Index = geo->calcMappingIndex(Geometry::MapTexCoords5    );
        TexCoords6Index = geo->calcMappingIndex(Geometry::MapTexCoords6    );
        TexCoords7Index = geo->calcMappingIndex(Geometry::MapTexCoords7    );

        if(! PositionData)
        {
            SWARNING << "masterPump: Geometry "
                << geo
                << "has no position index!?!"
                << std::endl;
            return;
        }

    }
    else if(IndexData)
    {
        nmappings = 1;
        PositionIndex =
            NormalIndex =
            ColorIndex =
            SecColorIndex =
            TexCoordsIndex =
            TexCoords1Index =
            TexCoords2Index =
            TexCoords3Index =
            TexCoords4Index =
            TexCoords5Index =
            TexCoords6Index =
            TexCoords7Index = 0;
    }

    // overall color?
    if(ColorData && ColorPtr->getSize() == 1)
        ColorFunc(ColorData);

    // Length handling. Special case: no length given

    UInt32 lendummy;
    UInt32 LengthSize;
    bool len16 = false;

    // no lengths? use all available data for the first type
    if(! LengthData)
    {
        LengthSize = 1;
        LengthData = reinterpret_cast<UChar8*>(&lendummy);
        if(IndexData)
        {
            lendummy = IndexPtr->getSize() / nmappings;
        }
        else
        {
            lendummy = PositionPtr->getSize();
        }
    }
    else
    {
        LengthSize = LengthPtr->getSize();
        len16 = (LengthPtr->getFormatSize() == 2);
    }

    bool ind16 = false;
    ind16 = (IndexPtr != NullFC && IndexPtr->getFormatSize() == 2);

    UInt32 l;

    for(LengthInd = 0; LengthInd < LengthSize; LengthInd++)
    {
        glBegin(*(TypeData + TypeInd++ * TypeStride));

        if(len16)
        {
            l = *reinterpret_cast<UInt16*>(LengthData +
                LengthInd * LengthStride);
        }
        else
        {
            l = *reinterpret_cast<UInt32*>(LengthData +
                LengthInd * LengthStride);
        }

        for( ; l > 0; l--)
        {
            if(IndexData && !ind16)
            {
                UInt32 * vind;

                vind = reinterpret_cast<UInt32*>(IndexData +
                    IndexStride * IndexInd);
                IndexInd += nmappings;

                if(ColorData && ColorIndex >= 0)
                {
                    ColorFunc(ColorData + ColorStride * vind[ColorIndex]);
                }

                if(SecColorData && SecColorIndex >= 0)
                {
                    SecColorFunc(SecColorData + SecColorStride *
                        vind[SecColorIndex]);
                }

                if(NormalData && NormalIndex >= 0)
                {
                    NormalFunc(NormalData + NormalStride * vind[NormalIndex]);
                }

                if(TexCoordsData && TexCoordsIndex >= 0)
                {
                    TexCoordsFunc(TexCoordsData + TexCoordsStride *
                        vind[TexCoordsIndex]);
                }

                if(TexCoords1Data && TexCoords1Index >= 0)
                {
                    TexCoords1Func(GL_TEXTURE1_ARB,
                        TexCoords1Data + TexCoords1Stride *
                        vind[TexCoords1Index]);
                }

                if(TexCoords2Data && TexCoords2Index >= 0)
                {
                    TexCoords2Func(GL_TEXTURE2_ARB,
                        TexCoords2Data + TexCoords2Stride *
                        vind[TexCoords2Index]);
                }

                if(TexCoords3Data && TexCoords3Index >= 0)
                {
                    TexCoords3Func(GL_TEXTURE3_ARB,
                        TexCoords3Data + TexCoords3Stride *
                        vind[TexCoords3Index]);
                }

                if(TexCoords4Data && TexCoords4Index >= 0)
                {
                    TexCoords4Func(GL_TEXTURE4_ARB,
                        TexCoords4Data + TexCoords4Stride *
                        vind[TexCoords4Index]);
                }

                if(TexCoords5Data && TexCoords5Index >= 0)
                {
                    TexCoords5Func(GL_TEXTURE5_ARB,
                        TexCoords5Data + TexCoords5Stride *
                        vind[TexCoords5Index]);
                }

                if(TexCoords6Data && TexCoords6Index >= 0)
                {
                    TexCoords6Func(GL_TEXTURE6_ARB,
                        TexCoords6Data + TexCoords6Stride *
                        vind[TexCoords6Index]);
                }

                if(TexCoords7Data && TexCoords7Index >= 0)
                {
                    TexCoords7Func(GL_TEXTURE7_ARB,
                        TexCoords7Data + TexCoords7Stride *
                        vind[TexCoords7Index]);
                }

                PositionFunc(PositionData + PositionStride * vind[PositionIndex]);
            }
            else if(IndexData && ind16)
            {
                UInt16 * vind;

                vind = reinterpret_cast<UInt16*>(IndexData +
                    IndexStride * IndexInd);
                IndexInd += nmappings;

                if(ColorData && ColorIndex >= 0)
                {
                    ColorFunc(ColorData + ColorStride * vind[ColorIndex]);
                }

                if(SecColorData && SecColorIndex >= 0)
                {
                    SecColorFunc(SecColorData + SecColorStride *
                        vind[SecColorIndex]);
                }

                if(NormalData && NormalIndex >= 0)
                {
                    NormalFunc(NormalData + NormalStride * vind[NormalIndex]);
                }

                if(TexCoordsData && TexCoordsIndex >= 0)
                {
                    TexCoordsFunc(TexCoordsData + TexCoordsStride *
                        vind[TexCoordsIndex]);
                }

                if(TexCoords1Data && TexCoords1Index >= 0)
                {
                    TexCoords1Func(GL_TEXTURE1_ARB,
                        TexCoords1Data + TexCoords1Stride *
                        vind[TexCoords1Index]);
                }

                if(TexCoords2Data && TexCoords2Index >= 0)
                {
                    TexCoords2Func(GL_TEXTURE2_ARB,
                        TexCoords2Data + TexCoords2Stride *
                        vind[TexCoords2Index]);
                }

                if(TexCoords3Data && TexCoords3Index >= 0)
                {
                    TexCoords3Func(GL_TEXTURE3_ARB,
                        TexCoords3Data + TexCoords3Stride *
                        vind[TexCoords3Index]);
                }

                if(TexCoords4Data && TexCoords4Index >= 0)
                {
                    TexCoords4Func(GL_TEXTURE4_ARB,
                        TexCoords4Data + TexCoords4Stride *
                        vind[TexCoords4Index]);
                }

                if(TexCoords5Data && TexCoords5Index >= 0)
                {
                    TexCoords5Func(GL_TEXTURE5_ARB,
                        TexCoords5Data + TexCoords5Stride *
                        vind[TexCoords5Index]);
                }

                if(TexCoords6Data && TexCoords6Index >= 0)
                {
                    TexCoords6Func(GL_TEXTURE6_ARB,
                        TexCoords6Data + TexCoords6Stride *
                        vind[TexCoords6Index]);
                }

                if(TexCoords7Data && TexCoords7Index >= 0)
                {
                    TexCoords7Func(GL_TEXTURE7_ARB,
                        TexCoords7Data + TexCoords7Stride *
                        vind[TexCoords7Index]);
                }

                PositionFunc(PositionData + PositionStride * vind[PositionIndex]);
            }
            else
            {
                if(ColorData)
                {
                    ColorFunc(ColorData + ColorStride * PositionInd);
                }

                if(SecColorData)
                {
                    SecColorFunc(SecColorData + SecColorStride * PositionInd);
                }

                if(NormalData)
                {
                    NormalFunc(NormalData + NormalStride * PositionInd);
                }

                if(TexCoordsData)
                {
                    TexCoordsFunc(TexCoordsData + TexCoordsStride *
                        PositionInd);
                }

                if(TexCoords1Data)
                {
                    TexCoords1Func(GL_TEXTURE1_ARB,
                        TexCoords1Data + TexCoords1Stride *
                        PositionInd);
                }

                if(TexCoords2Data)
                {
                    TexCoords2Func(GL_TEXTURE2_ARB,
                        TexCoords2Data + TexCoords2Stride *
                        PositionInd);
                }

                if(TexCoords3Data)
                {
                    TexCoords3Func(GL_TEXTURE3_ARB,
                        TexCoords3Data + TexCoords3Stride *
                        PositionInd);
                }

                if(TexCoords4Data)
                {
                    TexCoords4Func(GL_TEXTURE4_ARB,
                        TexCoords4Data + TexCoords4Stride *
                        PositionInd);
                }

                if(TexCoords5Data)
                {
                    TexCoords5Func(GL_TEXTURE5_ARB,
                        TexCoords5Data + TexCoords5Stride *
                        PositionInd);
                }

                if(TexCoords6Data)
                {
                    TexCoords6Func(GL_TEXTURE6_ARB,
                        TexCoords6Data + TexCoords6Stride *
                        PositionInd);
                }

                if(TexCoords7Data)
                {
                    TexCoords7Func(GL_TEXTURE7_ARB,
                        TexCoords7Data + TexCoords7Stride *
                        PositionInd);
                }

                PositionFunc(PositionData + PositionStride * PositionInd);

                PositionInd++;
            }
        }

        glEnd();
    }
}


void GeoPumpFactory::masterPartialGeoPump(Window   *OSG_CHECK_ARG(win     ),
    Geometry *OSG_CHECK_ARG(geo      ),
    UInt32    OSG_CHECK_ARG(primtype ),
    UInt32    OSG_CHECK_ARG(firstvert),
    UInt32    OSG_CHECK_ARG(nvert    ))
{
    FWARNING(("GeoPumpFactory::masterPartialGeoPump: not implemented yet!\n"));
}

void GeoPumpFactory::masterInterfacePump(
    Window                *OSG_CHECK_ARG(win       ),
    GeoPositionsInterface *OSG_CHECK_ARG(pos       ),
    GeoNormalsInterface   *OSG_CHECK_ARG(norm      ),
    GeoColorsInterface    *OSG_CHECK_ARG(col       ),
    GeoColorsInterface    *OSG_CHECK_ARG(seccol    ),
    GeoTexCoordsInterface *OSG_CHECK_ARG(texcoords ),
    GeoTexCoordsInterface *OSG_CHECK_ARG(texcoords1),
    GeoTexCoordsInterface *OSG_CHECK_ARG(texcoords2),
    GeoTexCoordsInterface *OSG_CHECK_ARG(texcoords3),
    GeoPTypesInterface    *OSG_CHECK_ARG(type      ),
    GeoPLengthsInterface  *OSG_CHECK_ARG(len       ),
    GeoIndicesInterface   *OSG_CHECK_ARG(ind       ),
    UInt16                *OSG_CHECK_ARG(pMap      ),
    UInt16                 OSG_CHECK_ARG(nmap      ))
{
    FWARNING(("GeoPumpFactory::masterInterfacePump: not implemented yet!\n"));
}

void GeoPumpFactory::masterPartialInterfacePump(
    Window                *OSG_CHECK_ARG(win       ),
    GeoPositionsInterface *OSG_CHECK_ARG(pos       ),
    GeoNormalsInterface   *OSG_CHECK_ARG(norm      ),
    GeoColorsInterface    *OSG_CHECK_ARG(col       ),
    GeoColorsInterface    *OSG_CHECK_ARG(seccol    ),
    GeoTexCoordsInterface *OSG_CHECK_ARG(texcoords ),
    GeoTexCoordsInterface *OSG_CHECK_ARG(texcoords1),
    GeoTexCoordsInterface *OSG_CHECK_ARG(texcoords2),
    GeoTexCoordsInterface *OSG_CHECK_ARG(texcoords3),
    GeoPTypesInterface    *OSG_CHECK_ARG(type      ),
    GeoPLengthsInterface  *OSG_CHECK_ARG(len       ),
    GeoIndicesInterface   *OSG_CHECK_ARG(ind       ),
    UInt16                *OSG_CHECK_ARG(pMap      ),
    UInt16                 OSG_CHECK_ARG(nmap      ),
    UInt32                 OSG_CHECK_ARG(primtype  ),
    UInt32                 OSG_CHECK_ARG(firstvert ),
    UInt32                 OSG_CHECK_ARG(nvert     ))
{
    FWARNING(
        ("GeoPumpFactory::masterPartialInterfacePump not implemented yet!\n"));
}

#ifdef __sgi
#pragma reset woff 1174,1209
#endif

bool GeoPumpFactory::glextInitFunction(void)
{
    _extSecondaryColor      =
        Window::registerExtension("GL_EXT_secondary_color");
    _extMultitexture        =
        Window::registerExtension("GL_ARB_multitexture");
    _extCompiledVertexArray =
        Window::registerExtension("GL_EXT_compiled_vertex_array");
    _extDrawRangeElements   =
        Window::registerExtension("GL_EXT_draw_range_elements");
    _arbVBO =
        Window::registerExtension("GL_ARB_vertex_buffer_object");
    _arbVAO =
        Window::registerExtension("GL_ARB_vertex_array_object");


    UInt16 i,j;
    for(i = 0; i < numFormats; i++)
    {
        for(j = 0; j < 4; j++)
        {
            SecColorIDs  [i][j] = Window::invalidFunctionID;
            TexCoords1IDs[i][j] = Window::invalidFunctionID;
        }
    }

    for(i = 0; i < 8; i++)
        secondaryColorInitFuncs[i].init(SecColorIDs, _extSecondaryColor);

    for(i = 0; i < 16; i++)
        multiTexCoordsInitFuncs[i].init(TexCoords1IDs, _extMultitexture);

    _funcglSecondaryColorPointer  = Window::registerFunction(
        OSG_DLSYM_UNDERSCORE"glSecondaryColorPointerEXT",
        _extSecondaryColor);
    _funcglClientActiveTextureARB = Window::registerFunction(
        OSG_DLSYM_UNDERSCORE"glClientActiveTextureARB",
        _extMultitexture);
    _funcglLockArraysEXT          = Window::registerFunction(
        OSG_DLSYM_UNDERSCORE"glLockArraysEXT",
        _extCompiledVertexArray);
    _funcglUnlockArraysEXT        = Window::registerFunction(
        OSG_DLSYM_UNDERSCORE"glUnlockArraysEXT",
        _extCompiledVertexArray);
    _funcglDrawRangeElementsEXT   = Window::registerFunction(
        OSG_DLSYM_UNDERSCORE"glDrawRangeElementsEXT",
        _extDrawRangeElements);

    _funcglGenBuffers           = Window::registerFunction(
        OSG_DLSYM_UNDERSCORE"glGenBuffers",
        _arbVBO);
    _funcglDeleteBuffers         = Window::registerFunction(
        OSG_DLSYM_UNDERSCORE"glDeleteBuffers",
        _arbVBO);
    _funcglBindBuffer           = Window::registerFunction(
        OSG_DLSYM_UNDERSCORE"glBindBuffer",
        _arbVBO);
    _funcglBufferData           = Window::registerFunction(
        OSG_DLSYM_UNDERSCORE"glBufferData",
        _arbVBO);
    _funcglBufferSubData         = Window::registerFunction(
        OSG_DLSYM_UNDERSCORE"glBufferSubData",
        _arbVBO);

    _funcglGenBuffersARB        = Window::registerFunction(
        OSG_DLSYM_UNDERSCORE"glGenBuffersARB",
        _arbVBO);
    _funcglDeleteBuffersARB     = Window::registerFunction(
        OSG_DLSYM_UNDERSCORE"glDeleteBuffersARB",
        _arbVBO);
    _funcglBindBufferARB        = Window::registerFunction(
        OSG_DLSYM_UNDERSCORE"glBindBufferARB",
        _arbVBO);
    _funcglBufferDataARB        = Window::registerFunction(
        OSG_DLSYM_UNDERSCORE"glBufferDataARB",
        _arbVBO);
    _funcglBufferSubDataARB     = Window::registerFunction(
        OSG_DLSYM_UNDERSCORE"glBufferSubDataARB",
        _arbVBO);

    _funcglGenVertexArrays      = Window::registerFunction(
        OSG_DLSYM_UNDERSCORE"glGenVertexArrays",
        _arbVAO);
    _funcglBindVertexArray      = Window::registerFunction(
        OSG_DLSYM_UNDERSCORE"glBindVertexArray",
        _arbVAO);
    _funcglDeleteVertexArrays    = Window::registerFunction(
        OSG_DLSYM_UNDERSCORE"glDeleteVertexArrays",
        _arbVAO);

    // bindless stuff

#ifdef OSG_NV_BINDLESS_SUPPORT
    // disabled for now!
    _nvVertexBufferUnifiedMemory = Window::registerExtension("GL_NV_vertex_buffer_unified_memory");
    _nvShaderBufferLoad = Window::registerExtension("GL_NV_shader_buffer_load");

    if(_nvVertexBufferUnifiedMemory != Window::invalidExtensionID && _nvShaderBufferLoad != Window::invalidExtensionID)
    {
        _funcglBufferAddressRangeNV     = Window::registerFunction( OSG_DLSYM_UNDERSCORE"glBufferAddressRangeNV", _nvVertexBufferUnifiedMemory);
        _funcglVertexFormatNV     = Window::registerFunction( OSG_DLSYM_UNDERSCORE"glVertexFormatNV", _nvVertexBufferUnifiedMemory);
        _funcglNormalFormatNV     = Window::registerFunction( OSG_DLSYM_UNDERSCORE"glNormalFormatNV", _nvVertexBufferUnifiedMemory);
        _funcglColorFormatNV     = Window::registerFunction( OSG_DLSYM_UNDERSCORE"glColorFormatNV", _nvVertexBufferUnifiedMemory);
        _funcglSecondaryColorFormatNV     = Window::registerFunction( OSG_DLSYM_UNDERSCORE"glSecondaryColorFormatNV", _nvVertexBufferUnifiedMemory);
        _funcglTexCoordFormatNV     = Window::registerFunction( OSG_DLSYM_UNDERSCORE"glTexCoordFormatNV", _nvVertexBufferUnifiedMemory);
        _funcglIndexFormatNV     = Window::registerFunction( OSG_DLSYM_UNDERSCORE"glIndexFormatNV", _nvVertexBufferUnifiedMemory);

        _funcglMakeBufferResidentNV     = Window::registerFunction( OSG_DLSYM_UNDERSCORE"glMakeBufferResidentNV", _nvShaderBufferLoad);
        _funcglGetBufferParameterui64vNV     = Window::registerFunction( OSG_DLSYM_UNDERSCORE"glGetBufferParameterui64vNV", _nvShaderBufferLoad);
    }
#endif

    return true;
}

// ------------------------------------------------------
// vbo stuff.
// ------------------------------------------------------

#ifndef GL_ARB_vertex_buffer_object
#define GL_BUFFER_SIZE_ARB                          0x8764
#define GL_BUFFER_USAGE_ARB                         0x8765
#define GL_ARRAY_BUFFER_ARB                         0x8892
#define GL_ELEMENT_ARRAY_BUFFER_ARB                 0x8893
#define GL_ARRAY_BUFFER_BINDING_ARB                 0x8894
#define GL_ELEMENT_ARRAY_BUFFER_BINDING_ARB         0x8895
#define GL_VERTEX_ARRAY_BUFFER_BINDING_ARB          0x8896
#define GL_NORMAL_ARRAY_BUFFER_BINDING_ARB          0x8897
#define GL_COLOR_ARRAY_BUFFER_BINDING_ARB           0x8898
#define GL_INDEX_ARRAY_BUFFER_BINDING_ARB           0x8899
#define GL_TEXTURE_COORD_ARRAY_BUFFER_BINDING_ARB   0x889A
#define GL_EDGE_FLAG_ARRAY_BUFFER_BINDING_ARB       0x889B
#define GL_SECONDARY_COLOR_ARRAY_BUFFER_BINDING_ARB 0x889C
#define GL_FOG_COORDINATE_ARRAY_BUFFER_BINDING_ARB  0x889D
#define GL_WEIGHT_ARRAY_BUFFER_BINDING_ARB          0x889E
#define GL_VERTEX_ATTRIB_ARRAY_BUFFER_BINDING_ARB   0x889F
#define GL_READ_ONLY_ARB                            0x88B8
#define GL_WRITE_ONLY_ARB                           0x88B9
#define GL_READ_WRITE_ARB                           0x88BA
#define GL_BUFFER_ACCESS_ARB                        0x88BB
#define GL_BUFFER_MAPPED_ARB                        0x88BC
#define GL_BUFFER_MAP_POINTER_ARB                   0x88BD
#define GL_STREAM_DRAW_ARB                          0x88E0
#define GL_STREAM_READ_ARB                          0x88E1
#define GL_STREAM_COPY_ARB                          0x88E2
#define GL_STATIC_DRAW_ARB                          0x88E4
#define GL_STATIC_READ_ARB                          0x88E5
#define GL_STATIC_COPY_ARB                          0x88E6
#define GL_DYNAMIC_DRAW_ARB                         0x88E8
#define GL_DYNAMIC_READ_ARB                         0x88E9
#define GL_DYNAMIC_COPY_ARB                         0x88EA
#endif

#ifndef GL_NV_vertex_buffer_unified_memory
#define GL_BUFFER_GPU_ADDRESS_NV 0x8F1D
#define GL_GPU_ADDRESS_NV 0x8F34
#define GL_MAX_SHADER_BUFFER_ADDRESS_NV 0x8F35

#define GL_VERTEX_ATTRIB_ARRAY_UNIFIED_NV 0x8F1E
#define GL_ELEMENT_ARRAY_UNIFIED_NV 0x8F1F
#define GL_VERTEX_ATTRIB_ARRAY_ADDRESS_NV 0x8F20
#define GL_VERTEX_ARRAY_ADDRESS_NV 0x8F21
#define GL_NORMAL_ARRAY_ADDRESS_NV 0x8F22
#define GL_COLOR_ARRAY_ADDRESS_NV 0x8F23
#define GL_INDEX_ARRAY_ADDRESS_NV 0x8F24
#define GL_TEXTURE_COORD_ARRAY_ADDRESS_NV 0x8F25
#define GL_EDGE_FLAG_ARRAY_ADDRESS_NV 0x8F26
#define GL_SECONDARY_COLOR_ARRAY_ADDRESS_NV 0x8F27
#define GL_FOG_COORD_ARRAY_ADDRESS_NV 0x8F28
#define GL_ELEMENT_ARRAY_ADDRESS_NV 0x8F29
#define GL_VERTEX_ATTRIB_ARRAY_LENGTH_NV 0x8F2A
#define GL_VERTEX_ARRAY_LENGTH_NV 0x8F2B
#define GL_NORMAL_ARRAY_LENGTH_NV 0x8F2C
#define GL_COLOR_ARRAY_LENGTH_NV 0x8F2D
#define GL_INDEX_ARRAY_LENGTH_NV 0x8F2E
#define GL_TEXTURE_COORD_ARRAY_LENGTH_NV 0x8F2F
#define GL_EDGE_FLAG_ARRAY_LENGTH_NV 0x8F30
#define GL_SECONDARY_COLOR_ARRAY_LENGTH_NV 0x8F31
#define GL_FOG_COORD_ARRAY_LENGTH_NV 0x8F32
#define GL_ELEMENT_ARRAY_LENGTH_NV 0x8F33
#endif

#define pumpSetup2( name, typename, getmethod )                             \
    typename name##Ptr;                                                     \
    GLubyte * name##Data = NULL;                                            \
    UInt32 name##Size = 0;                                                  \
    \
    name##Ptr = geo->getmethod();                                           \
    if ( name##Ptr != NullFC )                                              \
{                                                                       \
    name##Data = name##Ptr->getData();                                  \
    name##Size = name##Ptr->getSize() * name##Ptr->getFormatSize() * name##Ptr->getDimension(); \
}

#define pumpSetup3( name, Typename, Geo, GetMethod, TargetVector, IndexMap, IndexMapType)           \
    Typename name##Ptr = Geo->GetMethod();                                  \
    if( name##Ptr != NullFC) {                                              \
    GLubyte* data = name##Ptr->getData();                               \
    size_t dataSize = name##Ptr->getSize() * name##Ptr->getFormatSize() * name##Ptr->getDimension(); \
    for( size_t idx = 0; idx < dataSize; ++idx)                         \
    TargetVector.push_back( data[idx]);                             \
    }                                                                       \
    else {                                                                  \
    TargetVector.clear();                                               \
    IndexMap &= ~IndexMapType;                                          \
}


#define setupVBO2( Data, DataSize, VBO, Type, Usage )                        \
    if(Data.size() > 0) {                                                    \
    if(VBO != 0)                                                             \
    _glDeleteBuffersARB(1, &VBO);                                            \
    _glGenBuffersARB(1, &VBO);                                                \
    _glBindBufferARB(Type, VBO);                                              \
    _glBufferDataARB(Type, DataSize, &(Data[0]), Usage);                      \
    }                                                                       \
    else {                                                                  \
    if(VBO != 0) {                                                      \
    _glDeleteBuffersARB(1, &VBO);                                   \
    VBO = 0;                                                        \
    }                                                                   \
}

#define setupVBO( Data, DataSize, VBO, Type, Usage )                        \
    if(Data) {                                                              \
    if(VBO != 0)                                                        \
    _glDeleteBuffersARB(1, &VBO);                                   \
    _glGenBuffersARB(1, &VBO);                                          \
    _glBindBufferARB(Type, VBO);                                        \
    _glBufferDataARB(Type, DataSize, Data, Usage);                      \
    }                                                                       \
    else {                                                                  \
    if(VBO != 0) {                                                      \
    _glDeleteBuffersARB(1, &VBO);                                   \
    VBO = 0;                                                        \
    }                                                                   \
}

#define setupVBOBindless(Data, DataSize, VBO, VBOAddr, VBOSize, Type) \
    if(Data) \
    { \
        _glBindBufferARB(Type, VBO); \
        _glGetBufferParameterui64vNV(Type, GL_BUFFER_GPU_ADDRESS_NV, &VBOAddr); \
        _glMakeBufferResidentNV(Type, GL_READ_ONLY_ARB); \
        VBOSize = DataSize; \
    }

UInt32 GeoVBO::_render_sides = 1 << 1;

GeoVBO::GeoVBO(Window *win, Geometry *geo) :
    _vao(0),
    _positions(0),
    _normals(0),
    _colors(0),
    _secColors(0),
    _texCoords(0),
    _texCoords1(0),
    _texCoords2(0),
    _texCoords3(0),
    _texCoords4(0),
    _texCoords5(0),
    _texCoords6(0),
    _texCoords7(0),
    _indices(0),
    _numIndices(0),
    _maxIndex(0),
    _valid_draw_properties_mask(0xffff),
    _draw_properties_mask(0xffff),
    _win(win),
    _geo(geo),
    _cleared(true)
{
    update();
}

GeoVBO::~GeoVBO()
{
    clear();
}


void GeoVBO::showASides( bool enabled)
{
    if( enabled)
        _render_sides |= SHOW_A_SIDES;
    else
        _render_sides &= ~SHOW_A_SIDES;
}

void GeoVBO::showBSides( bool enabled)
{
    if( enabled)
        _render_sides |= SHOW_B_SIDES;
    else
        _render_sides &= ~SHOW_B_SIDES;
}

void GeoVBO::clear(void)
{
#ifdef OSG_NV_BINDLESS_SUPPORT
    _positions_addr = 0;
    _normals_addr = 0;
    _colors_addr = 0;
    _secColors_addr = 0;
    _texCoords_addr = 0;
    _texCoords1_addr = 0;
    _texCoords2_addr = 0;
    _texCoords3_addr = 0;
    _texCoords4_addr = 0;
    _texCoords5_addr = 0;
    _texCoords6_addr = 0;
    _texCoords7_addr = 0;
    _indices_addr = 0;
    _positions_size = 0;
    _normals_size = 0;
    _colors_size = 0;
    _secColors_size = 0;
    _texCoords_size = 0;
    _texCoords1_size = 0;
    _texCoords2_size = 0;
    _texCoords3_size = 0;
    _texCoords4_size = 0;
    _texCoords5_size = 0;
    _texCoords6_size = 0;
    _texCoords7_size = 0;
    _indices_size = 0;
#endif

    // well ATI supports the GL_ARB_vertex_buffer_object extension but
    // not the ARB functions ...
    void (OSG_APIENTRY*_glDeleteBuffersARB)(GLsizei, const GLuint *)=
        reinterpret_cast<void (OSG_APIENTRY*)(GLsizei, const GLuint *)>(
        _win->getFunction(GeoPumpFactory::_funcglDeleteBuffers));

    if(_glDeleteBuffersARB == NULL)
    {
        _glDeleteBuffersARB =
            reinterpret_cast<void (OSG_APIENTRY*)(GLsizei, const GLuint *)>(
            _win->getFunction(GeoPumpFactory::_funcglDeleteBuffersARB));
    }

    if(_positions != 0)
    {
        _glDeleteBuffersARB(1, &_positions);
        _positions = 0;
    }

    if(_normals != 0)
    {
        _glDeleteBuffersARB(1, &_normals);
        _normals = 0;
    }

    if(_colors != 0)
    {
        _glDeleteBuffersARB(1, &_colors);
        _colors = 0;
    }

    if(_secColors != 0)
    {
        _glDeleteBuffersARB(1, &_secColors);
        _secColors = 0;
    }

    if(_texCoords != 0)
    {
        _glDeleteBuffersARB(1, &_texCoords);
        _texCoords = 0;
    }

    if(_texCoords1 != 0)
    {
        _glDeleteBuffersARB(1, &_texCoords1);
        _texCoords1 = 0;
    }

    if(_texCoords2 != 0)
    {
        _glDeleteBuffersARB(1, &_texCoords2);
        _texCoords2 = 0;
    }

    if(_texCoords3 != 0)
    {
        _glDeleteBuffersARB(1, &_texCoords3);
        _texCoords3 = 0;
    }

    if(_texCoords4 != 0)
    {
        _glDeleteBuffersARB(1, &_texCoords4);
        _texCoords4 = 0;
    }

    if(_texCoords5 != 0)
    {
        _glDeleteBuffersARB(1, &_texCoords5);
        _texCoords5 = 0;
    }

    if(_texCoords6 != 0)
    {
        _glDeleteBuffersARB(1, &_texCoords6);
        _texCoords6 = 0;
    }

    if(_texCoords7 != 0)
    {
        _glDeleteBuffersARB(1, &_texCoords7);
        _texCoords7 = 0;
    }

    if(_indices != 0)
    {
        _glDeleteBuffersARB(1, &_indices);
        _indices = 0;
    }

    if(_vao != 0)
    {
        void (OSG_APIENTRY*_glDeleteVertexArrays)(GLsizei, const GLuint *)=
            reinterpret_cast<void (OSG_APIENTRY*)(GLsizei, const GLuint *)>(
            _win->getFunction(GeoPumpFactory::_funcglDeleteVertexArrays));

        _glDeleteVertexArrays(1, &_vao);
        _vao = 0;
    }

    _cleared = true;
}

#ifdef OSG_USE_VAO
void GeoVBO::update(void)
{
    Geometry *geo = _geo;
    if( geo == NULL)
        return;

    GLenum usage = GL_STATIC_DRAW_ARB;

    // first get all the function pointers

    // VAO
    void (OSG_APIENTRY*_glGenVertexArrays)(GLsizei, GLuint *)=
        reinterpret_cast<void (OSG_APIENTRY*)(GLsizei, GLuint *)>(
        _win->getFunction(GeoPumpFactory::_funcglGenVertexArrays));

    void (OSG_APIENTRY*_glDeleteVertexArrays)(GLsizei, const GLuint *)=
        reinterpret_cast<void (OSG_APIENTRY*)(GLsizei, const GLuint *)>(
        _win->getFunction(GeoPumpFactory::_funcglDeleteVertexArrays));

    void (OSG_APIENTRY*_glBindVertexArray)(GLuint)=
        reinterpret_cast<void (OSG_APIENTRY*)(GLuint)>(
        _win->getFunction(GeoPumpFactory::_funcglBindVertexArray));

    // VBO
    void (OSG_APIENTRY*_glGenBuffersARB)(GLsizei, GLuint *)=
        reinterpret_cast<void (OSG_APIENTRY*)(GLsizei, GLuint *)>(
        _win->getFunction(GeoPumpFactory::_funcglGenBuffers));

    if(_glGenBuffersARB == NULL)
    {
        _glGenBuffersARB =
            reinterpret_cast<void (OSG_APIENTRY*)(GLsizei, GLuint *)>(
            _win->getFunction(GeoPumpFactory::_funcglGenBuffersARB));
    }

    void (OSG_APIENTRY*_glDeleteBuffersARB)(GLsizei, const GLuint *)=
        reinterpret_cast<void (OSG_APIENTRY*)(GLsizei, const GLuint *)>(
        _win->getFunction(GeoPumpFactory::_funcglDeleteBuffers));

    if(_glDeleteBuffersARB == NULL)
    {
        _glDeleteBuffersARB =
            reinterpret_cast<void (OSG_APIENTRY*)(GLsizei, const GLuint *)>(
            _win->getFunction(GeoPumpFactory::_funcglDeleteBuffersARB));
    }

    void (OSG_APIENTRY*_glBindBufferARB)(GLenum, GLuint)=
        reinterpret_cast<void (OSG_APIENTRY*)(GLenum, GLuint)>(
        _win->getFunction(GeoPumpFactory::_funcglBindBuffer));

    if(_glBindBufferARB == NULL)
    {
        _glBindBufferARB =
            reinterpret_cast<void (OSG_APIENTRY*)(GLenum, GLuint)>(
            _win->getFunction(GeoPumpFactory::_funcglBindBufferARB));
    }

    // typedef ptrdiff_t GLsizeiptrARB;
    void (OSG_APIENTRY*_glBufferDataARB)(GLenum,
        ptrdiff_t,
        const GLvoid *,
        GLenum)=
        reinterpret_cast<void (OSG_APIENTRY*)(GLenum,
        ptrdiff_t,
        const GLvoid *,
        GLenum)>(
        _win->getFunction(GeoPumpFactory::_funcglBufferData));

    if(_glBufferDataARB == NULL)
    {
        _glBufferDataARB =
            reinterpret_cast<void (OSG_APIENTRY*)(GLenum,
            ptrdiff_t,
            const GLvoid *,
            GLenum)>(
            _win->getFunction(GeoPumpFactory::_funcglBufferDataARB));
    }

#if 0
    // typedef ptrdiff_t GLsizeiptrARB;
    // typedef ptrdiff_t GLintptrARB;
    void (OSG_APIENTRY*_glBufferSubDataARB)
        (GLenum, ptrdiff_t, ptrdiff_t, const GLvoid *)=
        (void (OSG_APIENTRY*)(GLenum, ptrdiff_t, ptrdiff_t, const GLvoid *))
        _win->getFunction(GeoPumpFactory::_funcglBufferSubDataARB);
#endif

    if(_vao != 0)
        _glDeleteVertexArrays(1, &_vao);

    _glGenVertexArrays(1, &_vao);
    _glBindVertexArray(_vao);

    GeometryPtr geoPtr = GeometryPtr(*geo);
    bool isShellGeo = geo->isShell();
    if( isShellGeo)
    {
        _maxIndex = 0;
        _numIndices = 0;
        _valid_draw_properties_mask = 0;
        if( geo->getParents().size() > 0)
        {
            NodePtr parentNode = geo->getParents()[0];
            if( parentNode != NullFC)
            {
                std::vector<GLubyte> PositionsData;
                std::vector<GLubyte> ColorsData;
                std::vector<GLubyte> SecColorsData;
                std::vector<GLubyte> TexCoordsData;
                std::vector<GLubyte> TexCoords1Data;
                std::vector<GLubyte> TexCoords2Data;
                std::vector<GLubyte> TexCoords3Data;
                std::vector<GLubyte> TexCoords4Data;
                std::vector<GLubyte> TexCoords5Data;
                std::vector<GLubyte> TexCoords6Data;
                std::vector<GLubyte> TexCoords7Data;
                std::vector<GLubyte> NormalsData;
                std::vector<UInt32> IndicesData; // indices are always 32 bit

                // start with a full index map
               
                _valid_draw_properties_mask =   Geometry::MapPosition | Geometry::MapNormal | Geometry::MapColor | Geometry::MapSecondaryColor | 
                    Geometry::MapTexCoords | Geometry::MapTexCoords1 | Geometry::MapTexCoords2 | Geometry::MapTexCoords3 | 
                    Geometry::MapTexCoords4 | Geometry::MapTexCoords5 | Geometry::MapTexCoords6 | Geometry::MapTexCoords7; 
                 
                OSG::UInt32 offsetValue = 0;
                for( size_t idx = 0; idx < parentNode->getNChildren(); ++idx)
                {
                    NodePtr child = parentNode->getChild(idx);
                    if( child != NullFC && child->getActive() && child->getCore() != NullFC)
                    {
                        GeometryPtr childGeo = GeometryPtr::dcast(child->getCore());
                        if( childGeo != NullFC)
                        {
                            // we can only support single indexed geometry so we convert it if it not already is single indexed
                            if( childGeo->getIndexMapping().size() != 1)
                            {
                                OSG::separateProperties(childGeo);
                                OSG::createSingleIndex(childGeo);
                            }

                            UInt32 sideFlag = childGeo->isBSide() ? SHOW_B_SIDES : SHOW_A_SIDES;
                            if( !(_render_sides & sideFlag))
                                continue;

                            // if no indices are present, just skip this geometry
                            if( childGeo->getIndexMapping().size() != 1)
                                continue;

                            // if no vertices are present, just skip
                            GeoPositionsPtr posPtr = childGeo->getPositions();
                            if( posPtr == NullFC)
                                continue;

                            // check the geometry type. Only pure triangles are allowed
                            GeoPTypesPtr types = childGeo->getTypes();
                            if(types == NullFC || types->getSize() != 1 || types->getValue(0) != GL_TRIANGLES)
                                continue;

                            // check the geometry type. Only pure triangles are allowed
                            GeoPLengthsPtr lens = childGeo->getLengths();
                            if(lens == NullFC || lens->getSize() != 1)
                                continue;
                                                
                            // check for indices first. We only support geometries with indices, otherwise we skip
                            OSG::GeoIndicesUI32Ptr indices32 = GeoIndicesUI32Ptr::dcast(childGeo->getIndices());
                            OSG::GeoIndicesUI16Ptr indices16 = GeoIndicesUI16Ptr::dcast(childGeo->getIndices());
                            if( indices32 != NullFC)
                            {
                                OSG::UInt32* data = reinterpret_cast<OSG::UInt32*>(indices32->getData());
                                size_t idxSize = indices32->getSize() * indices32->getDimension();
                                if( lens->getValue(0) != indices32->getSize())
                                    continue;

                                for( size_t idx = 0; idx < idxSize; ++idx)
                                    IndicesData.push_back( offsetValue + data[idx]);
                            }
                            else if( indices16 != NullFC)
                            {
                                OSG::UInt16* data = reinterpret_cast<OSG::UInt16*>(indices16->getData());
                                size_t idxSize = indices16->getSize() * indices16->getDimension();
                                if( lens->getValue(0) != indices16->getSize())
                                    continue;

                                for( size_t idx = 0; idx < idxSize; ++idx)
                                    IndicesData.push_back( offsetValue + static_cast<OSG::UInt32>(data[idx]));
                            }
                            else // no indices, ignore mesh
                                continue; 


                            Transform3DPtr trans = childGeo->getTransform();
                            if( trans != NullFC && !trans->isIdentity())
                            {
                                OSG::Matrix matrix = trans->getMatrix();
                                OSG::Matrix normalMatrix = matrix;
                                normalMatrix.invert();
                                normalMatrix.transpose();
                                GeoPositions3fPtr points3f = GeoPositions3fPtr::dcast(childGeo->getPositions());
                                //GeoPositions4fPtr points4f = GeoPositions4fPtr::dcast(childGeo->getPositions());
                                if(points3f != NullFC)
                                {
                                    size_t psize = points3f->getSize();
                                    for(size_t i=0;i<psize;++i)
                                    {
                                        // transformation mit reinrechnen.
                                        Pnt3f v = points3f->getValue(i);
                                        matrix.mult(v);
                                        GLubyte* posData = reinterpret_cast<GLubyte*>(&v);
                                        for( size_t bIdx = 0; bIdx < sizeof(Pnt3f); ++bIdx)
                                            PositionsData.push_back(posData[bIdx]);
                                    }
                                }
                                /*else if(points4f != NullFC)
                                {
                                    size_t psize = points4f->getSize();
                                    for(size_t i=0;i<psize;++i)
                                    {
                                        // transformation mit reinrechnen.
                                        Pnt4f v = points4f->getValue(i);
                                        matrix.mult(v);
                                        GLubyte* posData = reinterpret_cast<GLubyte*>(&v);
                                        for( size_t bIdx = 0; bIdx < sizeof(Pnt4f); ++bIdx)
                                            PositionsData.push_back(posData[bIdx]);
                                    }
                                }
                                */
                                else
                                {
                                    PositionsData.clear();
                                    _valid_draw_properties_mask &= ~Geometry::MapPosition;
                                }
                                GeoNormals3fPtr normals = GeoNormals3fPtr::dcast(childGeo->getNormals());
                                if(normals != NullFC)
                                {
                                    size_t psize = normals->getSize();
                                    for(size_t i=0;i<psize;++i)
                                    {
                                        // transformation mit reinrechnen.
                                        Vec3f v = normals->getValue(i);
                                        normalMatrix.mult(v);
                                        GLubyte* posData = reinterpret_cast<GLubyte*>(&v);
                                        for( size_t bIdx = 0; bIdx < sizeof(Vec3f); ++bIdx)
                                            NormalsData.push_back(posData[bIdx]);
                                    }
                                }
                                else
                                {
                                    NormalsData.clear();
                                    _valid_draw_properties_mask &= ~Geometry::MapNormal;
                                }
                            }
                            else
                            {
                                pumpSetup3( Positions, GeoPositionsPtr, childGeo, getPositions, PositionsData, _valid_draw_properties_mask, Geometry::MapPosition);
                                pumpSetup3( Normals, GeoNormalsPtr, childGeo, getNormals, NormalsData, _valid_draw_properties_mask, Geometry::MapNormal);
                            }
                            
                            pumpSetup3( Colors, GeoColorsPtr, childGeo, getColors, ColorsData, _valid_draw_properties_mask, Geometry::MapColor);
                            pumpSetup3( SecColors, GeoColorsPtr, childGeo, getSecondaryColors, SecColorsData, _valid_draw_properties_mask, Geometry::MapSecondaryColor);
                            pumpSetup3( TexCoords, GeoTexCoordsPtr, childGeo, getTexCoords, TexCoordsData, _valid_draw_properties_mask, Geometry::MapTexCoords);
                            pumpSetup3( TexCoords1, GeoTexCoordsPtr, childGeo, getTexCoords1, TexCoords1Data, _valid_draw_properties_mask, Geometry::MapTexCoords1);
                            pumpSetup3( TexCoords2, GeoTexCoordsPtr, childGeo, getTexCoords2, TexCoords2Data, _valid_draw_properties_mask, Geometry::MapTexCoords2);
                            pumpSetup3( TexCoords3, GeoTexCoordsPtr, childGeo, getTexCoords3, TexCoords3Data, _valid_draw_properties_mask, Geometry::MapTexCoords3);
                            pumpSetup3( TexCoords4, GeoTexCoordsPtr, childGeo, getTexCoords4, TexCoords4Data, _valid_draw_properties_mask, Geometry::MapTexCoords4);
                            pumpSetup3( TexCoords5, GeoTexCoordsPtr, childGeo, getTexCoords5, TexCoords5Data, _valid_draw_properties_mask, Geometry::MapTexCoords5);
                            pumpSetup3( TexCoords6, GeoTexCoordsPtr, childGeo, getTexCoords6, TexCoords6Data, _valid_draw_properties_mask, Geometry::MapTexCoords6);
                            pumpSetup3( TexCoords7, GeoTexCoordsPtr, childGeo, getTexCoords7, TexCoords7Data, _valid_draw_properties_mask, Geometry::MapTexCoords7);

                            if( posPtr != NullFC)
                                offsetValue += posPtr->getSize();
                        }
                    }
                }
                _numIndices = IndicesData.size();
                _maxIndex = offsetValue;
                if( _maxIndex > 0 && _numIndices > 0 && _valid_draw_properties_mask != 0)
                {
                    if( _valid_draw_properties_mask & Geometry::MapPosition)
                        setupVBO2( PositionsData, PositionsData.size(), _positions, GL_ARRAY_BUFFER_ARB, usage);
                    if( _valid_draw_properties_mask & Geometry::MapNormal)
                        setupVBO2( NormalsData, NormalsData.size(), _normals, GL_ARRAY_BUFFER_ARB, usage);
                    if( _valid_draw_properties_mask & Geometry::MapColor)
                        setupVBO2( ColorsData, ColorsData.size(), _colors, GL_ARRAY_BUFFER_ARB, usage);
                    if( _valid_draw_properties_mask & Geometry::MapSecondaryColor)
                        setupVBO2( SecColorsData, SecColorsData.size(), _secColors, GL_ARRAY_BUFFER_ARB, usage);
                    if( _valid_draw_properties_mask & Geometry::MapTexCoords)
                        setupVBO2( TexCoordsData, TexCoordsData.size(), _texCoords, GL_ARRAY_BUFFER_ARB, usage);
                    if( _valid_draw_properties_mask & Geometry::MapTexCoords1)
                        setupVBO2( TexCoords1Data, TexCoords1Data.size(), _texCoords1, GL_ARRAY_BUFFER_ARB, usage);
                    if( _valid_draw_properties_mask & Geometry::MapTexCoords2)
                        setupVBO2( TexCoords2Data, TexCoords2Data.size(), _texCoords2, GL_ARRAY_BUFFER_ARB, usage);
                    if( _valid_draw_properties_mask & Geometry::MapTexCoords3)
                        setupVBO2( TexCoords3Data, TexCoords3Data.size(), _texCoords3, GL_ARRAY_BUFFER_ARB, usage);
                    if( _valid_draw_properties_mask & Geometry::MapTexCoords4)
                        setupVBO2( TexCoords4Data, TexCoords4Data.size(), _texCoords4, GL_ARRAY_BUFFER_ARB, usage);
                    if( _valid_draw_properties_mask & Geometry::MapTexCoords5)
                        setupVBO2( TexCoords5Data, TexCoords5Data.size(), _texCoords5, GL_ARRAY_BUFFER_ARB, usage);
                    if( _valid_draw_properties_mask & Geometry::MapTexCoords6)
                        setupVBO2( TexCoords6Data, TexCoords6Data.size(), _texCoords6, GL_ARRAY_BUFFER_ARB, usage);
                    if( _valid_draw_properties_mask & Geometry::MapTexCoords7)
                        setupVBO2( TexCoords7Data, TexCoords7Data.size(), _texCoords7, GL_ARRAY_BUFFER_ARB, usage);

                    if( IndicesData.size() > 0)
                    {
                        setupVBO( reinterpret_cast<GLubyte*>(&(IndicesData[0])), IndicesData.size() * 4, _indices, GL_ELEMENT_ARRAY_BUFFER_ARB, usage);
                    }
                }
                else
                {
                    _numIndices = 0;
                    _maxIndex = 0;
                    _valid_draw_properties_mask = 0;
                }
            }
        }
    }
    else
    {
        _valid_draw_properties_mask = 0xffff;
        _numIndices = 0;
        pumpSetup2(Positions  , GeoPositionsPtr , getPositions );
        pumpSetup2(Colors     , GeoColorsPtr    , getColors    );
        pumpSetup2(SecColors  , GeoColorsPtr    , getSecondaryColors );
        pumpSetup2(TexCoords  , GeoTexCoordsPtr , getTexCoords  );
        pumpSetup2(TexCoords1 , GeoTexCoordsPtr , getTexCoords1 );
        pumpSetup2(TexCoords2 , GeoTexCoordsPtr , getTexCoords2 );
        pumpSetup2(TexCoords3 , GeoTexCoordsPtr , getTexCoords3 );
        pumpSetup2(TexCoords4 , GeoTexCoordsPtr , getTexCoords4 );
        pumpSetup2(TexCoords5 , GeoTexCoordsPtr , getTexCoords5 );
        pumpSetup2(TexCoords6 , GeoTexCoordsPtr , getTexCoords6 );
        pumpSetup2(TexCoords7 , GeoTexCoordsPtr , getTexCoords7 );
        pumpSetup2(Normals    , GeoNormalsPtr   , getNormals    );
        pumpSetup2(Indices    , GeoIndicesPtr   , getIndices    );

        if( !PositionsData || !IndicesData)
        {
            _glBindVertexArray(0);
            return;
        }

        setupVBO( PositionsData, PositionsSize, _positions, GL_ARRAY_BUFFER_ARB, usage)
        setupVBO( NormalsData, NormalsSize, _normals, GL_ARRAY_BUFFER_ARB, usage)
        setupVBO( ColorsData, ColorsSize, _colors, GL_ARRAY_BUFFER_ARB, usage)
        setupVBO( SecColorsData, SecColorsSize, _secColors, GL_ARRAY_BUFFER_ARB, usage)
        setupVBO( TexCoordsData, TexCoordsSize, _texCoords, GL_ARRAY_BUFFER_ARB, usage)
        setupVBO( TexCoords1Data, TexCoords1Size, _texCoords1, GL_ARRAY_BUFFER_ARB, usage)
        setupVBO( TexCoords2Data, TexCoords2Size, _texCoords2, GL_ARRAY_BUFFER_ARB, usage)
        setupVBO( TexCoords3Data, TexCoords3Size, _texCoords3, GL_ARRAY_BUFFER_ARB, usage)
        setupVBO( TexCoords4Data, TexCoords4Size, _texCoords4, GL_ARRAY_BUFFER_ARB, usage)
        setupVBO( TexCoords5Data, TexCoords5Size, _texCoords5, GL_ARRAY_BUFFER_ARB, usage)
        setupVBO( TexCoords6Data, TexCoords6Size, _texCoords6, GL_ARRAY_BUFFER_ARB, usage)
        setupVBO( TexCoords7Data, TexCoords7Size, _texCoords7, GL_ARRAY_BUFFER_ARB, usage)
        setupVBO( IndicesData, IndicesSize, _indices, GL_ELEMENT_ARRAY_BUFFER_ARB, usage)
    }

    // draw internal --------------------------------

    void (OSG_APIENTRY*_glClientActiveTextureARB) (GLenum type) = reinterpret_cast<void (OSG_APIENTRY*) (GLenum type)>( _win->getFunction( GeoPumpFactory::_funcglClientActiveTextureARB));
    void (OSG_APIENTRY*_glSecondaryColorPointerEXT) (GLint size,GLenum type,GLsizei stride,const GLvoid *pointer) = reinterpret_cast<void (OSG_APIENTRY*)(GLint size, GLenum type, GLsizei stride,const GLvoid *pointer)>(_win->getFunction(GeoPumpFactory::_funcglSecondaryColorPointer));

    if(_glBindBufferARB == NULL)
        _glBindBufferARB = reinterpret_cast<void (OSG_APIENTRY*)(GLenum, GLuint)>(_win->getFunction(GeoPumpFactory::_funcglBindBufferARB));

    // removed the hasExtension(GeoPumpFactory::_extDrawRangeElements) and
    // hasExtension(GeoPumpFactory::_extMultitexture) calls with VBO support
    // this extensions are also supported!
    Geometry* shellGeo = _geo;
    Int16 modified=0;

    bool isBside = geo->isBSide();
    if( isShellGeo)
    {
        if( geo->getParents().size() > 0)
        {
            NodePtr parentNode = geo->getParents()[0];
            if( parentNode != NullFC)
            {
                for( size_t childIdx = 0; childIdx < parentNode->getNChildren(); ++childIdx)
                {
                    NodePtr child = parentNode->getChild(childIdx);
                    if( child != NullFC && child->getCore() != NullFC)
                    {
                        GeometryPtr childGeo = GeometryPtr::dcast(child->getCore());
                        if( childGeo != NullFC)
                        {
                            geo = childGeo.getCPtr();
                            break;
                        }
                    }
                }
            }
        }
        if( geo == shellGeo) // if we have not found a valid child
            return;
    }

    GeoPositionsPtr PositionsPtr = geo->getPositions();
    if (PositionsPtr != NullFC &&
        (_valid_draw_properties_mask & _draw_properties_mask & Geometry::MapPosition))
    {
        _glBindBufferARB(GL_ARRAY_BUFFER_ARB, _positions);
        glVertexPointer(PositionsPtr->getDimension(), PositionsPtr->getFormat(),
            0, NULL);
        glEnableClientState(GL_VERTEX_ARRAY);
        modified|=(1<<0);
    }

    GeoNormalsPtr NormalsPtr = geo->getNormals();
    if (NormalsPtr != NullFC &&
        (_valid_draw_properties_mask & _draw_properties_mask & Geometry::MapNormal))
    {
        _glBindBufferARB(GL_ARRAY_BUFFER_ARB, _normals);
        glNormalPointer(NormalsPtr->getFormat(), 0, NULL);
        glEnableClientState(GL_NORMAL_ARRAY);
        modified|=(1<<3);
    }

    GeoColorsPtr ColorsPtr = geo->getColors();
    if (ColorsPtr != NullFC &&
        (_valid_draw_properties_mask & _draw_properties_mask & Geometry::MapColor))
    {
        _glBindBufferARB(GL_ARRAY_BUFFER_ARB, _colors);
        glColorPointer(ColorsPtr->getDimension(), ColorsPtr->getFormat(), 0,
            NULL);
        glEnableClientState(GL_COLOR_ARRAY);
        modified|=(1<<1);
    }

    GeoColorsPtr SecColorsPtr = geo->getSecondaryColors();
    if (SecColorsPtr != NullFC &&
        (_valid_draw_properties_mask & _draw_properties_mask & Geometry::MapSecondaryColor))
    {
        _glBindBufferARB(GL_ARRAY_BUFFER_ARB, _secColors);
        _glSecondaryColorPointerEXT(SecColorsPtr->getDimension(),
            SecColorsPtr->getFormat(),
            0, NULL);
        glEnableClientState(GL_SECONDARY_COLOR_ARRAY_EXT);
        modified|=(1<<2);
    }

    GeoTexCoordsPtr TexCoordsPtr = geo->getTexCoords();
    if (TexCoordsPtr != NullFC &&
        (_valid_draw_properties_mask & _draw_properties_mask & Geometry::MapTexCoords))
    {
        _glClientActiveTextureARB(GL_TEXTURE0_ARB);

        _glBindBufferARB(GL_ARRAY_BUFFER_ARB, _texCoords);
        glTexCoordPointer (TexCoordsPtr->getDimension(),
            TexCoordsPtr->getFormat(),
            0, NULL);
        glEnableClientState(GL_TEXTURE_COORD_ARRAY);
        modified|=(1<<4);
    }
    else
    {
        // special hack if we render a geometry without texture coordinates and with a material with enabled texture
        // we get flicker rendering because of random (last global texcoord state) texcoord values in our shaders.
        // so we upload one zero tex coord!
        glTexCoord2f(0.0f, 0.0f);
    }

     GeoTexCoordsPtr TexCoords1Ptr = geo->getTexCoords1();
    if (TexCoords1Ptr != NullFC &&
        (_valid_draw_properties_mask & _draw_properties_mask & Geometry::MapTexCoords1))
    {
        _glClientActiveTextureARB(GL_TEXTURE1_ARB);
        _glBindBufferARB(GL_ARRAY_BUFFER_ARB, _texCoords1);
        glTexCoordPointer (TexCoords1Ptr->getDimension(),
            TexCoords1Ptr->getFormat(),
            0, NULL);
        glEnableClientState(GL_TEXTURE_COORD_ARRAY);
        modified|=(1<<5);
    }

    GeoTexCoordsPtr TexCoords2Ptr = geo->getTexCoords2();
    if (TexCoords2Ptr != NullFC && (_valid_draw_properties_mask & _draw_properties_mask & Geometry::MapTexCoords2))
    {
        _glClientActiveTextureARB(GL_TEXTURE2_ARB);
        _glBindBufferARB(GL_ARRAY_BUFFER_ARB, _texCoords2);
        glTexCoordPointer (TexCoords2Ptr->getDimension(),
            TexCoords2Ptr->getFormat(),
            0, NULL);
        glEnableClientState(GL_TEXTURE_COORD_ARRAY);
        modified|=(1<<6);
    }

    GeoTexCoordsPtr TexCoords3Ptr = geo->getTexCoords3();
    if (TexCoords3Ptr != NullFC && (_valid_draw_properties_mask & _draw_properties_mask & Geometry::MapTexCoords3))
    {
        _glClientActiveTextureARB(GL_TEXTURE3_ARB);

        _glBindBufferARB(GL_ARRAY_BUFFER_ARB, _texCoords3);
        glTexCoordPointer (TexCoords3Ptr->getDimension(),
            TexCoords3Ptr->getFormat(),
            0, NULL);
        glEnableClientState(GL_TEXTURE_COORD_ARRAY);
        modified|=(1<<7);
    }

    GeoTexCoordsPtr TexCoords4Ptr = geo->getTexCoords4();
    if (TexCoords4Ptr != NullFC &&
        (_valid_draw_properties_mask & _draw_properties_mask & Geometry::MapTexCoords4))
    {

        _glClientActiveTextureARB(GL_TEXTURE4_ARB);

        _glBindBufferARB(GL_ARRAY_BUFFER_ARB, _texCoords4);
        glTexCoordPointer (TexCoords4Ptr->getDimension(),
            TexCoords4Ptr->getFormat(),
            0, NULL);
        glEnableClientState(GL_TEXTURE_COORD_ARRAY);
        modified|=(1<<8);
    }

    GeoTexCoordsPtr TexCoords5Ptr = geo->getTexCoords5();
    if (TexCoords5Ptr != NullFC &&
        (_valid_draw_properties_mask & _draw_properties_mask & Geometry::MapTexCoords5))
    {

        _glClientActiveTextureARB(GL_TEXTURE5_ARB);

        _glBindBufferARB(GL_ARRAY_BUFFER_ARB, _texCoords5);
        glTexCoordPointer (TexCoords5Ptr->getDimension(),
            TexCoords5Ptr->getFormat(),
            0, NULL);
        glEnableClientState(GL_TEXTURE_COORD_ARRAY);
        modified|=(1<<9);
    }

    GeoTexCoordsPtr TexCoords6Ptr = geo->getTexCoords6();
    if (TexCoords6Ptr != NullFC &&
        (_valid_draw_properties_mask & _draw_properties_mask & Geometry::MapTexCoords6))
    {

        _glClientActiveTextureARB(GL_TEXTURE6_ARB);

        _glBindBufferARB(GL_ARRAY_BUFFER_ARB, _texCoords6);
        glTexCoordPointer (TexCoords6Ptr->getDimension(),
            TexCoords6Ptr->getFormat(),
            0, NULL);
        glEnableClientState(GL_TEXTURE_COORD_ARRAY);
        modified|=(1<<10);
    }

    GeoTexCoordsPtr TexCoords7Ptr = geo->getTexCoords7();
    if (TexCoords7Ptr != NullFC &&
        (_valid_draw_properties_mask & _draw_properties_mask & Geometry::MapTexCoords7))
    {
        _glClientActiveTextureARB(GL_TEXTURE7_ARB);

        _glBindBufferARB(GL_ARRAY_BUFFER_ARB, _texCoords7);
        glTexCoordPointer (TexCoords7Ptr->getDimension(), TexCoords7Ptr->getFormat(),
            0, NULL);
        glEnableClientState(GL_TEXTURE_COORD_ARRAY);
        modified|=(1<<11);
    }
    // amz special nvidia HACK we need to upload something to tex coord 5-8!
    // without this we get strange memory allocations!
#if 1
    else if(PositionsPtr != NullFC)
    {
        _glClientActiveTextureARB(GL_TEXTURE7_ARB);
        _glBindBufferARB(GL_ARRAY_BUFFER_ARB, _positions);
        glTexCoordPointer (PositionsPtr->getDimension(), PositionsPtr->getFormat(), 0, NULL);
        glEnableClientState(GL_TEXTURE_COORD_ARRAY);
        modified|=(1<<11);
    }
#endif

    // ----------------------------------------------

    _glBindVertexArray(0);

    if(modified&(1<<0))
        glDisableClientState(GL_VERTEX_ARRAY);
    if(modified&(1<<1))
        glDisableClientState(GL_COLOR_ARRAY);
    if(modified&(1<<2))
        glDisableClientState(GL_SECONDARY_COLOR_ARRAY_EXT);
    if(modified&(1<<3))
        glDisableClientState(GL_NORMAL_ARRAY);

    if(modified&(1<<4))
    {
        _glClientActiveTextureARB(GL_TEXTURE0_ARB);
        glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    }

    if(modified&(1<<5))
    {
        _glClientActiveTextureARB(GL_TEXTURE1_ARB);
        glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    }

    if(modified&(1<<6))
    {
        _glClientActiveTextureARB(GL_TEXTURE2_ARB);
        glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    }

    if(modified&(1<<7))
    {
        _glClientActiveTextureARB(GL_TEXTURE3_ARB);
        glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    }

    if(modified&(1<<8))
    {
        _glClientActiveTextureARB(GL_TEXTURE4_ARB);
        glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    }

    if(modified&(1<<9))
    {
        _glClientActiveTextureARB(GL_TEXTURE5_ARB);
        glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    }

    if(modified&(1<<10))
    {
        _glClientActiveTextureARB(GL_TEXTURE6_ARB);
        glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    }

    if(modified&(1<<11))
    {
        _glClientActiveTextureARB(GL_TEXTURE7_ARB);
        glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    }
    _glBindBufferARB(GL_ARRAY_BUFFER_ARB, 0);
    _glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, 0);

    _cleared = false;
}

void GeoVBO::draw(void)
{
    if(_cleared)
        return;

    void (OSG_APIENTRY*_glBindVertexArray)(GLuint)= reinterpret_cast<void (OSG_APIENTRY*)(GLuint)>(_win->getFunction(GeoPumpFactory::_funcglBindVertexArray));
    void (OSG_APIENTRY*_glBindBufferARB)(GLenum, GLuint) = reinterpret_cast<void (OSG_APIENTRY*)(GLenum, GLuint)>(_win->getFunction(GeoPumpFactory::_funcglBindBuffer));
    void (OSG_APIENTRY *osgGLDrawRangeElementsEXT)(GLenum mode, GLuint start, GLuint end, GLsizei count, GLenum type, const GLvoid*indices) = reinterpret_cast<void (OSG_APIENTRY *)(GLenum mode, GLuint start, GLuint end, GLsizei count, GLenum type, const GLvoid*indices)>(_win->getFunction(GeoPumpFactory::_funcglDrawRangeElementsEXT));

    if(_glBindBufferARB == NULL)
        _glBindBufferARB = reinterpret_cast<void (OSG_APIENTRY*)(GLenum, GLuint)>(_win->getFunction(GeoPumpFactory::_funcglBindBufferARB));

    _glBindVertexArray(_vao);

    //_glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, _indices);

    Geometry *geo = _geo;

    GeoPositionsPtr PositionsPtr = geo->getPositions();

    if(geo->isShell())
    {
        if( _valid_draw_properties_mask)
        {
            // for a shell, assume we only have triangles
            osgGLDrawRangeElementsEXT(GL_TRIANGLES,  0, _maxIndex, _numIndices, GL_UNSIGNED_INT, NULL);
        }
    }
    else
    {
        pumpSetup(Lengths    , GeoPLengthsPtr  , getLengths    );
        pumpSetup(Types      , GeoPTypesPtr    , getTypes      );
        pumpSetup(Indices    , GeoIndicesPtr   , getIndices    );
        IndicesData = IndicesData; // workaround for irix compiler.

        UInt32 lendummy;
        UInt32 LengthsSize;
        bool   len16 = false;

        if(LengthsPtr != NullFC && LengthsData != NULL)
        {
            LengthsSize = LengthsPtr->getSize();
            len16 = (LengthsPtr->getFormatSize() == 2);
        }
        else
        {
            LengthsSize = 1;
            LengthsData = reinterpret_cast<GLubyte *>(&lendummy);
            lendummy = PositionsPtr->getSize();
        }

        UInt32 LengthsInd = 0, TypesInd = 0, IndicesInd = 0,
            IndicesSize = IndicesPtr->getSize();

        if(len16 == false)
        {
            for(LengthsInd = 0; LengthsInd < LengthsSize; LengthsInd++)
            {
                UInt32 count  = *reinterpret_cast<UInt32*>(LengthsData + LengthsInd * LengthsStride);
                // for vbo's we need only a offset!
                UInt8 *vind = static_cast<UInt8 *>(NULL) + IndicesStride * IndicesInd;
                IndicesInd += count;

                if(count != 0)
                {
                    osgGLDrawRangeElementsEXT(*(TypesData + TypesInd++ * TypesStride),
                        geo->getLowindices(LengthsInd),
                        geo->getHighindices(LengthsInd),
                        count,
                        IndicesPtr->getFormat(),
                        vind);
                }
            }
        }
        else if(len16 == true)
        {
            for(LengthsInd = 0; LengthsInd < LengthsSize; LengthsInd++)
            {
                UInt16 count  = *reinterpret_cast<UInt16*>(LengthsData + LengthsInd * LengthsStride);
                // for vbo's we need only a offset!
                UInt8 *vind = static_cast<UInt8 *>(NULL) + IndicesStride * IndicesInd;
                IndicesInd += count;

                if(count != 0)
                {
                    osgGLDrawRangeElementsEXT(*(TypesData + TypesInd++ * TypesStride), geo->getLowindices(LengthsInd), geo->getHighindices(LengthsInd), count, IndicesPtr->getFormat(), vind);
                }
            }
        }
        else
        {
            FWARNING(("VBO::draw: unknown Lengths format size\n"));
        }
    }

    //_glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, 0);

    _glBindVertexArray(0);
}

#else

// -----------------------------------------------------------------------
// VBO without VAO

void GeoVBO::update(void)
{
    Geometry *geo = _geo;
    if( geo == NULL)
        return;

    GLenum usage = GL_STATIC_DRAW_ARB;

    // first get all the function pointers
    void (OSG_APIENTRY*_glGenBuffersARB)(GLsizei, GLuint *)=
        reinterpret_cast<void (OSG_APIENTRY*)(GLsizei, GLuint *)>(
        _win->getFunction(GeoPumpFactory::_funcglGenBuffers));

    if(_glGenBuffersARB == NULL)
    {
        _glGenBuffersARB =
            reinterpret_cast<void (OSG_APIENTRY*)(GLsizei, GLuint *)>(
            _win->getFunction(GeoPumpFactory::_funcglGenBuffersARB));
    }

    void (OSG_APIENTRY*_glDeleteBuffersARB)(GLsizei, const GLuint *)=
        reinterpret_cast<void (OSG_APIENTRY*)(GLsizei, const GLuint *)>(
        _win->getFunction(GeoPumpFactory::_funcglDeleteBuffers));

    if(_glDeleteBuffersARB == NULL)
    {
        _glDeleteBuffersARB =
            reinterpret_cast<void (OSG_APIENTRY*)(GLsizei, const GLuint *)>(
            _win->getFunction(GeoPumpFactory::_funcglDeleteBuffersARB));
    }

    void (OSG_APIENTRY*_glBindBufferARB)(GLenum, GLuint)=
        reinterpret_cast<void (OSG_APIENTRY*)(GLenum, GLuint)>(
        _win->getFunction(GeoPumpFactory::_funcglBindBuffer));

    if(_glBindBufferARB == NULL)
    {
        _glBindBufferARB =
            reinterpret_cast<void (OSG_APIENTRY*)(GLenum, GLuint)>(
            _win->getFunction(GeoPumpFactory::_funcglBindBufferARB));
    }

    // typedef ptrdiff_t GLsizeiptrARB;
    void (OSG_APIENTRY*_glBufferDataARB)(GLenum,
        ptrdiff_t,
        const GLvoid *,
        GLenum)=
        reinterpret_cast<void (OSG_APIENTRY*)(GLenum,
        ptrdiff_t,
        const GLvoid *,
        GLenum)>(
        _win->getFunction(GeoPumpFactory::_funcglBufferData));

    if(_glBufferDataARB == NULL)
    {
        _glBufferDataARB =
            reinterpret_cast<void (OSG_APIENTRY*)(GLenum,
            ptrdiff_t,
            const GLvoid *,
            GLenum)>(
            _win->getFunction(GeoPumpFactory::_funcglBufferDataARB));
    }

#ifdef OSG_NV_BINDLESS_SUPPORT
    // bindles VBO stuff.
    bool bindless = (GeoPumpFactory::_nvVertexBufferUnifiedMemory != Window::invalidExtensionID && GeoPumpFactory::_nvShaderBufferLoad != Window::invalidExtensionID);

    void (OSG_APIENTRY*_glMakeBufferResidentNV)(GLenum, GLenum) = NULL;
    void (OSG_APIENTRY*_glGetBufferParameterui64vNV)(GLenum, GLenum, GLuint64 *) = NULL;

    if(bindless)
    {
        _glMakeBufferResidentNV = reinterpret_cast<void (OSG_APIENTRY*)(GLenum, GLenum)>(_win->getFunction(GeoPumpFactory::_funcglMakeBufferResidentNV));
        _glGetBufferParameterui64vNV = reinterpret_cast<void (OSG_APIENTRY*)(GLenum, GLenum, GLuint64 *)>(_win->getFunction(GeoPumpFactory::_funcglGetBufferParameterui64vNV));

        if(_glMakeBufferResidentNV == NULL || _glGetBufferParameterui64vNV == NULL)
        {
            GeoPumpFactory::_nvVertexBufferUnifiedMemory = Window::invalidExtensionID;
            GeoPumpFactory::_nvShaderBufferLoad = Window::invalidExtensionID;
            FWARNING(("Found a not working NV_vertex_buffer_unified_memory extension!\n"));
            bindless = false;
        }
    }
#endif


    // ------------------------

#if 0
    // typedef ptrdiff_t GLsizeiptrARB;
    // typedef ptrdiff_t GLintptrARB;
    void (OSG_APIENTRY*_glBufferSubDataARB)
        (GLenum, ptrdiff_t, ptrdiff_t, const GLvoid *)=
        (void (OSG_APIENTRY*)(GLenum, ptrdiff_t, ptrdiff_t, const GLvoid *))
        _win->getFunction(GeoPumpFactory::_funcglBufferSubDataARB);
#endif

    GeometryPtr geoPtr = GeometryPtr(*geo);
    bool isShellGeo = geo->isShell();
    if( isShellGeo)
    {
        _maxIndex = 0;
        _numIndices = 0;
        _valid_draw_properties_mask = 0;
        if( geo->getParents().size() > 0)
        {
            NodePtr parentNode = geo->getParents()[0];
            if( parentNode != NullFC)
            {
                std::vector<GLubyte> PositionsData;
                std::vector<GLubyte> ColorsData;
                std::vector<GLubyte> SecColorsData;
                std::vector<GLubyte> TexCoordsData;
                std::vector<GLubyte> TexCoords1Data;
                std::vector<GLubyte> TexCoords2Data;
                std::vector<GLubyte> TexCoords3Data;
                std::vector<GLubyte> TexCoords4Data;
                std::vector<GLubyte> TexCoords5Data;
                std::vector<GLubyte> TexCoords6Data;
                std::vector<GLubyte> TexCoords7Data;
                std::vector<GLubyte> NormalsData;
                std::vector<UInt32> IndicesData; // indices are always 32 bit

                // start with a full index map

                _valid_draw_properties_mask =   Geometry::MapPosition | Geometry::MapNormal | Geometry::MapColor | Geometry::MapSecondaryColor |
                    Geometry::MapTexCoords | Geometry::MapTexCoords1 | Geometry::MapTexCoords2 | Geometry::MapTexCoords3 |
                    Geometry::MapTexCoords4 | Geometry::MapTexCoords5 | Geometry::MapTexCoords6 | Geometry::MapTexCoords7;

                OSG::UInt32 offsetValue = 0;
                for( size_t idx = 0; idx < parentNode->getNChildren(); ++idx)
                {
                    NodePtr child = parentNode->getChild(idx);
                    if( child != NullFC && child->getActive() && child->getCore() != NullFC)
                    {
                        GeometryPtr childGeo = GeometryPtr::dcast(child->getCore());
                        if( childGeo != NullFC)
                        {
                            // we can only support single indexed geometry so we convert it if it not already is single indexed
                            if( childGeo->getIndexMapping().size() != 1)
                            {
                                OSG::separateProperties(childGeo);
                                OSG::createSingleIndex(childGeo);
                            }

                            UInt32 sideFlag = childGeo->isBSide() ? SHOW_B_SIDES : SHOW_A_SIDES;
                            if( !(_render_sides & sideFlag))
                                continue;

                            // if no indices are present, just skip this geometry
                            if( childGeo->getIndexMapping().size() != 1)
                                continue;

                            // if no vertices are present, just skip
                            GeoPositionsPtr posPtr = childGeo->getPositions();
                            if( posPtr == NullFC)
                                continue;

                            // check the geometry type. Only pure triangles are allowed
                            GeoPTypesPtr types = childGeo->getTypes();
                            if(types == NullFC || types->getSize() != 1 || types->getValue(0) != GL_TRIANGLES)
                                continue;

                            // check the geometry type. Only pure triangles are allowed
                            GeoPLengthsPtr lens = childGeo->getLengths();
                            if(lens == NullFC || lens->getSize() != 1)
                                continue;

                            // check for indices first. We only support geometries with indices, otherwise we skip
                            OSG::GeoIndicesUI32Ptr indices32 = GeoIndicesUI32Ptr::dcast(childGeo->getIndices());
                            OSG::GeoIndicesUI16Ptr indices16 = GeoIndicesUI16Ptr::dcast(childGeo->getIndices());
                            if( indices32 != NullFC)
                            {
                                OSG::UInt32* data = reinterpret_cast<OSG::UInt32*>(indices32->getData());
                                size_t idxSize = indices32->getSize() * indices32->getDimension();
                                if( lens->getValue(0) != indices32->getSize())
                                    continue;

                                for( size_t idx = 0; idx < idxSize; ++idx)
                                    IndicesData.push_back( offsetValue + data[idx]);
                            }
                            else if( indices16 != NullFC)
                            {
                                OSG::UInt16* data = reinterpret_cast<OSG::UInt16*>(indices16->getData());
                                size_t idxSize = indices16->getSize() * indices16->getDimension();
                                if( lens->getValue(0) != indices16->getSize())
                                    continue;

                                for( size_t idx = 0; idx < idxSize; ++idx)
                                    IndicesData.push_back( offsetValue + static_cast<OSG::UInt32>(data[idx]));
                            }
                            else // no indices, ignore mesh
                                continue;


                            Transform3DPtr trans = childGeo->getTransform();
                            if( trans != NullFC && !trans->isIdentity())
                            {
                                OSG::Matrix matrix = trans->getMatrix();
                                OSG::Matrix normalMatrix = matrix;
                                normalMatrix.invert();
                                normalMatrix.transpose();
                                GeoPositions3fPtr points3f = GeoPositions3fPtr::dcast(childGeo->getPositions());
                                //GeoPositions4fPtr points4f = GeoPositions4fPtr::dcast(childGeo->getPositions());
                                if(points3f != NullFC)
                                {
                                    size_t psize = points3f->getSize();
                                    for(size_t i=0;i<psize;++i)
                                    {
                                        // transformation mit reinrechnen.
                                        Pnt3f v = points3f->getValue(i);
                                        matrix.mult(v);
                                        GLubyte* posData = reinterpret_cast<GLubyte*>(&v);
                                        for( size_t bIdx = 0; bIdx < sizeof(Pnt3f); ++bIdx)
                                            PositionsData.push_back(posData[bIdx]);
                                    }
                                }
                                /*else if(points4f != NullFC)
                                {
                                    size_t psize = points4f->getSize();
                                    for(size_t i=0;i<psize;++i)
                                    {
                                        // transformation mit reinrechnen.
                                        Pnt4f v = points4f->getValue(i);
                                        matrix.mult(v);
                                        GLubyte* posData = reinterpret_cast<GLubyte*>(&v);
                                        for( size_t bIdx = 0; bIdx < sizeof(Pnt4f); ++bIdx)
                                            PositionsData.push_back(posData[bIdx]);
                                    }
                                }
                                */
                                else
                                {
                                    PositionsData.clear();
                                    _valid_draw_properties_mask &= ~Geometry::MapPosition;
                                }
                                GeoNormals3fPtr normals = GeoNormals3fPtr::dcast(childGeo->getNormals());
                                if(normals != NullFC)
                                {
                                    size_t psize = normals->getSize();
                                    for(size_t i=0;i<psize;++i)
                                    {
                                        // transformation mit reinrechnen.
                                        Vec3f v = normals->getValue(i);
                                        normalMatrix.mult(v);
                                        GLubyte* posData = reinterpret_cast<GLubyte*>(&v);
                                        for( size_t bIdx = 0; bIdx < sizeof(Vec3f); ++bIdx)
                                            NormalsData.push_back(posData[bIdx]);
                                    }
                                }
                                else
                                {
                                    NormalsData.clear();
                                    _valid_draw_properties_mask &= ~Geometry::MapNormal;
                                }
                            }
                            else
                            {
                                pumpSetup3( Positions, GeoPositionsPtr, childGeo, getPositions, PositionsData, _valid_draw_properties_mask, Geometry::MapPosition);
                                pumpSetup3( Normals, GeoNormalsPtr, childGeo, getNormals, NormalsData, _valid_draw_properties_mask, Geometry::MapNormal);
                            }

                            pumpSetup3( Colors, GeoColorsPtr, childGeo, getColors, ColorsData, _valid_draw_properties_mask, Geometry::MapColor);
                            pumpSetup3( SecColors, GeoColorsPtr, childGeo, getSecondaryColors, SecColorsData, _valid_draw_properties_mask, Geometry::MapSecondaryColor);
                            pumpSetup3( TexCoords, GeoTexCoordsPtr, childGeo, getTexCoords, TexCoordsData, _valid_draw_properties_mask, Geometry::MapTexCoords);
                            pumpSetup3( TexCoords1, GeoTexCoordsPtr, childGeo, getTexCoords1, TexCoords1Data, _valid_draw_properties_mask, Geometry::MapTexCoords1);
                            pumpSetup3( TexCoords2, GeoTexCoordsPtr, childGeo, getTexCoords2, TexCoords2Data, _valid_draw_properties_mask, Geometry::MapTexCoords2);
                            pumpSetup3( TexCoords3, GeoTexCoordsPtr, childGeo, getTexCoords3, TexCoords3Data, _valid_draw_properties_mask, Geometry::MapTexCoords3);
                            pumpSetup3( TexCoords4, GeoTexCoordsPtr, childGeo, getTexCoords4, TexCoords4Data, _valid_draw_properties_mask, Geometry::MapTexCoords4);
                            pumpSetup3( TexCoords5, GeoTexCoordsPtr, childGeo, getTexCoords5, TexCoords5Data, _valid_draw_properties_mask, Geometry::MapTexCoords5);
                            pumpSetup3( TexCoords6, GeoTexCoordsPtr, childGeo, getTexCoords6, TexCoords6Data, _valid_draw_properties_mask, Geometry::MapTexCoords6);
                            pumpSetup3( TexCoords7, GeoTexCoordsPtr, childGeo, getTexCoords7, TexCoords7Data, _valid_draw_properties_mask, Geometry::MapTexCoords7);

                            if( posPtr != NullFC)
                                offsetValue += posPtr->getSize();
                        }
                    }
                }
                _numIndices = IndicesData.size();
                _maxIndex = offsetValue;
                if( _maxIndex > 0 && _numIndices > 0 && _valid_draw_properties_mask != 0)
                {
                    if( _valid_draw_properties_mask & Geometry::MapPosition)
                        setupVBO2( PositionsData, PositionsData.size(), _positions, GL_ARRAY_BUFFER_ARB, usage);
                    if( _valid_draw_properties_mask & Geometry::MapNormal)
                        setupVBO2( NormalsData, NormalsData.size(), _normals, GL_ARRAY_BUFFER_ARB, usage);
                    if( _valid_draw_properties_mask & Geometry::MapColor)
                        setupVBO2( ColorsData, ColorsData.size(), _colors, GL_ARRAY_BUFFER_ARB, usage);
                    if( _valid_draw_properties_mask & Geometry::MapSecondaryColor)
                        setupVBO2( SecColorsData, SecColorsData.size(), _secColors, GL_ARRAY_BUFFER_ARB, usage);
                    if( _valid_draw_properties_mask & Geometry::MapTexCoords)
                        setupVBO2( TexCoordsData, TexCoordsData.size(), _texCoords, GL_ARRAY_BUFFER_ARB, usage);
                    if( _valid_draw_properties_mask & Geometry::MapTexCoords1)
                        setupVBO2( TexCoords1Data, TexCoords1Data.size(), _texCoords1, GL_ARRAY_BUFFER_ARB, usage);
                    if( _valid_draw_properties_mask & Geometry::MapTexCoords2)
                        setupVBO2( TexCoords2Data, TexCoords2Data.size(), _texCoords2, GL_ARRAY_BUFFER_ARB, usage);
                    if( _valid_draw_properties_mask & Geometry::MapTexCoords3)
                        setupVBO2( TexCoords3Data, TexCoords3Data.size(), _texCoords3, GL_ARRAY_BUFFER_ARB, usage);
                    if( _valid_draw_properties_mask & Geometry::MapTexCoords4)
                        setupVBO2( TexCoords4Data, TexCoords4Data.size(), _texCoords4, GL_ARRAY_BUFFER_ARB, usage);
                    if( _valid_draw_properties_mask & Geometry::MapTexCoords5)
                        setupVBO2( TexCoords5Data, TexCoords5Data.size(), _texCoords5, GL_ARRAY_BUFFER_ARB, usage);
                    if( _valid_draw_properties_mask & Geometry::MapTexCoords6)
                        setupVBO2( TexCoords6Data, TexCoords6Data.size(), _texCoords6, GL_ARRAY_BUFFER_ARB, usage);
                    if( _valid_draw_properties_mask & Geometry::MapTexCoords7)
                        setupVBO2( TexCoords7Data, TexCoords7Data.size(), _texCoords7, GL_ARRAY_BUFFER_ARB, usage);

                    if( IndicesData.size() > 0)
                    {
                        setupVBO( reinterpret_cast<GLubyte*>(&(IndicesData[0])), IndicesData.size() * 4, _indices, GL_ELEMENT_ARRAY_BUFFER_ARB, usage);
                    }
                }
                else
                {
                    _numIndices = 0;
                    _maxIndex = 0;
                    _valid_draw_properties_mask = 0;
                }
                _glBindBufferARB(GL_ARRAY_BUFFER_ARB, 0);
                _glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, 0);

                _cleared = false;
            }
        }
        return;
    }
    _valid_draw_properties_mask = 0xffff;
    _numIndices = 0;
    pumpSetup2(Positions  , GeoPositionsPtr , getPositions );
    pumpSetup2(Colors     , GeoColorsPtr    , getColors    );
    pumpSetup2(SecColors  , GeoColorsPtr    , getSecondaryColors );
    pumpSetup2(TexCoords  , GeoTexCoordsPtr , getTexCoords  );
    pumpSetup2(TexCoords1 , GeoTexCoordsPtr , getTexCoords1 );
    pumpSetup2(TexCoords2 , GeoTexCoordsPtr , getTexCoords2 );
    pumpSetup2(TexCoords3 , GeoTexCoordsPtr , getTexCoords3 );
    pumpSetup2(TexCoords4 , GeoTexCoordsPtr , getTexCoords4 );
    pumpSetup2(TexCoords5 , GeoTexCoordsPtr , getTexCoords5 );
    pumpSetup2(TexCoords6 , GeoTexCoordsPtr , getTexCoords6 );
    pumpSetup2(TexCoords7 , GeoTexCoordsPtr , getTexCoords7 );
    pumpSetup2(Normals    , GeoNormalsPtr   , getNormals    );
    pumpSetup2(Indices    , GeoIndicesPtr   , getIndices    );

    if( !PositionsData || !IndicesData)
        return;

    setupVBO( PositionsData, PositionsSize, _positions, GL_ARRAY_BUFFER_ARB, usage)
    setupVBO( NormalsData, NormalsSize, _normals, GL_ARRAY_BUFFER_ARB, usage)
    setupVBO( ColorsData, ColorsSize, _colors, GL_ARRAY_BUFFER_ARB, usage)
    setupVBO( SecColorsData, SecColorsSize, _secColors, GL_ARRAY_BUFFER_ARB, usage)
    setupVBO( TexCoordsData, TexCoordsSize, _texCoords, GL_ARRAY_BUFFER_ARB, usage)
    setupVBO( TexCoords1Data, TexCoords1Size, _texCoords1, GL_ARRAY_BUFFER_ARB, usage)
    setupVBO( TexCoords2Data, TexCoords2Size, _texCoords2, GL_ARRAY_BUFFER_ARB, usage)
    setupVBO( TexCoords3Data, TexCoords3Size, _texCoords3, GL_ARRAY_BUFFER_ARB, usage)
    setupVBO( TexCoords4Data, TexCoords4Size, _texCoords4, GL_ARRAY_BUFFER_ARB, usage)
    setupVBO( TexCoords5Data, TexCoords5Size, _texCoords5, GL_ARRAY_BUFFER_ARB, usage)
    setupVBO( TexCoords6Data, TexCoords6Size, _texCoords6, GL_ARRAY_BUFFER_ARB, usage)
    setupVBO( TexCoords7Data, TexCoords7Size, _texCoords7, GL_ARRAY_BUFFER_ARB, usage)
    setupVBO( IndicesData, IndicesSize, _indices, GL_ELEMENT_ARRAY_BUFFER_ARB, usage)

#ifdef OSG_NV_BINDLESS_SUPPORT
    if(bindless)
    {
        setupVBOBindless(PositionsData, PositionsSize, _positions, _positions_addr, _positions_size, GL_ARRAY_BUFFER_ARB)
        setupVBOBindless(NormalsData, NormalsSize,_normals,  _normals_addr, _normals_size, GL_ARRAY_BUFFER_ARB)
        setupVBOBindless(ColorsData, ColorsSize, _colors, _colors_addr, _colors_size, GL_ARRAY_BUFFER_ARB)
        setupVBOBindless(SecColorsData, SecColorsSize, _secColors, _secColors_addr, _secColors_size, GL_ARRAY_BUFFER_ARB)
        setupVBOBindless(TexCoordsData, TexCoordsSize, _texCoords, _texCoords_addr, _texCoords_size, GL_ARRAY_BUFFER_ARB)
        setupVBOBindless(TexCoords1Data, TexCoords1Size, _texCoords1, _texCoords1_addr, _texCoords1_size, GL_ARRAY_BUFFER_ARB)
        setupVBOBindless(TexCoords2Data, TexCoords2Size, _texCoords2, _texCoords2_addr, _texCoords2_size, GL_ARRAY_BUFFER_ARB)
        setupVBOBindless(TexCoords3Data, TexCoords3Size, _texCoords3, _texCoords3_addr, _texCoords3_size, GL_ARRAY_BUFFER_ARB)
        setupVBOBindless(TexCoords4Data, TexCoords4Size, _texCoords4, _texCoords4_addr, _texCoords4_size, GL_ARRAY_BUFFER_ARB)
        setupVBOBindless(TexCoords5Data, TexCoords5Size, _texCoords5, _texCoords5_addr, _texCoords5_size, GL_ARRAY_BUFFER_ARB)
        setupVBOBindless(TexCoords6Data, TexCoords6Size, _texCoords6, _texCoords6_addr, _texCoords6_size, GL_ARRAY_BUFFER_ARB)
        setupVBOBindless(TexCoords7Data, TexCoords7Size, _texCoords7, _texCoords7_addr, _texCoords7_size, GL_ARRAY_BUFFER_ARB)
        setupVBOBindless(IndicesData, IndicesSize, _indices, _indices_addr, _indices_size, GL_ELEMENT_ARRAY_BUFFER_ARB)
    }
#endif

    _glBindBufferARB(GL_ARRAY_BUFFER_ARB, 0);
    _glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, 0);

    _cleared = false;
}

void GeoVBO::draw(void)
{
    if(_cleared)
        return;

#ifdef OSG_NV_BINDLESS_SUPPORT
    bool bindless = (GeoPumpFactory::_nvVertexBufferUnifiedMemory != Window::invalidExtensionID && GeoPumpFactory::_nvShaderBufferLoad != Window::invalidExtensionID);
#endif

    void (OSG_APIENTRY*_glBindBufferARB)(GLenum, GLuint) = reinterpret_cast<void (OSG_APIENTRY*)(GLenum, GLuint)>(_win->getFunction(GeoPumpFactory::_funcglBindBuffer));
    void (OSG_APIENTRY*_glClientActiveTextureARB) (GLenum type) = reinterpret_cast<void (OSG_APIENTRY*) (GLenum type)>( _win->getFunction( GeoPumpFactory::_funcglClientActiveTextureARB));
    void (OSG_APIENTRY*_glSecondaryColorPointerEXT) (GLint size,GLenum type,GLsizei stride,const GLvoid *pointer) = reinterpret_cast<void (OSG_APIENTRY*)(GLint size, GLenum type, GLsizei stride,const GLvoid *pointer)>(_win->getFunction(GeoPumpFactory::_funcglSecondaryColorPointer));
    void (OSG_APIENTRY *osgGLDrawRangeElementsEXT)(GLenum mode, GLuint start, GLuint end, GLsizei count, GLenum type, const GLvoid*indices) = reinterpret_cast<void (OSG_APIENTRY *)(GLenum mode, GLuint start, GLuint end, GLsizei count, GLenum type, const GLvoid*indices)>(_win->getFunction(GeoPumpFactory::_funcglDrawRangeElementsEXT));

    if(_glBindBufferARB == NULL)
        _glBindBufferARB = reinterpret_cast<void (OSG_APIENTRY*)(GLenum, GLuint)>(_win->getFunction(GeoPumpFactory::_funcglBindBufferARB));

    void (OSG_APIENTRY*_glBufferAddressRangeNV)(GLenum, GLuint, GLuint64, ptrdiff_t) = NULL;
    void (OSG_APIENTRY*_glVertexFormatNV)(GLint, GLenum, GLsizei) = NULL;
    void (OSG_APIENTRY*_glNormalFormatNV)(GLenum, GLsizei) = NULL;
    void (OSG_APIENTRY*_glColorFormatNV)(GLint, GLenum, GLsizei) = NULL;
    void (OSG_APIENTRY*_glSecondaryColorFormatNV)(GLint, GLenum, GLsizei) = NULL;
    void (OSG_APIENTRY*_glTexCoordFormatNV)(GLint, GLenum, GLsizei) = NULL;
    void (OSG_APIENTRY*_glIndexFormatNV)(GLenum, GLsizei) = NULL;

#ifdef OSG_NV_BINDLESS_SUPPORT
    if(bindless)
    {
        _glBufferAddressRangeNV = reinterpret_cast<void (OSG_APIENTRY*)(GLenum, GLuint, GLuint64, ptrdiff_t)>(_win->getFunction(GeoPumpFactory::_funcglBufferAddressRangeNV));
        _glVertexFormatNV = reinterpret_cast<void (OSG_APIENTRY*)(GLint, GLenum, GLsizei)>(_win->getFunction(GeoPumpFactory::_funcglVertexFormatNV));
        _glNormalFormatNV = reinterpret_cast<void (OSG_APIENTRY*)(GLenum, GLsizei)>(_win->getFunction(GeoPumpFactory::_funcglNormalFormatNV));
        _glColorFormatNV = reinterpret_cast<void (OSG_APIENTRY*)(GLint, GLenum, GLsizei)>(_win->getFunction(GeoPumpFactory::_funcglColorFormatNV));
        _glSecondaryColorFormatNV = reinterpret_cast<void (OSG_APIENTRY*)(GLint, GLenum, GLsizei)>(_win->getFunction(GeoPumpFactory::_funcglSecondaryColorFormatNV));
        _glTexCoordFormatNV = reinterpret_cast<void (OSG_APIENTRY*)(GLint, GLenum, GLsizei)>(_win->getFunction(GeoPumpFactory::_funcglTexCoordFormatNV));
        _glIndexFormatNV = reinterpret_cast<void (OSG_APIENTRY*)(GLenum, GLsizei)>(_win->getFunction(GeoPumpFactory::_funcglIndexFormatNV));

        if(_glBufferAddressRangeNV == NULL)
            bindless = false;
    }
#endif

    // removed the hasExtension(GeoPumpFactory::_extDrawRangeElements) and
    // hasExtension(GeoPumpFactory::_extMultitexture) calls with VBO support
    // this extensions are also supported!
    Geometry *geo = _geo;
    Geometry* shellGeo = _geo;
    Int16 modified=0;

    bool isBside = geo->isBSide();
    bool isShellGeo = geo->isShell();
    if( isShellGeo)
    {
        if( geo->getParents().size() > 0)
        {
            NodePtr parentNode = geo->getParents()[0];
            if( parentNode != NullFC)
            {
                for( size_t childIdx = 0; childIdx < parentNode->getNChildren(); ++childIdx)
                {
                    NodePtr child = parentNode->getChild(childIdx);
                    if( child != NullFC && child->getCore() != NullFC)
                    {
                        GeometryPtr childGeo = GeometryPtr::dcast(child->getCore());
                        if( childGeo != NullFC)
                        {
                            geo = childGeo.getCPtr();
                            break;
                        }
                    }
                }
            }
        }
        if( geo == shellGeo) // if we have not found a valid child
            return;
    }

#ifdef OSG_NV_BINDLESS_SUPPORT
    if(bindless)
    {
        glEnableClientState(GL_VERTEX_ATTRIB_ARRAY_UNIFIED_NV);
        glEnableClientState(GL_ELEMENT_ARRAY_UNIFIED_NV);
    }
#endif

    GeoPositionsPtr PositionsPtr = geo->getPositions();
    if (PositionsPtr != NullFC &&
        (_valid_draw_properties_mask & _draw_properties_mask & Geometry::MapPosition))
    {
        glEnableClientState(GL_VERTEX_ARRAY);
#ifdef OSG_NV_BINDLESS_SUPPORT
        if(bindless)
        {
            _glVertexFormatNV(PositionsPtr->getDimension(), PositionsPtr->getFormat(), 0);
            _glBufferAddressRangeNV(GL_VERTEX_ARRAY_ADDRESS_NV, 0, _positions_addr, _positions_size);
        }
        else
#endif
        {
            _glBindBufferARB(GL_ARRAY_BUFFER_ARB, _positions);
            glVertexPointer(PositionsPtr->getDimension(), PositionsPtr->getFormat(), 0, NULL);
        }
        modified|=(1<<0);
    }

    GeoNormalsPtr NormalsPtr = geo->getNormals();
    if (NormalsPtr != NullFC &&
        (_valid_draw_properties_mask & _draw_properties_mask & Geometry::MapNormal))
    {
        glEnableClientState(GL_NORMAL_ARRAY);
#ifdef OSG_NV_BINDLESS_SUPPORT
        if(bindless)
        {
            _glNormalFormatNV(NormalsPtr->getFormat(), 0);
            _glBufferAddressRangeNV(GL_NORMAL_ARRAY_ADDRESS_NV, 0, _normals_addr, _normals_size);
        }
        else
#endif
        {
            _glBindBufferARB(GL_ARRAY_BUFFER_ARB, _normals);
            glNormalPointer(NormalsPtr->getFormat(), 0, NULL);
        }
        modified|=(1<<3);
    }

    GeoColorsPtr ColorsPtr = geo->getColors();
    if (ColorsPtr != NullFC &&
        (_valid_draw_properties_mask & _draw_properties_mask & Geometry::MapColor))
    {
        glEnableClientState(GL_COLOR_ARRAY);
#ifdef OSG_NV_BINDLESS_SUPPORT
        if(bindless)
        {
            _glColorFormatNV(ColorsPtr->getDimension(), ColorsPtr->getFormat(), 0);
            _glBufferAddressRangeNV(GL_COLOR_ARRAY_ADDRESS_NV, 0, _colors_addr, _colors_size);
        }
        else
#endif
        {
            _glBindBufferARB(GL_ARRAY_BUFFER_ARB, _colors);
            glColorPointer(ColorsPtr->getDimension(), ColorsPtr->getFormat(), 0, NULL);
        }
        modified|=(1<<1);
    }

    GeoColorsPtr SecColorsPtr = geo->getSecondaryColors();
    if (SecColorsPtr != NullFC &&
        (_valid_draw_properties_mask & _draw_properties_mask & Geometry::MapSecondaryColor))
    {
        glEnableClientState(GL_SECONDARY_COLOR_ARRAY_EXT);
#ifdef OSG_NV_BINDLESS_SUPPORT
        if(bindless)
        {
            _glSecondaryColorFormatNV(SecColorsPtr->getDimension(), SecColorsPtr->getFormat(), 0);
            _glBufferAddressRangeNV(GL_SECONDARY_COLOR_ARRAY_ADDRESS_NV, 0, _secColors_addr, _secColors_size);
        }
        else
#endif
        {
            _glBindBufferARB(GL_ARRAY_BUFFER_ARB, _secColors);
            _glSecondaryColorPointerEXT(SecColorsPtr->getDimension(), SecColorsPtr->getFormat(), 0, NULL);
        }
        modified|=(1<<2);
    }

    GeoTexCoordsPtr TexCoordsPtr = geo->getTexCoords();
    if (TexCoordsPtr != NullFC &&
        (_valid_draw_properties_mask & _draw_properties_mask & Geometry::MapTexCoords))
    {
        _glClientActiveTextureARB(GL_TEXTURE0_ARB);
        glEnableClientState(GL_TEXTURE_COORD_ARRAY);

#ifdef OSG_NV_BINDLESS_SUPPORT
        if(bindless)
        {
            _glTexCoordFormatNV(TexCoordsPtr->getDimension(), TexCoordsPtr->getFormat(), 0);
            _glBufferAddressRangeNV(GL_TEXTURE_COORD_ARRAY_ADDRESS_NV, 0, _texCoords_addr, _texCoords_size);
        }
        else
#endif
        {
            _glBindBufferARB(GL_ARRAY_BUFFER_ARB, _texCoords);
            glTexCoordPointer (TexCoordsPtr->getDimension(), TexCoordsPtr->getFormat(), 0, NULL);
        }
        modified|=(1<<4);
    }
    else
    {
        // special hack if we render a geometry without texture coordinates and with a material with enabled texture
        // we get flicker rendering because of random (last global texcoord state) texcoord values in our shaders.
        // so we upload one zero tex coord!
        glTexCoord2f(0.0f, 0.0f);
    }

    GeoTexCoordsPtr TexCoords1Ptr = geo->getTexCoords1();
    if (TexCoords1Ptr != NullFC &&
        (_valid_draw_properties_mask & _draw_properties_mask & Geometry::MapTexCoords1))
    {
        _glClientActiveTextureARB(GL_TEXTURE1_ARB);
        glEnableClientState(GL_TEXTURE_COORD_ARRAY);

#ifdef OSG_NV_BINDLESS_SUPPORT
        if(bindless)
        {
            _glTexCoordFormatNV(TexCoords1Ptr->getDimension(), TexCoords1Ptr->getFormat(), 0);
            _glBufferAddressRangeNV(GL_TEXTURE_COORD_ARRAY_ADDRESS_NV, 0, _texCoords1_addr, _texCoords1_size);
        }
        else
#endif
        {
            _glBindBufferARB(GL_ARRAY_BUFFER_ARB, _texCoords1);
            glTexCoordPointer (TexCoords1Ptr->getDimension(), TexCoords1Ptr->getFormat(), 0, NULL);
        }
        modified|=(1<<5);
    }

    GeoTexCoordsPtr TexCoords2Ptr = geo->getTexCoords2();
    if (TexCoords2Ptr != NullFC && (_valid_draw_properties_mask & _draw_properties_mask & Geometry::MapTexCoords2))
    {
        _glClientActiveTextureARB(GL_TEXTURE2_ARB);
        glEnableClientState(GL_TEXTURE_COORD_ARRAY);

#ifdef OSG_NV_BINDLESS_SUPPORT
        if(bindless)
        {
            _glTexCoordFormatNV(TexCoords2Ptr->getDimension(), TexCoords2Ptr->getFormat(), 0);
            _glBufferAddressRangeNV(GL_TEXTURE_COORD_ARRAY_ADDRESS_NV, 0, _texCoords2_addr, _texCoords2_size);
        }
        else
#endif
        {
            _glBindBufferARB(GL_ARRAY_BUFFER_ARB, _texCoords2);
            glTexCoordPointer (TexCoords2Ptr->getDimension(), TexCoords2Ptr->getFormat(), 0, NULL);
        }
        modified|=(1<<6);
    }

    GeoTexCoordsPtr TexCoords3Ptr = geo->getTexCoords3();
    if (TexCoords3Ptr != NullFC && (_valid_draw_properties_mask & _draw_properties_mask & Geometry::MapTexCoords3))
    {
        _glClientActiveTextureARB(GL_TEXTURE3_ARB);
        glEnableClientState(GL_TEXTURE_COORD_ARRAY);

#ifdef OSG_NV_BINDLESS_SUPPORT
        if(bindless)
        {
            _glTexCoordFormatNV(TexCoords3Ptr->getDimension(), TexCoords3Ptr->getFormat(), 0);
            _glBufferAddressRangeNV(GL_TEXTURE_COORD_ARRAY_ADDRESS_NV, 0, _texCoords3_addr, _texCoords3_size);
        }
        else
#endif
        {
            _glBindBufferARB(GL_ARRAY_BUFFER_ARB, _texCoords3);
            glTexCoordPointer (TexCoords3Ptr->getDimension(), TexCoords3Ptr->getFormat(), 0, NULL);
        }
        modified|=(1<<7);
    }

    GeoTexCoordsPtr TexCoords4Ptr = geo->getTexCoords4();
    if (TexCoords4Ptr != NullFC &&
        (_valid_draw_properties_mask & _draw_properties_mask & Geometry::MapTexCoords4))
    {
        _glClientActiveTextureARB(GL_TEXTURE4_ARB);
        glEnableClientState(GL_TEXTURE_COORD_ARRAY);

#ifdef OSG_NV_BINDLESS_SUPPORT
        if(bindless)
        {
            _glTexCoordFormatNV(TexCoords4Ptr->getDimension(), TexCoords4Ptr->getFormat(), 0);
            _glBufferAddressRangeNV(GL_TEXTURE_COORD_ARRAY_ADDRESS_NV, 0, _texCoords4_addr, _texCoords4_size);
        }
        else
#endif
        {
            _glBindBufferARB(GL_ARRAY_BUFFER_ARB, _texCoords4);
            glTexCoordPointer (TexCoords4Ptr->getDimension(), TexCoords4Ptr->getFormat(), 0, NULL);
        }
        modified|=(1<<8);
    }

    GeoTexCoordsPtr TexCoords5Ptr = geo->getTexCoords5();
    if (TexCoords5Ptr != NullFC &&
        (_valid_draw_properties_mask & _draw_properties_mask & Geometry::MapTexCoords5))
    {
        _glClientActiveTextureARB(GL_TEXTURE5_ARB);
        glEnableClientState(GL_TEXTURE_COORD_ARRAY);

#ifdef OSG_NV_BINDLESS_SUPPORT
        if(bindless)
        {
            _glTexCoordFormatNV(TexCoords5Ptr->getDimension(), TexCoords5Ptr->getFormat(), 0);
            _glBufferAddressRangeNV(GL_TEXTURE_COORD_ARRAY_ADDRESS_NV, 0, _texCoords5_addr, _texCoords5_size);
        }
        else
#endif
        {
            _glBindBufferARB(GL_ARRAY_BUFFER_ARB, _texCoords5);
            glTexCoordPointer (TexCoords5Ptr->getDimension(), TexCoords5Ptr->getFormat(), 0, NULL);
        }
        modified|=(1<<9);
    }

    GeoTexCoordsPtr TexCoords6Ptr = geo->getTexCoords6();
    if (TexCoords6Ptr != NullFC &&
        (_valid_draw_properties_mask & _draw_properties_mask & Geometry::MapTexCoords6))
    {
        _glClientActiveTextureARB(GL_TEXTURE6_ARB);
        glEnableClientState(GL_TEXTURE_COORD_ARRAY);

#ifdef OSG_NV_BINDLESS_SUPPORT
        if(bindless)
        {
            _glTexCoordFormatNV(TexCoords6Ptr->getDimension(), TexCoords6Ptr->getFormat(), 0);
            _glBufferAddressRangeNV(GL_TEXTURE_COORD_ARRAY_ADDRESS_NV, 0, _texCoords6_addr, _texCoords6_size);
        }
        else
#endif
        {
            _glBindBufferARB(GL_ARRAY_BUFFER_ARB, _texCoords6);
            glTexCoordPointer (TexCoords6Ptr->getDimension(), TexCoords6Ptr->getFormat(), 0, NULL);
        }
        modified|=(1<<10);
    }

    GeoTexCoordsPtr TexCoords7Ptr = geo->getTexCoords7();
    if (TexCoords7Ptr != NullFC &&
        (_valid_draw_properties_mask & _draw_properties_mask & Geometry::MapTexCoords7))
    {
        _glClientActiveTextureARB(GL_TEXTURE7_ARB);
        glEnableClientState(GL_TEXTURE_COORD_ARRAY);

#ifdef OSG_NV_BINDLESS_SUPPORT
        if(bindless)
        {
            _glTexCoordFormatNV(TexCoords7Ptr->getDimension(), TexCoords7Ptr->getFormat(), 0);
            _glBufferAddressRangeNV(GL_TEXTURE_COORD_ARRAY_ADDRESS_NV, 0, _texCoords7_addr, _texCoords7_size);
        }
        else
#endif
        {
            _glBindBufferARB(GL_ARRAY_BUFFER_ARB, _texCoords7);
            glTexCoordPointer (TexCoords7Ptr->getDimension(), TexCoords7Ptr->getFormat(), 0, NULL);
        }
        modified|=(1<<11);
    }
    // amz special nvidia HACK we need to upload something to tex coord 5-8!
    // without this we get strange memory allocations!
#if 1
    else if(PositionsPtr != NullFC)
    {
        _glClientActiveTextureARB(GL_TEXTURE7_ARB);
        glEnableClientState(GL_TEXTURE_COORD_ARRAY);

#ifdef OSG_NV_BINDLESS_SUPPORT
        if(bindless)
        {
            _glTexCoordFormatNV(PositionsPtr->getDimension(), PositionsPtr->getFormat(), 0);
            _glBufferAddressRangeNV(GL_TEXTURE_COORD_ARRAY_ADDRESS_NV, 0, _texCoords7_addr, _texCoords7_size);
        }
        else
#endif
        {
            _glBindBufferARB(GL_ARRAY_BUFFER_ARB, _positions);
            glTexCoordPointer (PositionsPtr->getDimension(), PositionsPtr->getFormat(), 0, NULL);
        }
        modified|=(1<<11);
    }
#endif

#ifdef OSG_NV_BINDLESS_SUPPORT
    if(!bindless)
#endif
    {
        _glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, _indices);
    }

    if( isShellGeo)
    {
        if( _valid_draw_properties_mask)
        {
            // for a shell, assume we only have triangles
            osgGLDrawRangeElementsEXT(GL_TRIANGLES,  0, _maxIndex, _numIndices, GL_UNSIGNED_INT, NULL);
        }

    }
    else
    {
        pumpSetup(Lengths    , GeoPLengthsPtr  , getLengths    );
        pumpSetup(Types      , GeoPTypesPtr    , getTypes      );
        pumpSetup(Indices    , GeoIndicesPtr   , getIndices    );
        IndicesData = IndicesData; // workaround for irix compiler.

#ifdef OSG_NV_BINDLESS_SUPPORT
        if(bindless)
        {
            //_glIndexFormatNV(IndicesPtr->getFormat(), 0);
            _glBufferAddressRangeNV(GL_ELEMENT_ARRAY_ADDRESS_NV, 0, _indices_addr, _indices_size);
        }
#endif

        UInt32 lendummy;
        UInt32 LengthsSize;
        bool   len16 = false;

        if(LengthsPtr != NullFC && LengthsData != NULL)
        {
            LengthsSize = LengthsPtr->getSize();
            len16 = (LengthsPtr->getFormatSize() == 2);
        }
        else
        {
            LengthsSize = 1;
            LengthsData = reinterpret_cast<GLubyte *>(&lendummy);
            lendummy = PositionsPtr->getSize();
        }

        UInt32 LengthsInd = 0, TypesInd = 0, IndicesInd = 0,
            IndicesSize = IndicesPtr->getSize();

        if(len16 == false)
        {
            for(LengthsInd = 0; LengthsInd < LengthsSize; LengthsInd++)
            {
                UInt32 count  = *reinterpret_cast<UInt32*>(LengthsData + LengthsInd * LengthsStride);
                // for vbo's we need only a offset!
                UInt8 *vind = static_cast<UInt8 *>(NULL) + IndicesStride * IndicesInd;
                IndicesInd += count;

                if(count != 0)
                {
                    osgGLDrawRangeElementsEXT(*(TypesData + TypesInd++ * TypesStride),
                        geo->getLowindices(LengthsInd),
                        geo->getHighindices(LengthsInd),
                        count,
                        IndicesPtr->getFormat(),
                        vind);
                }
            }
        }
        else if(len16 == true)
        {
            for(LengthsInd = 0; LengthsInd < LengthsSize; LengthsInd++)
            {
                UInt16 count  = *reinterpret_cast<UInt16*>(LengthsData + LengthsInd * LengthsStride);
                // for vbo's we need only a offset!
                UInt8 *vind = static_cast<UInt8 *>(NULL) + IndicesStride * IndicesInd;
                IndicesInd += count;

                if(count != 0)
                {
                    osgGLDrawRangeElementsEXT(*(TypesData + TypesInd++ * TypesStride), geo->getLowindices(LengthsInd), geo->getHighindices(LengthsInd), count, IndicesPtr->getFormat(), vind);
                }
            }
        }
        else
        {
            FWARNING(("VBO::draw: unknown Lengths format size\n"));
        }
    }

    if(modified&(1<<0))
        glDisableClientState(GL_VERTEX_ARRAY);
    if(modified&(1<<1))
        glDisableClientState(GL_COLOR_ARRAY);
    if(modified&(1<<2))
        glDisableClientState(GL_SECONDARY_COLOR_ARRAY_EXT);
    if(modified&(1<<3))
        glDisableClientState(GL_NORMAL_ARRAY);

    if(modified&(1<<4))
    {
        _glClientActiveTextureARB(GL_TEXTURE0_ARB);
        glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    }

    if(modified&(1<<5))
    {
        _glClientActiveTextureARB(GL_TEXTURE1_ARB);
        glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    }

    if(modified&(1<<6))
    {
        _glClientActiveTextureARB(GL_TEXTURE2_ARB);
        glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    }

    if(modified&(1<<7))
    {
        _glClientActiveTextureARB(GL_TEXTURE3_ARB);
        glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    }

    if(modified&(1<<8))
    {
        _glClientActiveTextureARB(GL_TEXTURE4_ARB);
        glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    }

    if(modified&(1<<9))
    {
        _glClientActiveTextureARB(GL_TEXTURE5_ARB);
        glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    }

    if(modified&(1<<10))
    {
        _glClientActiveTextureARB(GL_TEXTURE6_ARB);
        glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    }

    if(modified&(1<<11))
    {
        _glClientActiveTextureARB(GL_TEXTURE7_ARB);
        glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    }

#ifdef OSG_NV_BINDLESS_SUPPORT
    if(bindless)
    {
        glDisableClientState(GL_VERTEX_ATTRIB_ARRAY_UNIFIED_NV);
        glDisableClientState(GL_ELEMENT_ARRAY_UNIFIED_NV);
    }
    else
#endif
    {
        _glBindBufferARB(GL_ARRAY_BUFFER_ARB, 0);
        _glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, 0);
    }
}

#endif

/*------------------------------------------------------------------------*/
/*                              cvs id's                                  */

#ifdef OSG_SGI_CC
#pragma set woff 1174
#endif

#ifdef OSG_LINUX_ICC
#pragma warning( disable : 177 )
#endif

namespace
{
    static Char8 cvsid_cpp       [] = "@(#)$Id: $";
    static Char8 cvsid_hpp       [] = OSGGEOPUMPFACTORY_HEADER_CVSID;
    static Char8 cvsid_inl       [] = OSGGEOPUMPFACTORY_INLINE_CVSID;
}

