/*---------------------------------------------------------------------------*\
 *                                OpenSG                                     *
 *                                                                           *
 *                                                                           *
 *             Copyright (C) 2000-2002 by the OpenSG Forum                   *
 *                                                                           *
 *                            www.opensg.org                                 *
 *                                                                           *
 *   contact: dirk@opensg.org, gerrit.voss@vossg.org, jbehr@zgdv.de          *
 *                                                                           *
\*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*\
 *                                License                                    *
 *                                                                           *
 * This library is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU Library General Public License as published    *
 * by the Free Software Foundation, version 2.                               *
 *                                                                           *
 * This library is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU         *
 * Library General Public License for more details.                          *
 *                                                                           *
 * You should have received a copy of the GNU Library General Public         *
 * License along with this library; if not, write to the Free Software       *
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 *
 *                                                                           *
\*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*\
 *                                Changes                                    *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *                                                                           *
\*---------------------------------------------------------------------------*/

#include <stdlib.h>
#include <stdio.h>

#include <OSGConfig.h>

#include "OSGNodePtr.h"
#include "OSGMaterialPool.h"
#include "OSGMaterialGroup.h"
#include "OSGMaterialDrawable.h"

OSG_USING_NAMESPACE

/*! \class osg::MaterialPool
*/

/*-------------------------------------------------------------------------*/
/*                                Sync                                     */

void MaterialPool::changed(BitVector whichField, UInt32 origin)
{
    if(whichField & RootFieldMask)
    {
        if(origin & ChangedOrigin::Abstract)
        {
            if(origin & ChangedOrigin::AbstrIncRefCount)
            {
                addRefCP(_sfRoot.getValue());
            }
            else
            {
                NodePtr root = _sfRoot.getValue();

                _sfRoot.setValue(NullFC);

                setRoot(root);
            }
        }
    }

    if(whichField & LightRootFieldMask)
    {
        if(origin & ChangedOrigin::Abstract)
        {
            if(origin & ChangedOrigin::AbstrIncRefCount)
            {
                addRefCP(_sfLightRoot.getValue());
            }
            else
            {
                NodePtr root = _sfLightRoot.getValue();

                _sfLightRoot.setValue(NullFC);

                setLightRoot(root);
            }
        }
    }
    if(whichField & OverlayRootFieldMask)
    {
        if(origin & ChangedOrigin::Abstract)
        {
            if(origin & ChangedOrigin::AbstrIncRefCount)
            {
                addRefCP(_sfOverlayRoot.getValue());
            }
            else
            {
                NodePtr root = _sfOverlayRoot.getValue();
                
                _sfOverlayRoot.setValue(NullFC);
                
                setOverlayRoot(root);
            }
        }
    }
    if(origin & ChangedOrigin::Abstract)
    {
        MFMaterialPtr::iterator matIt  = _mfMaterials.begin();
        MFMaterialPtr::iterator matEnd = _mfMaterials.end ();

        while(matIt != matEnd)
        {
            addRefCP(*matIt);
            ++matIt;
        }
    }

    if(whichField & MaterialsFieldMask)
    {
        sync();
    }
    Inherited::changed(whichField, origin);
}

/*-------------------------------------------------------------------------*/
/*                                Dump                                     */

void MaterialPool::dump(      UInt32    uiIndent,
                         const BitVector bvFlags) const
{
    Inherited::dump(uiIndent, bvFlags);
}

/*-------------------------------------------------------------------------*/
/*                            Constructors                                 */

MaterialPool::MaterialPool(void) :
    Inherited(),
    _mats()
{
}

MaterialPool::MaterialPool(const MaterialPool &source) :
    Inherited(source),
    _mats(source._mats)
{
}

/*-------------------------------------------------------------------------*/
/*                             Destructor                                  */

MaterialPool::~MaterialPool(void)
{
    subRefCP(_sfRoot.getValue());
    subRefCP(_sfLightRoot.getValue());
    subRefCP(_sfOverlayRoot.getValue());
}

void MaterialPool::onDestroy(void)
{
    // if we're in shutdown this is the prototype ...
    if(OSG::GlobalSystemState == OSG::Shutdown)
        return;

    MFMaterialPtr::iterator       matIt        = _mfMaterials.begin();
    MFMaterialPtr::const_iterator endMaterials = _mfMaterials.end  ();

    FINFO(("MaterialPool::onDestroy : subrefing %u materials\n", _mfMaterials.size()));
    while(matIt != endMaterials)
    {
        subRefCP(*matIt);
        ++matIt;
    }
}

/*-------------------------------------------------------------------------*/
/*                                Init                                     */

void MaterialPool::initMethod(void)
{
}


void MaterialPool::sync(void)
{
    // should be safe as the materials interface is protected.
    if(_mfMaterials.size() == _mats.size())
    {
        FINFO(("MaterialPool::sync : %u materials already synced, nothing to do.\n", _mfMaterials.size()));
        return;
    }

    MFMaterialPtr::iterator       matIt        = _mfMaterials.begin();
    MFMaterialPtr::const_iterator endMaterials = _mfMaterials.end  ();

    while(matIt != endMaterials)
    {
        // this is already done in the osb loader!
        //addRefCP(*matIt);
        _mats.insert(*matIt);
        ++matIt;
    }
    FINFO(("MaterialPool::sync : synced %u materials\n", _mfMaterials.size()));
}

/*-------------------------------------------------------------------------*/
/*                             Access                                      */

bool MaterialPool::add(const MaterialPtr &mat)
{
    return add(mat, NullFC, true, 0);
}

bool MaterialPool::add(const MaterialPtr &mat, const NodePtr &parent, bool create_node, const UInt32 pos)
{
    if(mat == NullFC)
        return false;

    if(_mats.count(mat) > 0)
        return false;

    addRefCP(mat);
    _mats.insert(mat);
    _mfMaterials.push_back(mat);

    NodePtr parent_group = parent;

    if(parent_group == NullFC)
    {
        if(getRoot() == NullFC)
        {
            NodePtr root = Node::create();
            beginEditCP(root);
                root->setCore(Group::create());
            endEditCP(root);
            MaterialPoolPtr tmp(*this);
            beginEditCP(tmp, RootFieldMask);
                tmp->setRoot(root);
            endEditCP(tmp, RootFieldMask);
        }
        if(getRoot()->getNChildren() == 0)
        {
            NodePtr child0 = Node::create();
            beginEditCP(child0);
                child0->setCore(Group::create());
            endEditCP(child0);
            beginEditCP(getRoot(), RootFieldMask);
                getRoot()->addChild(child0);
            endEditCP(getRoot(), RootFieldMask);
        }
        parent_group = getRoot()->getChild(0);
    }

    if(create_node)
    {
        MaterialGroupPtr mgc = MaterialGroup::create();
        beginEditCP(mgc);
            mgc->setMaterial(mat);
        endEditCP(mgc);
        NodePtr mg = Node::create();
        beginEditCP(mg);
            mg->setCore(mgc);
        endEditCP(mg);
        beginEditCP(parent_group, Node::ChildrenFieldMask);
            parent_group->insertChild( (pos <= parent_group->getNChildren()) ? pos : 0 , mg);
        endEditCP(parent_group, Node::ChildrenFieldMask);
    }

    return true;
}

void MaterialPool::add(const NodePtr &root)
{
    if(root == NullFC)
        return;

    traverse(root, std::bind(&MaterialPool::addMaterialCB, this, std::placeholders::_1));
}

void MaterialPool::add(const MaterialPoolPtr &mp)
{
    if(mp == NullFC)
        return;

    if(mp == getPtr())
        return;

    if(getRoot() == NullFC)
    {
        NodePtr root = Node::create();
        beginEditCP(root);
            root->setCore(Group::create());
        endEditCP(root);
        MaterialPoolPtr tmp(*this);
        beginEditCP(tmp, RootFieldMask);
            tmp->setRoot(root);
        endEditCP(tmp, RootFieldMask);
    }
    if(getRoot()->getNChildren() == 0)
    {
        NodePtr child0 = Node::create();
        beginEditCP(child0);
            child0->setCore(Group::create());
        endEditCP(child0);
        beginEditCP(getRoot(), RootFieldMask);
            getRoot()->addChild(child0);
        endEditCP(getRoot(), RootFieldMask);
    }
   
    bool create_node = false;
    if(mp->getRoot() != NullFC && mp->getRoot()->getNChildren() > 0 && mp->getRoot()->getChild(0)->getNChildren() > 0)
    {
        beginEditCP(getRoot());
            for(UInt32 i=0;i<mp->getRoot()->getChild(0)->getNChildren();++i)
                getRoot()->getChild(0)->addChild(OSG::deepCloneTree(mp->getRoot()->getChild(0)->getChild(i), "Material"));
        endEditCP(getRoot());
    }
    else
    {
        // there was no root nor children so we add all materials to the root node later.
        create_node = true;
    }

    if(getLightRoot() == NullFC)
    {
        NodePtr root = Node::create();
        beginEditCP(root);
        root->setCore(Group::create());
        root->setFlags(Node::NFAuxiliary);
        endEditCP(root);
        
        MaterialPoolPtr tmp(*this);
        beginEditCP(tmp, LightRootFieldMask);
        tmp->setLightRoot(root);
        endEditCP(tmp, LightRootFieldMask);
    }
    
    if(mp->getLightRoot() != NullFC && mp->getLightRoot()->getNChildren() > 0)
    {
        beginEditCP(getLightRoot());

        while (mp->getLightRoot()->getNChildren())
        {
            //getLightRoot()->addChild(OSG::deepCloneTree(mp->getLightRoot()->getChild(i), "Light"));
            // we can not copy these since this would break the connection of the nodes in the associated field
            getLightRoot()->addChild( mp->getLightRoot()->getChild(0));
        }
        endEditCP(getLightRoot());
    }
    
    if(getOverlayRoot() == NullFC)
    {
        NodePtr root = Node::create();
        beginEditCP(root);
        root->setCore(Group::create());
        root->setFlags(Node::NFAuxiliary);
        endEditCP(root);
        
        MaterialPoolPtr tmp(*this);
        beginEditCP(tmp, OverlayRootFieldMask);
        tmp->setOverlayRoot(root);
        endEditCP(tmp, OverlayRootFieldMask);
    }
    
    if(mp->getOverlayRoot() != NullFC && mp->getOverlayRoot()->getNChildren() > 0)
    {
        beginEditCP(getOverlayRoot());
        
        while (mp->getOverlayRoot()->getNChildren())
        {
            //getLightRoot()->addChild(OSG::deepCloneTree(mp->getLightRoot()->getChild(i), "Light"));
            // we can not copy these since this would break the connection of the nodes in the associated field
            getOverlayRoot()->addChild( mp->getOverlayRoot()->getChild(0));
        }
        endEditCP(getOverlayRoot());
    }
    
    
    MFMaterialPtr::iterator       matIt        = mp->_mfMaterials.begin();
    MFMaterialPtr::const_iterator endMaterials = mp->_mfMaterials.end  ();

    while(matIt != endMaterials)
    {
        add(*matIt, NullFC, create_node, 0);
        ++matIt;
    }
}

Int32 MaterialPool::find(const MaterialPtr &mat) const
{
    if(mat == NullFC)
        return -1;

    if(_mats.count(mat) == 0)
        return -1;

    UInt32 index;

    for(index = 0; index < _mfMaterials.size(); ++index)
    {
        if(_mfMaterials[index] == mat)
            return index;
    }
    return -1;
}

NodePtr MaterialPool::find(const NodePtr &node, const MaterialPtr &mat) const
{
    if(node == NullFC || mat == NullFC)
        return NullFC;

    NodePtr r = NullFC;
    MaterialGroupPtr mgc = MaterialGroupPtr::dcast(node->getCore());

    if(mgc != NullFC && mgc->getMaterial() == mat)
        return node;

    for(UInt32 i=0;i<node->getNChildren();++i)
    {
        r = find(node->getChild(i), mat);
        if(r != NullFC)
            break;
    }

    return r;
}

void MaterialPool::find(std::vector<NodePtr> &nodes, const NodePtr &node, const MaterialPtr &mat) const
{
    if(node == NullFC || mat == NullFC)
        return;

    MaterialGroupPtr mgc = MaterialGroupPtr::dcast(node->getCore());

    if(mgc != NullFC && mgc->getMaterial() == mat)
        nodes.push_back(node);

    for(UInt32 i=0;i<node->getNChildren();++i)
    {
        find(nodes, node->getChild(i), mat);
    }
}

void MaterialPool::replace(const MaterialPtr &old_mat, const MaterialPtr &new_mat)
{
    if(old_mat == NullFC || new_mat == NullFC || old_mat == new_mat)
        return;

    if(_mats.count(old_mat) == 0)
    {
        SWARNING << "MaterialPool(" << this << ")::replace: " << old_mat
                 << " is not one of my materials!" << std::endl;
        return;
    }

    MFMaterialPtr::iterator matIt = _mfMaterials.find(old_mat);

    if(matIt != _mfMaterials.end())
    {
        bool new_mat_referenced = (_mats.count(new_mat) > 0);
        if(new_mat_referenced)
        {
            // the new mat is already in the pool so we increase the ref count to keep it alive.
            // and remove it from the pool.
            addRefCP(new_mat);
            sub(new_mat); // might invalidate matIt as new_mat is erased from _mfMaterials
            matIt = _mfMaterials.find(old_mat); // get valid iterator 
            if(matIt == _mfMaterials.end())
                return;
        }

        std::vector<NodePtr> nodes;
        find(nodes, getRoot(), old_mat);

        for(size_t i = 0; i < nodes.size(); i++)
        {
            NodePtr node = nodes[i];
            if(node != NullFC)
            {
                MaterialGroupPtr mg = MaterialGroupPtr::dcast(node->getCore());
                if(mg != NullFC)
                {
                    beginEditCP(mg);
                    mg->setMaterial(new_mat);
                    endEditCP(mg);
                }
            }
        }
        _mats.erase(old_mat);
        subRefCP(old_mat);

        *matIt = new_mat;
        if(!new_mat_referenced)
            addRefCP(new_mat);
        _mats.insert(new_mat);
    }
    else
    {
        SWARNING << "MaterialPool(" << this << ")::replace: " << old_mat
                 << " is not one of my materials!" << std::endl;
    }
}

void MaterialPool::sub(const MaterialPtr &mat)
{
    if(mat == NullFC)
        return;

    if(_mats.count(mat) == 0)
    {
        SWARNING << "MaterialPool(" << this << ")::sub: " << mat
                 << " is not one of my materials!" << std::endl;
        return;
    }

    MFMaterialPtr::iterator matIt = _mfMaterials.find(mat);

    if(matIt != _mfMaterials.end())
    {
        std::vector<NodePtr> nodes;
        find(nodes, getRoot(), mat);

        for(size_t i = 0; i < nodes.size(); i++)
        {
            NodePtr node = nodes[i];
            if(node != NullFC)
            {
                NodePtr parent = node->getParent();
                if(parent != NullFC)
                {
                    beginEditCP(parent, Node::ChildrenFieldMask);
                        parent->subChild(node);
                    endEditCP(parent, Node::ChildrenFieldMask);
                }
                else
                {
                    SWARNING << "MaterialPool(" << this << ")::sub: " << mat
                             << " node parent is NULL!" << std::endl;
                }
            }
        }
        _mats.erase(mat);
        subRefCP(mat);

        _mfMaterials.erase(matIt);
    }
    else
    {
        SWARNING << "MaterialPool(" << this << ")::sub: " << mat
                 << " is not one of my materials!" << std::endl;
    }
}

void MaterialPool::sub(const NodePtr &root)
{
    if(root == NullFC)
        return;

    traverse(root, std::bind(&MaterialPool::subMaterialCB, this, std::placeholders::_1));
}

void MaterialPool::sub(const MaterialPoolPtr &mp)
{
    if(mp == NullFC)
        return;

    if(mp == getPtr())
        return;

    MFMaterialPtr::iterator       matIt        = mp->_mfMaterials.begin();
    MFMaterialPtr::const_iterator endMaterials = mp->_mfMaterials.end  ();

    while(matIt != endMaterials)
    {
        sub(*matIt);
        ++matIt;
    }
}

void MaterialPool::sub(UInt32 index)
{
    MFMaterialPtr::iterator matIt = _mfMaterials.begin();

    matIt += index;

    if(matIt != _mfMaterials.end())
    {
        subRefCP(*matIt);
        _mats.erase(*matIt);
        _mfMaterials.erase(matIt);
    }
}

void MaterialPool::get(std::vector<MaterialPtr> &mats)
{
    mats.clear();
    mats.reserve(_mfMaterials.getSize());
    MFMaterialPtr::iterator       matIt        = _mfMaterials.begin();
    MFMaterialPtr::const_iterator endMaterials = _mfMaterials.end  ();

    while(matIt != endMaterials)
    {
        mats.push_back(*matIt);
        ++matIt;
    }
}

const std::set<MaterialPtr> &MaterialPool::get(void)
{
    return _mats;
}

void MaterialPool::clear(void)
{
    MFMaterialPtr::iterator       matIt        = _mfMaterials.begin();
    MFMaterialPtr::const_iterator endMaterials = _mfMaterials.end  ();

    while(matIt != endMaterials)
    {
        subRefCP(*matIt);
        ++matIt;
    }
    _mfMaterials.clear();
    _mats.clear();
    subRefCP(getRoot());
    setRoot(NullFC);

    subRefCP(getLightRoot());
    setLightRoot(NullFC);
    
    subRefCP(getOverlayRoot());
    setOverlayRoot(NullFC);
}

Action::ResultE MaterialPool::addMaterialCB(const NodePtr &node)
{
    FieldContainerPtr core = node->getCore();

    if(core == NullFC)
        return Action::Continue;

    MaterialGroupPtr mg = MaterialGroupPtr::dcast(core);
    if(mg != NullFC)
    {
        // ignore the materials form our light sources geometry!
        FieldContainerType *value_pair_type = FieldContainerFactory::the()->findType("ValuePair");
        if(value_pair_type != NULL)
        {
            AttachmentPtr vp = node->findAttachment(*value_pair_type);
            if(vp != NullFC)
            {
                Field *field = vp->getField("key");
                if(field != NULL)
                {
                     MFString *keys = dynamic_cast<MFString *>(field);
                     if(keys != NULL)
                     {
                         if(!keys->empty())
                         {
                             if(keys->getValue(0) == "LightBeaconGroup")
                             {
                                 // we skip all children!
                                 return Action::Skip;
                             }
                         }
                     }
                }
            }
        }
        add(mg->getMaterial(), NullFC, false, 0);
        return Action::Continue;
    }

    MaterialPoolPtr mp = MaterialPoolPtr::dcast(core);
    if(mp != NullFC)
    {
        add(mp);
        return Action::Continue;
    }

    MaterialDrawablePtr md = MaterialDrawablePtr::dcast(core);
    if(md != NullFC)
    {
        add(md->getMaterial(), NullFC, false, 0);
        return Action::Continue;
    }

    const char *typename2 = core->getType().getName().str();
    if(!strcmp(typename2, "DVRVolume"))
    {
        Field *field = core->getField("renderMaterial");
        if(field != NULL)
        {
            add(MaterialPtr::dcast((static_cast<SFFieldContainerPtr *>(field))->getValue()), NullFC, false, 0);
        }
    }

    return Action::Continue;
}

Action::ResultE MaterialPool::subMaterialCB(const NodePtr &node)
{
    FieldContainerPtr core = node->getCore();

    if(core == NullFC)
        return Action::Continue;

    MaterialGroupPtr mg = MaterialGroupPtr::dcast(core);
    if(mg != NullFC)
    {
        sub(mg->getMaterial());
        return Action::Continue;
    }

    // ignore MaterialPool node!
#if 0
    MaterialPoolPtr mp = MaterialPoolPtr::dcast(core);
    if(mp != NullFC)
    {
        sub(mp);
        return Action::Continue;
    }
#endif

    MaterialDrawablePtr md = MaterialDrawablePtr::dcast(core);
    if(md != NullFC)
    {
        sub(md->getMaterial());
        return Action::Continue;
    }

    const char *typename2 = core->getType().getName().str();
    if(!strcmp(typename2, "DVRVolume"))
    {
        Field *field = core->getField("renderMaterial");
        if(field != NULL)
        {
            sub(MaterialPtr::dcast((static_cast<SFFieldContainerPtr *>(field))->getValue()));
        }
    }

    return Action::Continue;
}

/*-------------------------------------------------------------------------*/
/*                              cvs id's                                   */

#ifdef __sgi
#pragma set woff 1174
#endif

#ifdef OSG_LINUX_ICC
#pragma warning( disable : 177 )
#endif

namespace
{
    static Char8 cvsid_cpp[] = "@(#)$Id: OSGMaterialPool.cpp,v 1.2 2008/06/05 05:01:18 vossg Exp $";
    static Char8 cvsid_hpp[] = OSGMATERIALPOOL_HEADER_CVSID;
    static Char8 cvsid_inl[] = OSGMATERIALPOOL_INLINE_CVSID;
}
